/*======================================================================*\
|*		Editor mined						*|
|*		keyboard input						*|
|*		currently included from io.c				*|
\*======================================================================*/


voidfunc keyproc = I;	/* function addressed by entered function key */
char keyshift = '0';	/* shift state of entered function key */
#ifdef deprecated_accent_prefix
charfunc accentproc;
#endif

/* _readchar may precompose an accented character if an 'accentproc' 
   function key is read as a prefix. The resulting character is 
   delivered as 1 Latin-1-encoded byte, regardless of effective 
   terminal encoding.
 */
FLAG last_char_latin1 = False;	/* was accented character composed? */

int keymap_delay = 900;		/* wait to detect mapped key sequence */
int dont_suppress_keymap = 1;


/*======================================================================*\
|*		Handling function key and keyboard mapping tables	*|
\*======================================================================*/

#ifndef msdos

extern struct fkeyentry {
	char * fk;
	voidfunc fp;
} fkeymap [];

extern struct fkeyentry fkeymap_xterm [];
extern struct fkeyentry fkeymap_hp [];
extern struct fkeyentry fkeymap_siemens [];

struct fkeyentry * fkeymap2 = fkeymap_xterm;

void
set_fkeymap (term)
  char * term;
{
  if (term == NIL_PTR) {
	return;
  } else if (* term == 'x') {
	fkeymap2 = fkeymap_xterm;
  } else if (* term == 'h') {
	fkeymap2 = fkeymap_hp;
  } else if (* term == 's') {
	fkeymap2 = fkeymap_siemens;
  }
}


struct keymap {
	char * fk;
	char * fp;
};

#include "keytrans.h"

char * keyboard_mapping = "--";
static char * last_keyboard_mapping = "--";
static char * next_keyboard_mapping = "";

static struct keymap * keyboard_map = (struct keymap *) NIL_PTR;
static struct keymap * last_keyboard_map = (struct keymap *) NIL_PTR;

/**
   Set keymap.
 */
static
void
set_keymap (new_keymap, current, next)
  struct keymap * new_keymap;
  char * current;
  char * next;
{
  last_keyboard_map = keyboard_map;
  last_keyboard_mapping = keyboard_mapping;
  keyboard_map = new_keymap;
  keyboard_mapping = current;
  next_keyboard_mapping = next;
}

/**
   Exchange current and previous keyboard mapping; set previous active.
   With HOP, reset keyboard mapping to none.
 */
void
toggleKEYMAP ()
{
  if (allow_keymap) {
	if (hop_flag > 0) {
		set_keymap ((struct keymap *) NIL_PTR, "--", "gr");
	} else {
		set_keymap (last_keyboard_map, last_keyboard_mapping, last_keyboard_mapping);
	}
	displayflags ();
  } else {
	error ("Keyboard mapping not active", NIL_PTR);
  }
}

/**
   Set keyboard mapping.
 */
static
FLAG
selectKEYMAP (script)
  char * script;
{
	if (script == NIL_PTR || * script == '\0') {
		set_keymap ((struct keymap *) NIL_PTR, "--", "gr");
	} else if (strisprefix ("gr", script)) {
		set_keymap (keymap_GreekMonotonic, "gr", "cy");
	} else if (strisprefix ("el", script)) {
		set_keymap (keymap_GreekMonotonic, "gr", "cy");
	} else if (strisprefix ("cy", script)) {
		set_keymap (keymap_Cyrillic, "cy", "ar");
	} else if (strisprefix ("ru", script)) {
		set_keymap (keymap_Cyrillic, "cy", "ar");
	} else if (strisprefix ("ar", script)) {
		set_keymap (keymap_arabic, "ar", "he");
	} else if (strisprefix ("he", script)) {
		set_keymap (keymap_Hebrew, "he", "");
	} else {
		return False;
	}
	displayflags ();
	return True;
}

/**
   Set keyboard mapping.
 */
void
setKEYMAP (script)
  char * script;
{
  if (script != NIL_PTR && * script == '-') {
	/* set keymap as secondary map */
	script ++;
	selectKEYMAP (script);
	selectKEYMAP (NIL_PTR);
  } else {
	selectKEYMAP (script);
  }
}

/**
   Cycle through keyboard mappings.
 */
void
cycleKEYMAP ()
{
  selectKEYMAP (next_keyboard_mapping);
}

/**
   Enable keyboard mapping interactively.
   With HOP, cycle through keyboard mappings.
 */
void
setupKEYMAP ()
{
  if (allow_keymap) {
	if (hop_flag > 0) {
		hop_flag = 0;
		selectKEYMAP (next_keyboard_mapping);
	} else {
		handleKeymapmenu (0);
	}
  } else {
	error ("Keyboard mapping not active", NIL_PTR);
  }
}

/**
   get_string_nokeymap prompts a string like get_string
   but doesn't apply international keyboard mapping
 */
int
get_string_nokeymap (prompt, inbuf, statfl, term_input)
  char * prompt;
  char * inbuf;
  FLAG statfl;
  char * term_input;
{
  int ret;

  dont_suppress_keymap = 0;
  ret = get_string (prompt, inbuf, statfl, term_input);
  dont_suppress_keymap = 1;

  return ret;
}

#endif


/*======================================================================*\
|*				Terminal input				*|
\*======================================================================*/

#define max_stamps 10
static long last_sec = 0;
static long last_msec = 0;
static int n_stamps = 0;
static int i_stamp = 0;
static long stamps [max_stamps];
static long total_deltas = 0;
long average_delta_readchar = 0;
long last_delta_readchar = 0;

static
void
timestamp_readchar ()
{
  struct timeval now;
  long sec;
  long msec;

  gettimeofday (& now, 0);
  sec = now.tv_sec;
  msec = now.tv_usec / 1000;
  if (last_sec == 0) {
	last_delta_readchar = 1000;
  } else {
	last_delta_readchar = (sec - last_sec) * 1000 + msec - last_msec;
  }
  last_sec = sec;
  last_msec = msec;

  if (n_stamps < max_stamps) {
	n_stamps ++;
  } else {
	/* remove oldest stamp from statistics */
	total_deltas -= stamps [i_stamp];
  }
  stamps [i_stamp] = last_delta_readchar;
  /* add stamp to statistics */
  total_deltas += last_delta_readchar;
  i_stamp ++;
  if (i_stamp == max_stamps) {
	i_stamp = 0;
  }

  average_delta_readchar = total_deltas / n_stamps;
}

/*
   readchar_allbuttons reads like readchar
   but always reports mouse button release
 */
int
readchar_allbuttons ()
{
  report_release = True;
  return readchar ();
}

/*
   readchar_nokeymap reads like readchar
   but doesn't apply international keyboard mapping
 */
int
readchar_nokeymap ()
{
  int res;
  int prev_dont_suppress_keymap = dont_suppress_keymap;

  dont_suppress_keymap = 0;
  res = readchar ();
  dont_suppress_keymap = prev_dont_suppress_keymap;

  return res;
}

/*
   readcharacter () reads one Unicode character
 */
unsigned long
readcharacter ()
{
  int utfcount = 1;
  unsigned long unichar;

  unichar = readchar ();

  if ((unichar & 0x80) == 0x00) {
	return unichar;
  } else if (Chinese == True && multichar (unichar)) {
	return unichar << 8 | readchar ();
  } else if (utf8_input == False) {
	return unichar;
  } else {
	if ((unichar & 0xE0) == 0xC0) {
		utfcount = 2;
		unichar = unichar & 0x1F;
	} else if ((unichar & 0xF0) == 0xE0) {
		utfcount = 3;
		unichar = unichar & 0x0F;
	} else if ((unichar & 0xF8) == 0xF0) {
		utfcount = 4;
		unichar = unichar & 0x07;
	} else if ((unichar & 0xFC) == 0xF8) {
		utfcount = 5;
		unichar = unichar & 0x03;
	} else if ((unichar & 0xFE) == 0xFC) {
		utfcount = 6;
		unichar = unichar & 0x01;
	} else /* illegal UTF-8 code */ {
		return unichar;
	}
	while (utfcount > 1) {
		unichar = (unichar << 6) | readchar ();
		utfcount --;
	}
  }
}

/*
 * Readchar () reads one character from the terminal.
 * There are problems due to interruption of the read operation by signals
 * (QUIT, WINCH). The waitingforinput flag is only a partial solution.
 * Unix doesn't provide sufficient facilities to handle these situations
 * neatly and properly. Moreover, different Unix versions yield different
 * surprising effects. However, the use of select () could still be
 * an improvement.
 */
int
readchar ()
{
  register character c;

#ifdef msdos
#define winchhere
#endif

#ifdef winchhere
  FLAG waiting;

  if (winchg == True && waitingforinput == False) {
	/* In the Unix version, this is now done in __readchar () */
	RDwin ();
  }
  /* must save waitingforinput flag since in the MSDOS version, 
     readchar can be called recursively */
  waiting = waitingforinput;
#endif
  waitingforinput = True;

  c = _readchar ();
  while (c == (character) FUNcmd && (* keyproc) == MOUSEescape 
	&& mouse_button == releasebutton && report_release == False)
  {
	c = _readchar ();
  }
  report_release = False;

#ifdef winchhere
  waitingforinput = waiting;
#else
  waitingforinput = False;
#endif

  /* the modification   if (quit == True) c = QUITCHAR;   (now in __readchar)
     must not be placed after resetting the flag  waitingforinput = False;  .
     Otherwise a QUIT signal coming in just between these two would
     discard the last valid character just taken up. */

  timestamp_readchar ();

  return c;
}

/*-------------------------------------------------------------------------*/

/*
 * DIRECT...getxy () reads in the information part of a cursor escape sequence
 */
static
void
DIRECTxtermgetxy (code)
  char code;	/* 'M' for normal mouse report, 't'/'T' for tracking */
{
  character button;
  int xpos;
  int ypos;
  char track [22];

  mouse_lastbutton = mouse_button;
  if (mouse_button == leftbutton || mouse_button == rightbutton || mouse_button == middlebutton) {
	mouse_prevbutton = mouse_button;
  }
  mouse_lastxpos = mouse_xpos;
  mouse_lastypos = mouse_ypos;

  if (code == 't') {
	mouse_button = releasebutton;
  } else if (code == 'T') {
	/* could be Siemens 97801/97808 page down ... !?! */
	button = _readchar_nokeymap ();
	button = _readchar_nokeymap ();
	button = _readchar_nokeymap ();
	button = _readchar_nokeymap ();
	mouse_button = releasebutton;
  } else {	/* code == 'M' */
	button = _readchar_nokeymap ();
	if (button == QUITCHAR) {
		quit = True;
	}
	button -= 32;
	mouse_shift = button & 0x1C;
	if (button & 32) {
		mouse_button = movebutton;
	} else {
		button &= 0x03;
		if (button == 0x00) {
			mouse_button = leftbutton;
		} else if (button == 0x01) {
			mouse_button = middlebutton;
		} else if (button == 0x02) {
			mouse_button = rightbutton;
		} else {
			mouse_button = releasebutton;
		}
	}
  }

  xpos = _readchar_nokeymap ();
  if (xpos == QUITCHAR) {
	quit = True;
  }
  mouse_xpos = xpos - 33;
  ypos = _readchar_nokeymap ();
  if (ypos == QUITCHAR) {
	quit = True;
  }
  mouse_ypos = ypos - 33 - MENU;

#ifdef mouse_hilite_tracking
  if (mouse_button == leftbutton) {
	/* handle hilite tracking */
	if (mouse_ypos < 0 || mouse_ypos > last_y 
	    || (mouse_xpos == XMAX && scrollbar == True) 
	    || is_menu_open ())
	{
		/* abort hilite tracking */
		putstring ("\033[0;0;0;0;0T");
		flush ();
	} else {
#define track_trick
#ifdef track_trick
		move_to (mouse_xpos, mouse_ypos);
		build_string (track, "\033[1;%d;%d;%d;%dT", 
				x + 1, y + 1 + MENU, 
				1 + MENU, last_y + 3);
#else
		build_string (track, "\033[1;%d;%d;%d;%dT", 
				mouse_xpos + 1, mouse_ypos + 1 + MENU, 
				1 + MENU, last_y + 3);
#endif
		putstring (track);
		flush ();
	}
  }
#endif
}

static
void
DIRECTcrttoolgetxy ()
{
  character c;
  int xpos;
  int ypos;

  mouse_lastbutton = mouse_button;
  if (mouse_button == leftbutton || mouse_button == rightbutton || mouse_button == middlebutton) {
	mouse_prevbutton = mouse_button;
  }
  mouse_lastxpos = mouse_xpos;
  mouse_lastypos = mouse_ypos;

  c = get_digits (& ypos); /* c should be ';' */
  c = get_digits (& xpos);
  if (c == 'l') {
	mouse_button = leftbutton;
  } else if (c == 'm') {
	mouse_button = middlebutton;
  } else if (c == 'r') {
	mouse_button = rightbutton;
  } else {
	mouse_button = releasebutton;
  }

  mouse_ypos = ypos - 1 - MENU;
  mouse_xpos = xpos - 1;
}

/*
 * DIRECT... () reads in a direct cursor movement input sequence.
 * Then it either performs a direct function (move/copy/paste) or 
 * invokes the menu functions.
 * Is no longer called as this is already handled in _readchar ().
 */
void
DIRECTxterm ()
{
  DIRECTxtermgetxy ('M');
  MOUSEescape ();
}
void
TRACKxterm ()
{
  DIRECTxtermgetxy ('t');
  MOUSEescape ();
}
void
TRACKxtermT ()
{
  DIRECTxtermgetxy ('T');
  MOUSEescape ();
}

void
DIRECTcrttool ()
{
  DIRECTcrttoolgetxy ();
  MOUSEescape ();
}


#ifdef CURSES
#ifdef KEY_MOUSE

#ifdef __PDCURSES__

/*
 * CURSgetxy () reads in the information of a mouse event
 */
static
void
CURSgetxy ()
{
# define mouse_state Mouse_status.changes
# define mouse_x MOUSE_X_POS
# define mouse_y MOUSE_Y_POS

  int button = 0;

  mouse_lastbutton = mouse_button;
  if (mouse_button == leftbutton || mouse_button == rightbutton || mouse_button == middlebutton) {
	mouse_prevbutton = mouse_button;
  }
  mouse_lastxpos = mouse_xpos;
  mouse_lastypos = mouse_ypos;

  request_mouse_pos ();

  if (BUTTON_CHANGED (1)) {
	button = 1;
  } else if (BUTTON_CHANGED (2)) {
	button = 2;
  } else if (BUTTON_CHANGED (3)) {
	button = 3;
  }

  mouse_shift = (BUTTON_STATUS (button) & BUTTON_MODIFIER_MASK) == BUTTON_SHIFT;

  if ((BUTTON_STATUS (button) & BUTTON_ACTION_MASK) == BUTTON_PRESSED) {
	if (button == 1) {
		mouse_button = leftbutton;
	} else if (button == 2) {
		mouse_button = middlebutton;
	} else if (button == 3) {
		mouse_button = rightbutton;
	} else {
		mouse_button = releasebutton;
	}
  } else {
	mouse_button = releasebutton;
  }

  mouse_xpos = mouse_x;
  mouse_ypos = mouse_y - MENU;
}

# else

static
void
CURSgetxy ()
{
  MEVENT mouse_info;
# define mouse_state mouse_info.bstate
# define mouse_x mouse_info.x
# define mouse_y mouse_info.y

  mouse_lastbutton = mouse_button;
  if (mouse_button == leftbutton || mouse_button == rightbutton || mouse_button == middlebutton) {
	mouse_prevbutton = mouse_button;
  }
  mouse_lastxpos = mouse_xpos;
  mouse_lastypos = mouse_ypos;

  getmouse (& mouse_info);

  mouse_shift = mouse_state & BUTTON_SHIFT;
  if (mouse_state & BUTTON1_CLICKED) {
	mouse_button = leftbutton;
  } else if (mouse_state & BUTTON2_CLICKED) {
	mouse_button = middlebutton;
  } else if (mouse_state & BUTTON3_CLICKED) {
	mouse_button = rightbutton;
  } else {
	mouse_button = releasebutton;
  }

  mouse_xpos = mouse_x - 1;
  mouse_ypos = mouse_y - 1 - MENU;
}

#endif

#endif
#endif


/*-------------------------------------------------------------------------*/

#ifndef msdos

#define MAXCODELEN 7 /* max. length of function key sequence to be detected,
			depending on the fkeymap table */

/*
 * queue collects the keys of an Escape sequence typed in until the 
 * sequence can be detected or rejected.
 * If the queue is not empty, queue [0] contains the character next 
 * to be delivered by _readchar () (it's not a ring buffer).
 * The queue contents is always terminated by a '\0', so queue can also 
 * be taken as a character string.
 */
static character queue [MAXCODELEN + 1], * endp = queue;
static character ctrl_queue [MAXCODELEN + 1], * endcp = ctrl_queue;

#ifdef unused
static
int
q_empty ()
{
  return endp == queue ? 1 : 0;
}
#endif

static
int
q_notfull ()
{
  return endp - queue == MAXCODELEN ? 0 : 1;
}

static
void
q_clear ()
{
  endp = queue;
  endcp = ctrl_queue;
}

static
int
q_len ()
{
  return endp - queue;
}

static
void
q_put (c)
  character c;
/* queue must not be full prior to this call! */
{
  * endp = c;
  * ++ endp = '\0';
  if (c == '\0') {
	* endcp = '@';
  } else {
	* endcp = c;
  }
  * ++ endcp = '\0';
}

static
character
q_unput ()
{
  character c;
  if (q_len () == 0) {
	return '\0';
  } else {
	endp --;
	endcp --;
	c = * endp;
	* endp = '\0';
	* endcp = '\0';
	return c;
  }
}

static
character
q_get ()
{
  character c;
  register character * pd;
  register character * ps;

  c = * queue;
  pd = queue;
  ps = pd + 1;
  while (ps <= endp) {
	* pd ++ = * ps ++;
  }
  if (endp > queue) {
	endp --;
	endcp --;
  }
  return c;
}

/*
 * Look up key sequence in fkeymap table.
 * if matchmode == 0:
 * findkey (str) >=  0: escape sequence found
 *				(str == fkeymap [findkey (str)].fk)
 *			keyproc = fkeymap [i].fp (side effect)
 *		 == -1: escape sequence prefix recognised
 *				(str is prefix of some entry in fkeymap)
 *		 == -2: escape sequence not found
 *				(str is not contained in fkeymap)
 * * found == index of exact match if found
 * if matchmode == 1:
 * return -1 if any prefix detected (even if other exact match found)
 * if matchmode == 2:
 * don't return -1, only report exact or no match
 */
static
int
findkeyin (str, fkeymap, matchmode, found)
  char * str;
  struct fkeyentry * fkeymap;
  int matchmode;
  int * found;
{
  int lastmatch = 0;	/* last index with string matching prefix */
  register int i;
  int ret = -2;

  * found = -1;

  if (fkeymap [0].fk == NIL_PTR) {
	return ret;
  }

  i = lastmatch;
  do {
	if (strncmp (str, fkeymap [i].fk, strlen (str)) == 0) {
		/* str is prefix of current entry */
		lastmatch = i;
		if (strlen (str) == strlen (fkeymap [i].fk)) {
			/* str is equal to current entry */
			* found = i;
			keyproc = fkeymap [i].fp;
			if (matchmode == 1) {
				/* return index unless other prefix 
				   was or will be found */
				if (ret == -2) {
					ret = i;
				}
			} else {
				return i;
			}
		} else {
			/* str is partial prefix of current entry */
			if (matchmode == 1) {
				/* return -1 but continue to search 
				   for exact match */
				ret = -1;
			} else if (matchmode != 2) {
				return -1;
			} else {
			}
		}
	}
	++ i;
	if (fkeymap [i].fk == NIL_PTR) {
		i = 0;
		return ret;
	}
  } while (i != lastmatch);

  return ret;
}

static
int
findkey (str)
  char * str;
{
  int dummy;
  int res = findkeyin (str, fkeymap, 0, & dummy);
  if (res == -2) {
	return findkeyin (str, fkeymap2, 0, & dummy);
  } else {
	return res;
  }
}

#endif	/* ifndef msdos */

#ifndef msdos
/*
 * Is a character available within a specified number of milliseconds ?
 */
int
char_ready_within (msec)
  int msec;
{
  return (q_len () > 0) || inputreadyafter (input_fd, msec);
}

#else	/* ifndef msdos */

int
char_ready_within (msec)
  int msec;
{
  return inputreadyafter (input_fd, msec);
}

#endif


#ifdef CURSES
chtype key;

static
void
showfkey ()
{
  build_string (text_buffer, "Curses function key entered: %03X / %04o", key, key);
  status_msg (text_buffer);
}


/*
 * Read a character, ignoring uninterpreted key strokes (SHIFT etc.)
 * also handles resize events
 */
static
int
curs_readchar ()
{
  int keycode;
  keycode = __readchar ();
  while (keycode >= KEY_MIN) {
	switch (keycode) {
#ifdef KEY_RESIZE
	case KEY_RESIZE:
printf ("curses resize key received \n"); sleep (1);
			keyproc = showfkey; return FUNcmd;
#endif
#ifdef KEY_SHIFT_L
	case KEY_SHIFT_L:
	case KEY_SHIFT_R:
	case KEY_CONTROL_L:
	case KEY_CONTROL_R:
	case KEY_ALT_L:
	case KEY_ALT_R:
			break;
#endif
	default:	return keycode;
	}
	keycode = __readchar ();
  }

  return keycode;
}

/*
 * Mapping KEY_ code (curses) -> mined function
 */
struct {
	chtype keycode;
	voidfunc func;
} cursfunc [] =
{
#ifdef PAD0
	{PAD0, PT},
#endif
	{KEY_F(1), HELP},
	{KEY_F(2), WT},
	{KEY_F(3), EDIT},
	{KEY_F(4), INSFILE},
	{KEY_F(5), insert_diaeresis},
	{KEY_F(6), insert_diaeresis},
	{KEY_F(7), SFW},
	{KEY_F(8), SFW},
	{KEY_F(9), RS},
	{KEY_F(10), FILEMENU},
	{KEY_F(20), QUICKMENU},
	{KEY_F(11), LOWCAP},
	{KEY_F(12), UML},
#ifdef __PDCURSES__
	{PADSLASH, I},
	{PADSTAR, I},
	{PADMINUS, I},
	{PADPLUS, I},
	{PADSTOP, DT},
	{PADENTER, SNL},
	{ALT_F, FILEMENU},
	{ALT_E, EDITMENU},
	{ALT_S, SEARCHMENU},
	{ALT_X, EXTRAMENU},
	{ALT_P, PARAMENU},
#endif
	{KEY_IC, PT},
	{KEY_DC, DCC},
	{KEY_HOME, MARK},
	{KEY_END, YA},
	{0x1BF, BLINE},
	{0x1C0, ELINE},
	{KEY_PPAGE, PU},
	{KEY_NPAGE, PD},

	{0}
};

/*
 * Look up mined function by curses key code
 * lookup_curskey (str) >=  0: keycode == cursfunc [lookup_curskey (str)].fk
 *			== -1: keycode is not contained in mapping table
 */
static
voidfunc
lookup_curskey (keycode)
  chtype keycode;
{
  register int i;

  i = 0;
  while (cursfunc [i].keycode != 0 && cursfunc [i].keycode != keycode) {
	i ++;
  }
  if (cursfunc [i].keycode == 0) {
	return showfkey;
  } else {
	return cursfunc [i].func;
  }
}

#endif


/*
 * Read a character from terminal, considering function keys and 
 * composing special character of an 8 bit character set.
 * _readchar () takes the following actions:
 *  -	function key sequences according to the table fkeymap are 
 *	transformed into a special controlling character which is 
 *	assigned the function FUNKEY. Also the intended editor function, 
 *	as taken from the table fkeymap, is saved in a variable for 
 *	use by FUNKEY.
 *  -	the prefix keys for diacritic and special characters are 
 *	combined with the following key to make up the character.
 */
#ifdef CURSES
#define read1char curs_readchar
#else
#define read1char __readchar
#endif

static FLAG is_char_stored = False;
static character ch_stored;

int
_readchar_nokeymap ()
{
  int res;
  int prev_dont_suppress_keymap = dont_suppress_keymap;

  dont_suppress_keymap = 0;
  res = _readchar ();
  dont_suppress_keymap = prev_dont_suppress_keymap;

  return res;
}

int
_readchar ()
{
  register character ch;
#ifndef msdos
  char second_esc_char;
#endif
#ifdef dosmouse
  int ich;
#endif
#ifndef msdos
  int res;
  int found;
  int prev_res = -2;
#endif
  char * transformed;


#ifndef msdos
  if (q_len () > 0) {
	return q_get ();
  }
#endif

  keyshift = '0';

#ifndef msdos
  if (is_char_stored == True) {
	ch = ch_stored;
	is_char_stored = False;
  } else {
#endif

#ifdef CURSES

  key = curs_readchar ();
  while (key >= KEY_MIN) {
	switch (key) {
#ifdef KEY_A2
	case KEY_A2:
#endif
	case KEY_UP:	keyproc = MUP; return FUNcmd;
#ifdef KEY_C2
	case KEY_C2:
#endif
	case KEY_DOWN:	keyproc = MDN; return FUNcmd;
#ifdef KEY_B1
	case KEY_B1:
#endif
	case KEY_LEFT:	keyproc = MLF; return FUNcmd;
#ifdef KEY_B3
	case KEY_B3:
#endif
	case KEY_RIGHT:	keyproc = MRT; return FUNcmd;
	case KEY_B2:	keyproc = HOP; return FUNcmd;
	case KEY_A1:	keyproc = MARK; return FUNcmd;
	case KEY_A3:	keyproc = PU; return FUNcmd;
	case KEY_C1:	keyproc = YA; return FUNcmd;
	case KEY_C3:	keyproc = PD; return FUNcmd;
#ifdef KEY_MOUSE
	case KEY_MOUSE:
			CURSgetxy ();
			keyproc = MOUSEescape;
			return FUNcmd;
#endif

# ifdef curs_readchar_not_defined
#ifdef KEY_RESIZE
	case KEY_RESIZE:
			keyproc = showfkey; return FUNcmd;
#endif
#ifdef KEY_SHIFT_L
	case KEY_SHIFT_L:
	case KEY_SHIFT_R:
	case KEY_CONTROL_L:
	case KEY_CONTROL_R:
	case KEY_ALT_L:
	case KEY_ALT_R:
			break;
#endif
# endif

	default:
		keyproc = lookup_curskey (key); return FUNcmd;
	}
	/* loop should not occur, handled in curs_readchar */
	key = curs_readchar ();
  }
  ch = key;

#else	/* ifdef CURSES */

#ifdef dosmouse
  ich = __readchar ();
  last_char_latin1 = False;
  if (ich < 0) {
	DIRECTdosmousegetxy ();
	keyproc = MOUSEescape;
	return FUNcmd /* index of FUNKEY */;
  } else {
	ch = ich;
  }
#else
  ch = __readchar ();
#endif

#endif	/* else CURSES */

#ifndef msdos
  }	/* else is_char_stored */
#endif


  last_char_latin1 = False;

  if (ch == '\000') {
#ifdef dosmouse
	ch = getch ();
#else
	ch = __readchar ();
#endif
	keyproc = pc_key_map [ch];
#ifdef DEBUG
	if ((voidfunc) keyproc == (voidfunc) I) {
		return ch;
	}
	/* (voidfunc) is an identity cast here. It seems to be required 
	   for the sake of the apparently totally rotten microvax compiler */
#endif
#ifdef deprecated_accent_prefix
	accentproc = (charfunc) keyproc;
	if ((accentproc == grave) || (accentproc == circumflex)
		 || (accentproc == acute) || (accentproc == diaeresis)
		 || (accentproc == tilde) || (accentproc == angstrom))
	{
		ch = (* accentproc) (readchar ());
		 keyproc = I;
		 last_char_latin1 = True;
		 return ch;
	} else
#endif
	{
		return FUNcmd /* index of FUNKEY */;
	}
  }

#ifndef msdos

    else if (ch == '\033') {
	/* q_clear (); */
	q_put (ch);
	second_esc_char = ch;
	while  ((res = findkey (ctrl_queue)) == -1 /* prefix of table entry */
		&& q_notfull ()
		&& inputreadyafter (input_fd, escape_delay)
	       )
	{
		ch = read1char ();

		/* handle generic shift state of ANSI sequences: */
		if (ch == ';' && second_esc_char == '[') {
			keyshift = read1char () - 1;
			ch = read1char ();
		} else if (ch >= '0' && ch <= '9' && second_esc_char == 'O') {
			keyshift = ch - 1;
			ch = read1char ();
		}

		q_put (ch);
		if (second_esc_char == '\033') {
			second_esc_char = ch;
		}
	}

	if (quit == True) {
		return '\0';
	} else if (res < 0) { /* key pattern not detected in fkeymap table */
		return q_get () /* just deliver the typed characters */;
	} else {
		q_clear ();

		if (keyproc == DIRECTxterm) {
			DIRECTxtermgetxy ('M');
			keyproc = MOUSEescape;
		} else if (keyproc == TRACKxterm) {
			DIRECTxtermgetxy ('t');
			keyproc = MOUSEescape;
		} else if (keyproc == TRACKxtermT) {
			DIRECTxtermgetxy ('T');
			keyproc = MOUSEescape;
		} else if (keyproc == DIRECTcrttool) {
			DIRECTcrttoolgetxy ();
			keyproc = MOUSEescape;
		}

#ifdef deprecated_accent_prefix
		accentproc = (charfunc) keyproc;
		if ((accentproc == grave) || (accentproc == circumflex)
		    || (accentproc == acute) || (accentproc == diaeresis)
		    || (accentproc == tilde) || (accentproc == angstrom)
		   )
		{
			ch = (* accentproc) (readchar ());
			keyproc = I;
			last_char_latin1 = True;
			return ch;
		} else {
			return FUNcmd /* index of FUNKEY */;
		}
#else
		return FUNcmd /* index of FUNKEY */;
#endif
	}
  }
    else if (allow_keymap && dont_suppress_keymap && keyboard_map != (struct keymap *) NIL_PTR) {
	q_put (ch);
	while  ((res = findkeyin (ctrl_queue, keyboard_map, 1, & found)) == -1
		 /* prefix of some table entry */
		&& q_notfull ()
		&& inputreadyafter (input_fd, keymap_delay)
	       )
	{
		q_put (read1char ());
		/* exact match found? note for later consideration */
		if (found >= 0) {
			prev_res = found;
		}
	}
	/* possible results:
		res >= 0: match found, no further prefix match
		res == -1, prev_res >= 0: prefix with previous match
		res == -1: prefix only, no further key typed
		res == -2: no match with previous match
	*/

	if (res < 0 && prev_res >= 0) {
		res = prev_res;
		ch_stored = q_unput ();
		is_char_stored = True;
	} else if (res == -1) {
		res = findkeyin (ctrl_queue, keyboard_map, 2, & found);
	}

	if (res >= 0) { /* key pattern detected in fkeymap table */
		q_clear ();
		transformed = (char *) keyproc;
		while (* transformed != '\0') {
			q_put (* transformed);
			transformed ++;
		}
	}

	return q_get () /* deliver the first transformed byte */;
  }

#endif

    else if (transdeutsch == True) {
	switch (ch) {
		case '{': return '\344';
		case '|': return '\366';
		case '}': return '\374';
		case '~': return '\337';
		case '[': return '\304';
		case '\\': return '\326';
		case ']': return '\334';
		case (unsigned char) '\344': return '{';
		case (unsigned char) '\366': return '|';
		case (unsigned char) '\374': return '}';
		case (unsigned char) '\337': return '~';
		case (unsigned char) '\304': return '[';
		case (unsigned char) '\326': return '\\';
		case (unsigned char) '\334': return ']';
		default: return ch;
	}
  } else {
	return ch;
  }
}


/*======================================================================*\
|*				End					*|
\*======================================================================*/
