/*======================================================================*\
|*		Editor mined						*|
|*		keyboard mappings to commands and characters		*|
\*======================================================================*/

#include "mined.h"
#include "io.h"


/*======================================================================*\
|*			8 bit character constant stuff			*|
\*======================================================================*/

#ifdef pc_charset
char TABdefault = (char) 250;
char RETdefault = (char) 0xAE;
char DOSRETdefault = (char) 0xE6;
char PARAdefault = (char) 0x9E;	/* Char displayed to indicate end of paragraph */
char SHIFT_MARK = (char) 175;	/* Char indicating that line continues */
char UNI_MARK = (char) 207;	/* char to be shown in place of Unicode char */
character ring = (character) 248;
#else
char TABdefault = '';
char RETdefault = '';
char DOSRETdefault = '';
char PARAdefault = '';		/* Char displayed to indicate end of paragraph */
char SHIFT_MARK = '';		/* Char indicating that line continues */
char UNI_MARK = '';		/* char to be shown in place of Unicode char */
character ring = (character) '';
#endif

int
idfchar (ch)
  character ch;
{
  return ('0' <= ch && ch <= '9') ||
	 ('A' <= ch && ch <= 'Z') ||
	 ('a' <= ch && ch <= 'z') ||
#ifdef pc_charset
	 ((character) '\200' <= ch && ch <= (character) '\232') ||
	 ((character) '\240' <= ch && ch <= (character) '\245') ||
	 ((character) '\340' <= ch && ch <= (character) '\353') ||
#else
	 ((character) '\300' <= ch && ch != (character) '' && ch != (character) '') ||
	 (ch == (character) '') ||
#endif
	 (ch == '_');
}


/*======================================================================*\
|*			MSDOS Video mode switching table		*|
\*======================================================================*/

struct {
	int mode, cols, lins;
} modetab [] =
{ /* list of available modes, unsorted, terminated by a '-1' mode */
  /* the entries contain:
	a video mode number (restriction to text modes preferable),
	the numbers of columns and of lines which can be displayed 
		in that mode if a font of height 16 is used */
  {42, 100, 40},
  {38, 80, 60},
  {36, 132, 28},
  {35, 132, 25},
  {34, 132, 44},
  { 3, 80, 25},
  { 1, 40, 25},
  {-1, 0, 0}
};


/*======================================================================*\
|*			Character set coding				*|
\*======================================================================*/

/*
 * Determine a default diacritic character value for _readchar ().
 * The characters are given literally except for the MSDOS version in order
 * not to depend on the function of a character set translator.
 * For all non-MSDOS platforms, ISO-Latin1 or the very similar 
 * DEC-Supplemental are assumed.
 * For MSDOS, characters for the "International Codepage" ("850") are 
 * generated which includes the real letters (not all graphic and Greek 
 * characters) of the old IBM codepage ("437") (which, by the way, some 
 * time ago someone must have designed in an attack of mental sickness).
 */
static
unsigned long
diacritic (dia, c)
  character dia;
  character c;
{
  unsigned int composed = compose (dia, c);
  if (composed != '') {
	return composed;
  }

  /*	the following characters cannot be inserted by this function:
	 (ISO-Latin-1) and MSDOS graphics characters */
  switch (c) {
#ifdef pc_charset
	case 'C' : return (character) 128;
	case 'c' : return (character) 135;
	case 'N' : return (character) 165;
	case 'n' : return (character) 164;
	case 's' : return (character) 225;
	case '!' : return (character) 173;
	case '$' : return (character) 189;
	case '/' : return (character) 189;
	case 'l' : return (character) 156;
	case 'L' : return (character) 156;
	case 'Y' : return (character) 190;
	case '+' : return (character) 241;
	case '?' : return (character) 168;
	case '@' : return (character) 248;
	case '1' : return (character) 251;
	case '2' : return (character) 253;
	case '3' : return (character) 252;
	case '<' : return (character) 174;
	case '>' : return (character) 175;
	case ':' : return (character) 246;
	case 'x' : return (character) 158;
	case 'f' : return (character) 159;
	case 'D' : return (character) 209;
	case 'd' : return (character) 208;
	case 'P' : return (character) 232;
	case 'p' : return (character) 231;
	case 'i' : return (character) 213;
	case '|' : return (character) 221;
	case '&' : return (character) 244;
	case '\\' : return (character) 170;
	case '_' : return (character) 238;
	case '-' : return (character) 196;
	case 'm' : return (character) 230;
	case '=' : return (character) 242;
	case '#' : return (character) 245;
	case '.' : return (character) 250;
	case ',' : return (character) 247;
	case '\'' : return (character) 239;
	case 'X' : return (character) 207;
#else
	case 'C' : return (character) '';
	case 'c' : return (character) '';
	case 'N' : return (character) '';
	case 'n' : return (character) '';
	case 's' : return (character) '';
	case '!' : return (character) '';
	case '$' : return (character) '';
	case '/' : return (character) '';
	case 'l' : return (character) '';
	case 'L' : return (character) '';
	case '+' : return (character) '';
	case '?' : return (character) '';
	case '@' : return (character) '';
	case '2' : return (character) '';
	case '<' : return (character) '';
	case '>' : return (character) '';
	case 'D' : return (character) '';
	case 'd' : return (character) '';
	case 'P' : return (character) '';
	case 'p' : return (character) '';
	case '|' : return (character) '';
	case '&' : return (character) '';
	case '\\' : return (character) '';
	case '_' : return (character) '';
	case '-' : return (character) '';
	case 'm' : return (character) '';
	case '#' : return (character) '';
	case '.' : return (character) '';
	case '*' : return (character) '';
	case ',' : return (character) '';
	case '\'' : return (character) '\'';
	case '1' : return (character) '';
	case '3' : return (character) '';
	case 'X' : return (character) '';
	case '"' : return (character) '"';
	case 'x' : return (character) '';
	case ':' : return (character) '';
#endif
	default: return c;
  }
}

unsigned long
grave (ch)
  character ch;
{
	switch (ch) {
#ifdef pc_charset
		case 'A' : return (character) 183;
		case 'E' : return (character) 212;
		case 'I' : return (character) 222;
		case 'O' : return (character) 227;
		case 'U' : return (character) 235;
		case 'a' : return (character) 133;
		case 'e' : return (character) 138;
		case 'i' : return (character) 141;
		case 'o' : return (character) 149;
		case 'u' : return (character) 151;
#else
		case 'A' : return (character) '';
		case 'E' : return (character) '';
		case 'I' : return (character) '';
		case 'O' : return (character) '';
		case 'U' : return (character) '';
		case 'a' : return (character) '';
		case 'e' : return (character) '';
		case 'i' : return (character) '';
		case 'o' : return (character) '';
		case 'u' : return (character) '';
#endif
		default : return diacritic ('`', ch);
	}
}

unsigned long
circumflex (ch)
  character ch;
{
	switch (ch) {
		case '^' : return (character) '';
#ifdef pc_charset
		case 'A' : return (character) 182;
		case 'E' : return (character) 210;
		case 'I' : return (character) 215;
		case 'O' : return (character) 226;
		case 'U' : return (character) 234;
		case 'a' : return (character) 131;
		case 'e' : return (character) 136;
		case 'i' : return (character) 140;
		case 'o' : return (character) 147;
		case 'u' : return (character) 150;
#else
		case 'A' : return (character) '';
		case 'E' : return (character) '';
		case 'I' : return (character) '';
		case 'O' : return (character) '';
		case 'U' : return (character) '';
		case 'a' : return (character) '';
		case 'e' : return (character) '';
		case 'i' : return (character) '';
		case 'o' : return (character) '';
		case 'u' : return (character) '';
#endif
		default : return diacritic ('^', ch);
	}
}

unsigned long
acute (ch)
  character ch;
{
	switch (ch) {
#ifdef pc_charset
		case 'A' : return (character) 181;
		case 'E' : return (character) 144;
		case 'I' : return (character) 214;
		case 'O' : return (character) 224;
		case 'U' : return (character) 233;
		case 'a' : return (character) 160;
		case 'e' : return (character) 130;
		case 'i' : return (character) 161;
		case 'o' : return (character) 162;
		case 'u' : return (character) 163;
		case 'Y' : return (character) 237;
		case 'y' : return (character) 236;
		case ' ' : return (character) 239;
#else
		case 'A' : return (character) '';
		case 'E' : return (character) '';
		case 'I' : return (character) '';
		case 'O' : return (character) '';
		case 'U' : return (character) '';
		case 'a' : return (character) '';
		case 'e' : return (character) '';
		case 'i' : return (character) '';
		case 'o' : return (character) '';
		case 'u' : return (character) '';
		case 'Y' : return (character) '';
		case 'y' : return (character) '';
#endif
		default : return diacritic ('\'', ch);
	}
}

unsigned long
diaeresis (ch)
  character ch;
{
	switch (ch) {
#ifdef pc_charset
		case 'A' : return (character) 142;
		case 'E' : return (character) 211;
		case 'I' : return (character) 216;
		case 'O' : return (character) 153;
		case 'U' : return (character) 154;
		case 'a' : return (character) 132;
		case 'e' : return (character) 137;
		case 'i' : return (character) 139;
		case 'o' : return (character) 148;
		case 'u' : return (character) 129;
		case 'y' : return (character) 152;
		case ' ' : return (character) 249;
#else
		case 'A' : return (character) '';
		case 'E' : return (character) '';
		case 'I' : return (character) '';
		case 'O' : return (character) '';
		case 'U' : return (character) '';
		case 'a' : return (character) '';
		case 'e' : return (character) '';
		case 'i' : return (character) '';
		case 'o' : return (character) '';
		case 'u' : return (character) '';
		case ' ' : return (character) '';
		case 'y' : return (character) '\377'; /* '' aborts sun compiler */
#endif
		default : return diacritic ('"', ch);
	}
}

unsigned long
tilde (ch)
  character ch;
{
	switch (ch) {
#ifdef pc_charset
		case 'A' : return (character) 199;
		case 'O' : return (character) 229;
		case 'a' : return (character) 198;
		case 'o' : return (character) 228;
#else
		case 'A' : return (character) '';
		case 'O' : return (character) '';
		case 'a' : return (character) '';
		case 'o' : return (character) '';
#endif
		default : return diacritic ('~', ch);
	}
}

unsigned long
angstrom (ch)
  character ch;
{
	switch (ch) {
#ifdef pc_charset
		case 'A' : return (character) 143;
		case 'E' : return (character) 146;
		case 'O' : return (character) 157;
		case 'a' : return (character) 134;
		case 'e' : return (character) 145;
		case 'o' : return (character) 155;
		case 'R' :
		case 'r' : return (character) 169;
		case 'C' :
		case 'c' : return (character) 184;
#else
		case 'A' : return (character) '';
		case 'E' : return (character) '';
		case 'O' : return (character) '';
		case 'a' : return (character) '';
		case 'e' : return (character) '';
		case 'o' : return (character) '';
#endif
		default : return diacritic ('', ch);
	}
}

struct {
	char * mnemo;
	int uni;
} unimnemo3 [] =
{
#include "mnemos.h"
	{NIL_PTR}
};

unsigned int
compose_mnemo (mnemo)
  char * mnemo;
{
  int i = 0;
  char * mnemo2 = & mnemo [1];

  while (unimnemo3 [i].mnemo != NIL_PTR) {
	if (strcmp (unimnemo3 [i].mnemo, mnemo) == 0
	    || (mnemo [0] == '&' && strcmp (unimnemo3 [i].mnemo, mnemo2) == 0)
	   )
	{
		return unimnemo3 [i].uni;
	}
	i ++;
  }

  return '';
}

unsigned int
compose (c1, c2)
  character c1;
  character c2;
{
  character cs [3];
  unsigned int ret;

  cs [0] = c1;
  cs [1] = c2;
  cs [2] = '\0';

  ret = compose_mnemo (cs);
  if (ret == '') {
	cs [0] = c2;
	cs [1] = c1;
	ret = compose_mnemo (cs);
  }
  if (utf8_text == True || ret < 256) {
	return ret;
  } else {
	return '';
  }
}


/*======================================================================*\
|*			Key and function key -> command mapping		*|
\*======================================================================*/

/* The mapping between input codes and functions */

#ifdef unix
#define NULLcmd FUNKEY
#define BSPcmd MLF
#endif
#ifdef msdos
#define NULLcmd FUNKEY	/* since FUNcmd is 0 */
#define BSPcmd DPC
#endif
#ifdef vms
#define NULLcmd FUNKEY
#define BSPcmd MLF
#endif

voidfunc key_map [256] =
{	/* characters to functions map */
   /* 000-017 */ NULLcmd, MPW, DCC, PD, MRT, MUP, MNW, GOTO,
		 /*BSPcmd*/ DPC, S, SNL, DLN, HOP, SNL, RS, LIB,
   /* 020-037 */ PT, HOP, PU, MLF, DNW, DT, CTRl, SU,
		 MDN, YA, SD, ESCAPE, CANCEL, MARK, DPW, BAD,
   /* 040-057 */ S, S, Sdoublequote, S, S, S, S, Ssinglequote,
		 S, S, S, S, S, S, S, S,
   /* 060-077 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 100-117 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 120-137 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 140-157 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 160-177 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, /*DPC*/ DCC,
   /* 200-217 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 220-237 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 240-257 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 260-277 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 300-317 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 320-337 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 340-357 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
   /* 360-377 */ S, S, S, S, S, S, S, S, S, S, S, S, S, S, S, S,
};

voidfunc ws_key_map [32] =
{	/* characters to functions map */
   /* 000-017 */ NULLcmd, MPW, JUS, PD, MRT, MUP, MNW, DCC,
		 MLF, S, S, ctrlK, RS, SNL, LIB, ctrlO,
   /* 020-037 */ CTRl, ctrlQ, PU, MLF, DNW, BAD, TOGINS, SU,
		 MDN, DLINE, SD, ESCAPE, CANCEL, BAD, BAD, BAD,
};

/*
function keys deliver NUL and the following codes:
Block1: RGISOQHKPM
Block2: GHIKMOPQRS	Pos1 Up PgUp Left Center Right End Down PdDn Ins
Cntrl:	st <-/->
Cntrl:	wu Pos1/End
	v PgUp / PgDn	<- for MSDOS: first char is  (a-with-")
F1-F10: ;<=>?@ABCD
shift:	TUVWXYZ[\]
cntrl:	^_`abcdefg
alt:	hijklmnopq
*/

void
HELPF ()
{ status_msg ("       F2 write  F3 edit  F5 \371  F6 \357  F8 search     F9 next       F10 hop"); }
void
HELPS ()
{ status_msg ("shift- F2 write!          F5 ~  F6 `  F8 search rev F9 next       F10 hop"); }
void
HELPC ()
{ status_msg ("cntrl- F2 exit   F3 view  F5 \370  F6 ^  F8 search nxt F9 cycle lns  F10 hop"); }
void
HELPA ()
{ status_msg ("alt-   F2 write buffer                F8 search rev F9 switch lns F10 hop"); }


/* mapping function key codes -> functions */
voidfunc pc_key_map [256] =
{
   /* 00-0F */ NULLcmd, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,
   /* 10 alt-Q */ I,
   /* 11 alt-W */ I,
   /* 12 alt-E */ EDITMENU,
   /* 13 alt-R */ I,
   /* 14 alt-T */ I,
   /* 15 alt-Y */ I,
   /* 16 alt-U */ I,
   /* 17 alt-I */ I,
   /* 18 alt-O */ I,
   /* 19 alt-P */ PARAMENU,
   /* 1A-1D */ I, I, I, I,
   /* 1E alt-A */ I,
   /* 1F alt-S */ SEARCHMENU,
   /* 20 alt-D */ I,
   /* 21 alt-F */ FILEMENU,
   /* 22 alt-G */ I,
   /* 23 alt-H */ HTML,
   /* 24 alt-J */ I,
   /* 25 alt-K */ I,
   /* 26 alt-L */ I,
   /* 27-2B */ I, I, I, I, I,
   /* 2C alt-Z */ I,
   /* 2D alt-X */ EXTRAMENU,
   /* 2E alt-C */ I,
   /* 2F alt-V */ I,
   /* 30 alt-B */ I,
   /* 31 alt-N */ I,
   /* 32 alt-M */ I,
   /* 33-36 */ I, I, I, I,
   /* 37 alt-* */ I,
   /* 38 */ I,
   /* 39 alt-control-blank */ QUICKMENU,
   /* 3A */ I,
   /* 3B F1 */ HELP,	/* HELPF ? */
   /* 3C F2 */ WT,
   /* 3D F3 */ EDIT,
   /* 3E F4 */ I,
   /* 3F F5 */ insert_diaeresis,
   /* 40 F6 */ insert_acute,
   /* 41 F7 */ I, 
   /* 42 F8 */ SFW,
   /* 43 F9 */ RS,
   /* 44 F10 */ GOTO,
   /* 45-46 */ I, I,
   /* 47 pos1 */ MARK,
   /* 48 up */ MUP,
   /* 49 pgup */ PU,
   /* 4A alt-- */ I,
   /* 4B lf */ MLF,
   /* 4C 5 */ HOP,
   /* 4D rt */ MRT,
   /* 4E alt-+ */ I,
   /* 4F end */ YA,
   /* 50 dn */ MDN,
   /* 51 pgdn */ PD,
   /* 52 ins */ PT,
   /* 53 del */ DCC,
   /* 54 shift-F1 */ HELP,	/* HELPS ? */
   /* 55 shift-F2 */ WTU,
   /* 56 shift-F3 */ I,
   /* 57 shift-F4 */ I,
   /* 58 shift-F5 */ insert_tilde,
   /* 59 shift-F6 */ insert_grave,
   /* 5A shift-F7 */ I,
   /* 5B shift-F8 */ SRV,
   /* 5C shift-F9 */ RS,
   /* 5D shift-F10 */ GOTO,
   /* 5E ctrl-F1 */ HELP,	/* HELPC ? */
   /* 5F ctrl-F2 */ EXED,
   /* 60 ctrl-F3 */ VIEW,
   /* 61 ctrl-F4 */ I,
   /* 62 ctrl-F5 */ insert_angstrom,
   /* 63 ctrl-F6 */ insert_circumflex,
   /* 64 ctrl-F7 */ I,
   /* 65 ctrl-F8 */ RS,
   /* 66 ctrl-F9 */ LNCI,
   /* 67 ctrl-F10 */ GOTO,
   /* 68 alt-F1 */ HELP,	/* HELPA ? */
   /* 69 alt-F2 */ WB,
   /* 6A alt-F3 */ I,
   /* 6B alt-F4 */ I,
   /* 6C alt-F5 */ I,
   /* 6D alt-F6 */ I,
   /* 6E alt-F7 */ I,
   /* 6F alt-F8 */ SRV,
   /* 70 alt-F9 */ LNSW,
   /* 71 alt-F10 */ GOTO,
   /* 72 */ I,
   /* 73 ctrl-lf */ MPW,
   /* 74 ctrl-rt */ MNW,
   /* 75 ctrl-end */ DT,
   /* 76 ctrl-pgdn */ SD,
   /* 77 ctrl-pos1 */ HOP,
   /* 78 alt-1 */ I,
   /* 79 alt-2 */ I,
   /* 7A alt-3 */ I,
   /* 7B alt-4 */ I,
   /* 7C alt-5 */ I,
   /* 7D alt-6 */ I,
   /* 7E alt-7 */ I,
   /* 7F alt-8 */ I,
   /* 80 alt-9 */ I,
   /* 81 alt-0 */ I,
   /* 82-83 */ I, I,
   /* 84 ctrl-pgup */ SU,
   /* 85 F11 */ LOWCAP,
   /* 86 F12 */ I,
   /* 87 shift-F11 */ I,
   /* 88 shift-F12 */ I,
   /* 89 ctrl-F11 */ I,
   /* 8A ctrl-F12 */ I,
   /* 8B alt-F11 */ I,
   /* 8C alt-F12 */ I,
   /* 8D ctrl-up */ MUP,
   /* 8E ctrl-- */ I,
   /* 8F ctrl-5 */ HOP,
   /* 90 ctrl-+ */ I,
   /* 91 ctrl-dn */ MDN,
   /* 92 ctrl-ins */ I,
   /* 93 ctrl-del */ I,
   /* 94 */ I,
   /* 95 ctrl-/ */ I,
   /* 96 ctrl-* */ I,
   /* 97 alt-pos1 */ I,
   /* 98 alt-up */ I,
   /* 99 alt-pgup */ I,
   /* 9A */ I,
   /* 9B alt-lf */ I,
   /* 9C */ I,
   /* 9D alt-rt */ I,
   /* 9E */ I,
   /* 9F alt-end */ I,
   /* A0 alt-dn */ I,
   /* A1 alt-pgdn */ I,
   /* A2 alt-ins */ I,
   /* A3 alt-del */ I,
   /* A4 alt-/ */ I,
   /* A5-AF */ I, I, I, I, I, I, I, I, I, I, I,
   /* B0-BF */ I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,
   /* C0-CF */ I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,
   /* D0-DF */ I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,
   /* E0-EF */ I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,
   /* F0-FF */ I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,
};

#ifndef msdos

/* The mapping between function key sequences and functions */

struct fkeyentry {
	char * fk;
	voidfunc fp;
} fkeymap [] =
{	/* Escape-key sequences of function keys */

	/* Alt-Letter sequences for manual menu activation */
	{"\033f",	FILEMENU},
	{"\033e",	EDITMENU},
	{"\033s",	SEARCHMENU},
	{"\033x",	EXTRAMENU},
	{"\033p",	PARAMENU},

	/* Alt-Letter sequences for keyboard switching functions 
	   to be recognised on prompt line */
	{"\033k",	toggleKEYMAP},
	{"\033K",	setupKEYMAP},

	/* VT100 and SUN */
	{"\033[A",	MUP},	/* move cursor up	*/
	{"\033[B",	MDN},	/* more cursor down	*/
	{"\033[C",	MRT},	/* more cursor right	*/
	{"\033[D",	MLF},	/* more cursor left	*/

	/* SUN raw function key codes */
	{"\033[215z",	MUP},	/* move cursor up	*/
	{"\033[221z",	MDN},	/* more cursor down	*/
	{"\033[219z",	MRT},	/* more cursor right	*/
	{"\033[217z",	MLF},	/* more cursor left	*/

	/* SuSe Linux console */
	{"\033[1~",	MARK},	/* Pos1	*/
	{"\033[4~",	YA},	/* End	*/

	/* RedHat Linux console */
	{"\033[[A",	HELP},	/* F1	*/
	{"\033[[B",	WT},	/* F2	*/
	{"\033[24~",	WTU},	/* shift F2 / F12	*/
	{"\033[[C",	EDIT},	/* F3	*/
	{"\033[[D",	INSFILE},	/* F4	*/
	{"\033[26~",	WB},	/* shift F4	*/
	{"\033[[E",	insert_diaeresis},	/* F5	*/
	{"\033[G",	HOP},	/* 5	*/

	/* Linux X / xterm */
	{"\033OP",	HELP},	/* F1	*/
	{"\033OQ",	WT},	/* F2	*/
	{"\033OR",	EDIT},	/* F3	*/
	{"\033OS",	INSFILE},	/* F4	*/
	{"\033OH",	HOMEkey},	/* Pos1	*/
	{"\033OF",	ENDkey},	/* End	*/
	{"\033OE",	HOP},	/* 5	*/
	{"\033Ol",	DT},	/* Del	*/
	/* non-application mode xterm */
	{"\033[H",	BLINE},	/* Pos1	*/
	{"\033[F",	ELINE},	/* Ende	*/
	{"\033[E",	HOP},	/* 5	*/

	/* xterm shifted keypad keys */
	{"\033O2H",	shiftHOMEkey},	/* shift-Pos1	*/
	{"\033O2F",	shiftENDkey},	/* shift-End	*/
	{"\033O5H",	BLINE},	/* ctrl-Pos1	*/
	{"\033O5F",	ELINE},	/* ctrl-End	*/
	{"\033O2A",	HIGH},	/* shift-Up	*/
	{"\033O2D",	BLINE},	/* shift-Left	*/
	{"\033O2B",	LOW},	/* shift-Down	*/
	{"\033O2C",	ELINE},	/* shift-Right	*/
	{"\033O5A",	MUP},	/* control-Up	*/
	{"\033O5D",	ctrl_MLF},	/* control-Left	*/
	{"\033O5B",	MDN},	/* control-Down	*/
	{"\033O5C",	ctrl_MRT},	/* control-Right	*/
	{"\033O6A",	HIGH},	/* control-shift-Up	*/
	{"\033O6D",	BLINE},	/* control-shift-Left	*/
	{"\033O6B",	LOW},	/* control-shift-Down	*/
	{"\033O6C",	ELINE},	/* control-shift-Right	*/
	/* non-application mode */
	{"\033[2H",	MARK},	/* shift-Pos1	*/
	{"\033[2F",	YA},	/* shift-Ende	*/
	{"\033[5H",	BLINE},	/* ctrl-Pos1	*/
	{"\033[5F",	ELINE},	/* ctrl-Ende	*/
	{"\0332A",	HIGH},	/* shift-Up	*/
	{"\0332D",	BLINE},	/* shift-Left	*/
	{"\0332B",	LOW},	/* shift-Down	*/
	{"\0332C",	ELINE},	/* shift-Right	*/
	{"\0335A",	MUP},	/* control-Up	*/
	{"\0335D",	ctrl_MLF},	/* control-Left	*/
	{"\0335B",	MDN},	/* control-Down	*/
	{"\0335C",	ctrl_MRT},	/* control-Right	*/
	{"\0336A",	HIGH},	/* control-shift-Up	*/
	{"\0336D",	BLINE},	/* control-shift-Left	*/
	{"\0336B",	LOW},	/* control-shift-Down	*/
	{"\0336C",	ELINE},	/* control-shift-Right	*/

	/* hanterm */
	{"\033O ",	HOP},	/* middle keypad key ("5")	*/

	/* VT100 - some have been disabled by linux sequences */
	{"\033[6~",	PD},	/* Next Screen	*/
	{"\033[5~",	PU},	/* Prev Screen	*/
/*	{"\033[4~",	MARK},	*//* Select	*/
	{"\033[3~",	DT},	/* Remove	*/
	{"\033[2~",	PT},	/* Insert Here	*/
/*	{"\033[1~",	SFW},	*//* Find		*/
/*	{"\033[29~",	YA},	*//* Do		*/
	{"\033[29~",	SFW},	/* Do		*/

	{"\033[28~",	HELP},	/* Help		*/

	{"\033[11~",	HELP},	/* F1	*/
	{"\033[12~",	WT},	/* F2	*/
	{"\033[13~",	EDIT},	/* F3	*/
	{"\033[14~",	INSFILE},	/* F4	*/
	{"\033[15~",	insert_diaeresis},	/* F5	*/

	{"\033[17~",	insert_acute}, /* F6	*/
	{"\033[17~",	insert_acute}, /* F6	*/
	{"\033[18~",	SFW},	/* F7	*/
	{"\033[19~",	SFW},	/* F8	*/
	{"\033[20~",	RS},	/* F9	*/
	{"\033[21~",	FILEMENU},	/* F10	*/
	{"\033[23~",	LOWCAP},	/* F11	*/
	{"\033[24~",	UML},	/* F12	*/

	{"\033[25~",	I},	/* F13	*/
	{"\033[26~",	I},	/* F14	*/
	{"\033[31~",	I},	/* F17	*/
	{"\033[32~",	I},	/* F18	*/
	{"\033[33~",	I},	/* F19	*/
	{"\033[34~",	QUICKMENU},	/* F20	*/
/*	{"\033OP",	HOP},	*//* PF1	*/
/*	{"\033OQ",	SU},	*//* PF2	*/
/*	{"\033OR",	SD},	*//* PF3	*/
/*	{"\033OS",	SU},	*//* PF4	*/
	{"\033[@",	BLINE},	/* Pos1, PC-xterm	*/
	{"\033[e",	ELINE},	/* End, PC-xterm	*/
	/* application cursor keys */
	{"\033OA",	MUP},	/* up	*/
	{"\033OB",	MDN},	/* dn	*/
	{"\033OC",	MRT},	/* rt	*/
	{"\033OD",	MLF},	/* lf	*/
	{"\033O@",	BLINE},	/* Pos1, PC-xterm	*/
	{"\033Oe",	ELINE},	/* End, PC-xterm	*/
	/* application keypad */
	{"\033Oo",	I},	/* PF1, /	*/
	{"\033Oj",	I},	/* PF2, *	*/
	{"\033Om",	SU},	/* PF3, -	*/
	{"\033Ok",	SD},	/* PF4, +	*/
	{"\033OM",	SNL},	/* Enter */
	{"\033Ow",	MARK},	/* Pos1	*/
	{"\033Ox",	MUP},	/* up	*/
	{"\033Oy",	PU},	/* PgUp	*/
	{"\033Ot",	MLF},	/* lf	*/
	{"\033Ou",	HOP},	/* mid	*/
	{"\033Ov",	MRT},	/* rt	*/
	{"\033Oq",	YA},	/* End	*/
	{"\033Or",	MDN},	/* dn	*/
	{"\033Os",	PD},	/* PgDn	*/
	{"\033Op",	PT},	/* Ins	*/
	{"\033On",	DT},	/* Del	*/

	/* xterm shifted function keys */
	{"\033[11;2~",	I},	/* shift-F1	*/
	{"\033[12;2~",	WTU},	/* shift-F2	*/
	{"\033[13;2~",	I},	/* shift-F3	*/
	{"\033[14;2~",	I},	/* shift-F4	*/
	{"\033[15;2~",	insert_tilde},	/* shift-F5	*/
	{"\033[17;2~",	insert_grave},	/* shift-F6	*/
	{"\033[18;2~",	I},	/* shift-F7	*/
	{"\033[19;2~",	GR},	/* shift-F8	*/
	{"\033[20;2~",	SIDFW},	/* shift-F9	*/
	{"\033[21;2~",	QUICKMENU},	/* shift-F10	*/
	{"\033[23;2~",	I},	/* shift-F11	*/
	{"\033[24;2~",	I},	/* shift-F12	*/
	{"\033[15;5~",	insert_angstrom},	/* control-F5 */
	{"\033[17;5~",	insert_circumflex},	/* control-F6 */
	{"\033[15;6~",	insert_angstrom},	/* control-shift-F5 */
	{"\033[17;6~",	insert_circumflex},	/* control-shift-F6 */

	/* SUN */
	{"\033[208z",	insert_grave},	/* R1 */
	{"\033[209z",	insert_circumflex},	/* R2 */
	{"\033[210z",	insert_acute},	/* R3 */
	{"\033[211z",	insert_diaeresis},	/* R4 */
	{"\033[212z",	insert_tilde},	/* R5 */
	{"\033[213z",	insert_angstrom},	/* R6 */
	{"\033[214z",	SU},	/* R7	*/
	{"\033[216z",	PU},	/* R9	*/
	{"\033[218z",	HOP},	/* R11	*/
	{"\033[220z",	SD},	/* R13	*/
	{"\033[222z",	PD},	/* R15	*/
	{"\033[254z",	GOMA},	/* - */
	{"\033[253z",	MARK},	/* + */
	{"\033[250z",	YA},	/* Enter */
	{"\033[249z",	DT},	/* Del */
	{"\033[247z",	PT},	/* Ins */
	{"\033[1z",	BLINE},	/* Pos1, non-appl */
	{"\033[4z",	ELINE},	/* End, non-appl */
	{"\033[2z",	PT},	/* Ins on VT100 block */
	{"\033[3z",	DT},	/* Del (xterm/Sun) */
	{"\033[5z",	PU},	/* PgUp on VT100 block */
	{"\033[6z",	PD},	/* PgDn on VT100 block */
	{"\033[192z",	MARK},	/* Stop */
	{"\033[193z",	RS},	/* Again */
	{"\033[194z",	FS},	/* Props */
	{"\033[195z",	I},	/* Undo */
	{"\033[197z",	YA},	/* Copy */
	{"\033[199z",	PT},	/* Paste */
	{"\033[200z",	RS},	/* Find */
	{"\033[201z",	DT},	/* Cut */
	{"\033[196z",	HELP},	/* Help */

	{"\033[225z",	WT},	/* F2 */
	{"\033[226z",	EDIT},	/* F3 */
	{"\033[227z",	INSFILE},	/* F4 */
	{"\033[228z",	insert_diaeresis},	/* F5 */
	{"\033[229z",	PU},	/* F6 */
	{"\033[230z",	PD},	/* F7 */
	{"\033[231z",	SFW},	/* F8 */
	{"\033[232z",	RS},	/* F9 */
	{"\033[233z",	FILEMENU},	/* F10 */
	{"\033[234z",	LOWCAP},	/* F11 */
	{"\033[235z",	I},	/* F12 */
	{"\033[224z",	HELP},	/* F1, xterm */
/*	{"\033[192z",	I},	*//* F11, xterm; Sun "Stop" -> MARK */
/*	{"\033[193z",	I},	*//* F12, xterm; Sun "Again" */

	/* xterm shifted function keys, Sun mode */
	{"\033[224;2z",	I},	/* shift-F1 */
	{"\033[225;2z",	WTU},	/* shift-F2 */
	{"\033[226;2z",	I},	/* shift-F3 */
	{"\033[227;2z",	I},	/* shift-F4 */
	{"\033[228;2z",	insert_tilde},	/* shift-F5 */
	{"\033[229;2z",	insert_grave},	/* shift-F6 */
	{"\033[230;2z",	I},	/* shift-F7 */
	{"\033[231;2z",	GR},	/* shift-F8 */
	{"\033[232;2z",	SIDFW},	/* shift-F9 */
	{"\033[233;2z",	I},	/* shift-F10 */
	{"\033[192;2z",	I},	/* shift-F11 */
	{"\033[193;2z",	I},	/* shift-F12 */
	{"\033[228;5z",	insert_angstrom},	/* control-F5 */
	{"\033[229;5z",	insert_circumflex},	/* control-F6 */
	{"\033[228;6z",	insert_angstrom},	/* control-shift-F5 */
	{"\033[229;6z",	insert_circumflex},	/* control-shift-F6 */

	/* Iris */
	{"\033[209q",	MARK},	/* Print Screen */
	{"\033[213q",	YA},	/* Scroll Lock */
/*	{"\033[217q",	I},	*//* Pause */
	{"\033[139q",	PT},	/* Insert, Ins */
	{"\033[P",	DT},	/* Del */
/*	{"\033[H",	SU},	*//* Home */
	{"\033[150q",	PU},	/* PgUp */
	{"\033[146q",	SD},	/* End */
	{"\033[154q",	PD},	/* PgDn */
	{"\033[000q",	HOP},	/* 5 */
	{"\033[001q",	HELP},	/* F1 */
	{"\033[002q",	WT},	/* F2 */
	{"\033[005q",	insert_diaeresis},	/* F5 */
	{"\033[006q",	insert_acute},	/* F6 */
	{"\033[007q",	I},	/* F7 */
	{"\033[008q",	SFW},	/* F8 */
	{"\033[009q",	RS},	/* F9 */
	{"\033[022q",	insert_grave},	/* Shift-F10 */
	{"\033[023q",	insert_circumflex},	/* Shift-F11 */
	{"\033[024q",	insert_acute},	/* Shift-F12 */
	{"\033[010q",	insert_diaeresis},	/* F10 */
	{"\033[011q",	insert_tilde},	/* F11 */
	{"\033[012q",	insert_angstrom},	/* F12 */

	/* HP workstation keyboard without NumLock */
	{"\033OU",	MARK},	/* Pos1	*/
	{"\033OW",	MUP},	/* up	*/
	{"\033OZ",	PU},	/* PgUp	*/
	{"\033OV",	MLF},	/* lf	*/
	{"\033OX",	MRT},	/* rt	*/
	{"\033OY",	MDN},	/* dn	*/
	{"\033O?",	PD},	/* PgDn, End, Ins, Del	*/

	{NIL_PTR}
};

struct fkeyentry fkeymap_xterm [] = {
	/* xterm */
	{"\033[M",	DIRECTxterm},	/* direct cursor address */
	{"\033[t",	TRACKxterm},	/* cursor hilite tracking */
	{"\033[T",	TRACKxtermT},	/* cursor hilite tracking */

	{NIL_PTR}
};

struct fkeyentry fkeymap_hp [] = {
	/* crttool */
	{"\033m",	DIRECTcrttool},	/* direct cursor address */

	/* HP, overrides some of Siemens and VT100 PC-emulator */
	{"\033h",	MARK},	/* Pos1 */
	{"\033A",	MUP},	/* up	*/
	{"\033T",	PU},	/* PgUp	*/
	{"\033D",	MLF},	/* lf	*/
	{"\033C",	MRT},	/* rt	*/
	{"\033F",	YA},	/* End	*/
	{"\033B",	MDN},	/* dn	*/
	{"\033S",	PD},	/* PgDn	*/
	{"\033Q",	PT},	/* Ins	*/
	{"\033P",	DT},	/* Del	*/
	{"\033p",	HELP},	/* F1 */
	{"\033q",	WT},	/* F2 */
	{"\033r",	EDIT},	/* F3 */
	{"\033s",	INSFILE},	/* F4 */
	{"\033t",	insert_diaeresis},	/* F5 */
	{"\033u",	PU},	/* F6 */
	{"\033v",	PD},	/* F7 */
	{"\033w",	SFW},	/* F8 */
	{"\0332p",	I},	/* shift-F1 */
	{"\0332q",	WTU},	/* shift-F2 */
	{"\0332r",	I},	/* shift-F3 */
	{"\0332s",	I},	/* shift-F4 */
	{"\0332t",	insert_tilde},	/* shift-F5 */
	{"\0332u",	insert_grave},	/* shift-F6 */
	{"\0332v",	I},	/* shift-F7 */
	{"\0332w",	GR},	/* shift-F8 */

	{NIL_PTR}
};

struct fkeyentry fkeymap_siemens [] = {
	/* VT100 emulator at IBM-PC */
	{"\033M",	PU},	/* Pg Up */
	{"\033D",	PD},	/* Pg Dn */
/*	{"\033[H",	SU},	*//* Home	, same at Iris */
	{"\033[24H",	SD},	/* End; actually "\033[24;1H", ";1" filtered out */
	{"\0331",	HOP},	/* Ins	*/
		/* cursor keys and F1-4 same as VT100 */

	/* Siemens 97801/97808 */
	{"\033A",	WT},
	{"\033;",	WTU},
	{"\033B",	EDIT},
	{"\033\"",	VIEW},
	{"\033N",	insert_diaeresis},	/* F14 */
	{"\033P",	RS},			/* F16 */
	{"\0332",	insert_grave},		/* shift-F18 */
	{"\0333",	insert_circumflex},	/* shift-F19 */
	{"\033U",	insert_acute},		/* shift-F20 */
	{"\033T",	insert_diaeresis},	/* F20 */
	{"\033d",	insert_tilde},		/* F19 */
	{"\033_",	insert_angstrom},	/* F18 */
	{"\0330",	UML},			/* F17 */
	{"\033[Z",	MPW},		/* |<- */
	{"\0339",	MPW},		/* <~- */
	{"\033:",	MNW},		/* -~> */
	{"\0336",	HOP},		/* neben END, ber Cursorblock */
	{"\033Y",	TOGGO},		/* shift-F22 */
/*
INSERT:	[@	o	[L
DELETE:	([P)	p	[M
|<-:	[Z
*/
	{"\033[S",	PU},
	{"\033[T",	PD},
	{"\033^",	HOP},
	{NIL_PTR}
};

#define MAXCODELEN 7 /* max. length of function key sequence to be detected, 
			depending on the fkeymap table => used in keyboard.c */

#endif


/*======================================================================*\
|*				End					*|
\*======================================================================*/
