/*======================================================================*\
|*		Editor mined						*|
|*		part 1							*|
\*======================================================================*/

#include "mined.h"
#include "io.h"

#include <errno.h>


/*======================================================================*\
|*			Definitions specific for mined1.c		*|
\*======================================================================*/

# ifdef unix
# define helpcommand "man mined"
# define helpfilename "%s/mined.help"
# endif

# ifdef vms
# define helpcommand "help mined"
# define helpfilename "%s/mined.help"
# endif

# ifdef msdos
# define helpcommand "more < %smined.hlp"
# define helpfilename "%s/mined.hlp"
# endif

# ifdef unix
#  ifdef sysV
#  define printcommand "lp %s"
#  else
#  define printcommand "lpr %s"
#  endif
# endif

# ifdef vms
# define printcommand "print %s"
# endif

# ifdef msdos
# define printcommand "copy %s prn: > nul:"
# endif

#define mark_file	"@mined.mar"


#ifdef msdos
#ifdef __TURBOC__
#define msdos_screenfunctions
#endif
#endif

/* forward */
void set_file_type_flags ();


/*======================================================================*\
|*			Screen stuff					*|
\*======================================================================*/

#define clearscreen()	clear_screen (); top_line_scrolled = True;

/*======================================================================*\
|*			Data section					*|
\*======================================================================*/

LINE * header;			/* Head of line list */
LINE * tail;			/* Last line in line list */
LINE * cur_line;		/* Current line in use */
LINE * top_line;		/* First line of screen */
LINE * bot_line;		/* Last line of screen */
char * cur_text;		/* Current char on current line in use */
int last_y;			/* Last y of screen. Usually SCREENMAX */
int x = 0, y = 0;		/* x, y coordinates on screen */
int line_number;		/* current line # determined by file_status */
int lines_per_page = 0;		/* assumption for file_status */
int open_linum;			/* line # to re-position to */
int open_col;			/* column to re-position to */
LINE * open_line;		/* line to re-position to */

short YMAX, XMAX;
short MENU = 1;
mousebutton mouse_button, mouse_lastbutton, mouse_prevbutton;
FLAG report_release = False;
int mouse_shift = 0;
int mouse_xpos, mouse_ypos, mouse_lastxpos, mouse_lastypos;

char linesplitmarker = '\0';	/* Leave 1 char. space BEFORE screen [] ! */
char screen [screen_BUFL + 1];	/* I/O buffer for "writes" and "reads" */

FLAG flags_changed = False;	/* should flags menu area be redrawn? */
FLAG quickmenu = True;		/* Right mouse button pops up menu */
FLAG scrollbar = True;		/* use scrollbar */
int total_lines = 0;		/* Number of lines in file */
long total_chars = -1L;		/* Number of characters in file */
FLAG modified = False;		/* Set when file is modified */
FLAG plus_opt = False;		/* Set if options start with '+' */
FLAG viewonly = False;		/* Set when view only mode is selected */
FLAG append_flag = False;	/* Set when buffer should be appended to */
FLAG restricted = False;	/* Set when edited file shall not be switched */
FLAG mined_keypad = True;	/* Apply mined keypad assignments */
FLAG overwriteOK = False;	/* Set if current file is OK for overwrite */
FLAG tab_left = True;		/* Set if moving up/down on TAB should go left */
FLAG writable;			/* Set if file cannot be written */
int JUSlevel = 0;		/* keep justified while typing? */
int JUSmode = 0;		/* 1: paragraphs end at empty line */
FLAG autoindent = True;		/* auto indent on input of Enter? */
FLAG dim_HTML = True;		/* display HTML dimmed ? */
FLAG loading = True;		/* Loading a file? Init True for error handling */
FLAG quit = False;		/* Set when quit character is typed */
FLAG intr_char = False;		/* Set when intr character is typed */
FLAG winchg = False;		/* Set when window size has changed */
FLAG interrupted = False;	/* Set when a signal interrupts */
FLAG isscreenmode = False;	/* Set when screen mode is on */
FLAG stat_visible;		/* Set if status_line is visible */
FLAG top_line_scrolled = False;	/* was menu line scrolled away? */
FLAG fstat_always = False;	/* Permanent file status display wanted ? */
FLAG always_disp_code = False;	/* Permanent char code display on status line? */
FLAG waitingforinput = False;	/* Set while waiting for the next command key */
FLAG rpipe = False;		/* Set if file should be read from stdin */
FLAG wpipe = False;		/* Set if file should be written to stdout */
FLAG multiexit = True;		/* Should exit command go to next file? */
FLAG wordnonblank = False;	/* handle all non-blank sequences as words */
FLAG proportional = False;	/* Enable support for proportional fonts? */
FLAG transdeutsch = False;	/* Transform input {->, ..., ~-> ? */
FLAG auto_transdeutsch = False;	/* Auto detect transdeutsch mode from file ? */
FLAG auto_UTF_detect = True;	/* Auto detect UTF-8 encoding from file ? */
FLAG translate_output = False;	/* Transform output diacritics to strings */
int tabsize = 8;		/* width of tab positions, 4 or 8 */
char * transout;		/* Output transformation table */
char * dimansi;			/* special character dimming ANSI sequence */
char * ctrlansi;		/* control character display ANSI sequence */
char * uniansi;			/* Unicode character display ANSI sequence */
char * unimarkansi;		/* Unicode (lineend) marker display ANSI sequence */
char * combiningansi;		/* combining character display ANSI sequence */
char * menuansi;		/* menu line ANSI sequence */
char * HTMLansi;		/* HTML display ANSI sequence */
char * diagansi;		/* dialog (bottom status) line ANSI sequence */
char * scrollfgansi;		/* scrollbar foreground ANSI sequence */
char * scrollbgansi;		/* scrollbar background ANSI sequence */
int translen;			/* length of " */
FLAG controlQS = False;		/* must respect ^Q/^S handshake ? */
FLAG insert_mode = True;	/* insert or overwrite */
character control_prefix = '\026';	/* ^V/^P character to prefix control chars */
character erase_char = '\010';	/* effective (configured) char for erase left */
FLAG Chinese = False;		/* set if two-byte characters are enabled */
FLAG Turkish = False;		/* enable Turkish case toggle specials ? */
FLAG smart_quotes = True;	/* replace " with typographic quote ? */
FLAG utf8_screen = False;	/* screen driven in UTF-8 mode ? */
FLAG utf8_input = False;	/* keyboard input in UTF-8 mode ? */
FLAG utf8_text = False;		/* text in UTF-8 representation ? */
FLAG utf8_lineends = False;	/* detect UTF-8 LS and PS line ends ? */
FLAG bidi = False;		/* poor man's bidi support ? */
FLAG disp_scrollbar = True;	/* shall scrollbar be displayed ? */
int scrollbar_width = 1;
FLAG update_scrollbar_lazy = False;	/* avoid much scrollbar output ? */
FLAG combining_mode = False;	/* UTF-8 combining character display support ? */
FLAG combining_screen = False;	/* UTF-8 combining character display enabled by terminal ? */
FLAG enforce_block_graphics_menu = False; /* use block graphics as menu border */
FLAG enable_block_graphics = False;	/* display control chars as graphics */
FLAG no_window_title = False;	/* suppress filename display in window title? */

FLAG page_scroll = False;	/* use scroll for page up/down */
FLAG page_stay = False;		/* stay at edge of screen after page up/down */
#ifdef pc_delay
int display_delay = 9;		/* delay between display lines */
#else
int display_delay = -1;		/* Unix terminals are slow enough anyway */
#endif

#ifdef msdos_with_auto_crlf
char RET_opt = 'r';		/* handle RET chars: ignore / newline */
#else
char RET_opt = ' ';		/* handle RET chars: ignore / newline */
#endif
lineend_type default_lineend = lineend_LF;	/* used for some inserts */

long chars_saved;		/* # of chars in paste buffer */
long bytes_saved;		/* # of bytes in paste buffer */
int lines_saved;		/* # of lines in paste buffer */
int input_fd = STD_IN;		/* File descriptors for terminal dialog */
#ifdef __EMX__
int output_fd = STD_OUT;
#else
int output_fd = STD_ERR;
#endif
unsigned int out_count = 0;	/* Index in output buffer */
char file_name [maxLINE_LEN];	/* Name of file in use */
char text_buffer [MAX_CHARS];	/* for get_line, modifications, build_string */
int escape_delay = 450;		/* wait to detect escape sequence */
int hop_flag = 0;		/* Counter for the HOP function */
char TABchar = ' ';		/* Char to be shown in place of tab chars */
char TABchar0 = '\0';		/* Char to be shown at start of tab chars */
char TABchar2 = '\0';		/* Char to be shown at end of tab chars */
char TABcharmid = '\0';		/* Char to be shown in middle of tab chars */
char SHIFT_BEG = '\0';		/* Char indicating that line continues left */
char RET_MARK = '\0';		/* Char indicating end of line */
char DOSRET_MARK = '\0';	/* Char indicating DOS end of line */
char PARA_MARK = '\0';		/* Char indicating end of paragraph */
char RET_BLANK = '\0';		/* Char to fill the end of line with */
char RET_BLANK2 = '\0';		/* Char to fill last position of line with */
char * UTF_TAB = NIL_PTR;	/* Char to be shown in place of tab chars */
char * UTF_TAB0 = NIL_PTR;	/* Char to be shown at start of tab chars */
char * UTF_TAB2 = NIL_PTR;	/* Char to be shown at end of tab chars */
char * UTF_TABmid = NIL_PTR;	/* Char to be shown in middle of tab chars */
char * UTF_RET = NIL_PTR;	/* Char indicating end of line */
char * UTF_DOSRET = NIL_PTR;	/* Char indicating DOS end of line */
char * UTF_PARA = NIL_PTR;	/* Char indicating end of paragraph */
char * UTF_RETblank = NIL_PTR;	/* Char to fill the end of line with */
char * UTF_RETblank2 = NIL_PTR;	/* Char to fill last position of line with */
FLAG paradisp = False;		/* Shall paragraph end be distinguished? */
#ifdef vms
unsigned int fprot = 0;		/* To be used for file creatings */
unsigned int bufprot = 0;	/* To be used for paste buffer file */
unsigned int exeprot = 0;	/* To be used for creation of executables */
#else
PROT fprot = 0644;	/* To be used for file creatings */
PROT bufprot = 0600;	/* To be used for paste buffer file */
PROT exeprot = 0111;	/* To be used for creation of executables */
#endif
PROT xprot = 0;
int fnami;			/* Parameter index of current file name */
int fnami_min, fnami_max, fnami_cnt;
/* char * (* fnamv) []; */
char * * fnamv;			/* Copy of argv. Points to program params */

char * inisearch = NIL_PTR;	/* Optional startup search string */

int first_left_margin = 0;
int next_left_margin = 0;
int right_margin = 71;

#ifdef unix
/* window headline and icon text setting */
char * window_string_code = "";
char * mined_modf = " (*)";
#endif

/*
 * Yank variables.
 */
char * temp_dir;
char yank_file [maxLINE_LEN];
char yankie_file [maxLINE_LEN];
char html_file [maxLINE_LEN];	/* Temp. file for HTML embedding buffer */
char panic_file [maxLINE_LEN];


/*======================================================================*\
|*			Text string routines				*|
\*======================================================================*/

int
UTF_len (c)
  char c;
{
	if ((c & 0x80) == 0x00)
		return 1;
	else if ((c & 0xE0) == 0xC0)
		return 2;
	else if ((c & 0xF0) == 0xE0)
		return 3;
	else if ((c & 0xF8) == 0xF0)
		return 4;
	else if ((c & 0xFC) == 0xF8)
		return 5;
	else if ((c & 0xFE) == 0xFC)
		return 6;
	else /* illegal UTF-8 code */
		return 1;
}
int
UTFseq_len (text)
  char * text;
{
	int follow = UTF_len (* text) - 1;
	int len = 1;
	text ++;
	while (follow > 0 && (* text & 0xC0) == 0x80) {
		len ++;
		text ++;
		follow --;
	}
	return len;
}

int
uniUTF (unichar, buf)
  unsigned long unichar;
  character * buf;
{
  int len;

  if (unichar < 0x80) {
	len = 1;
	* buf ++ = unichar;
  } else if (unichar < 0x800) {
	len = 2;
	* buf ++ = 0xC0 | (unichar >> 6);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x10000) {
	len = 3;
	* buf ++ = 0xE0 | (unichar >> 12);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x200000) {
	len = 4;
	* buf ++ = 0xF0 | (unichar >> 18);
	* buf ++ = 0x80 | ((unichar >> 12) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x4000000) {
	len = 5;
	* buf ++ = 0xF8 | (unichar >> 24);
	* buf ++ = 0x80 | ((unichar >> 18) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 12) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x80000000) {
	len = 6;
	* buf ++ = 0xFC | (unichar >> 30);
	* buf ++ = 0x80 | ((unichar >> 24) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 18) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 12) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else {
	len = 0;
  }
  * buf = '\0';
  return len;
}

void
utf8_info (u, length, unicode)
  char * u;
  int * length;
  unsigned long * unicode;
{
  char * textpoi = u;
  character c = * textpoi;
  int utfcount;
  unsigned long unichar;

	if ((c & 0x80) == 0x00) {
		utfcount = 1;
		unichar = c;
	} else if ((c & 0xE0) == 0xC0) {
		utfcount = 2;
		unichar = c & 0x1F;
	} else if ((c & 0xF0) == 0xE0) {
		utfcount = 3;
		unichar = c & 0x0F;
	} else if ((c & 0xF8) == 0xF0) {
		utfcount = 4;
		unichar = c & 0x07;
	} else if ((c & 0xFC) == 0xF8) {
		utfcount = 5;
		unichar = c & 0x03;
	} else if ((c & 0xFE) == 0xFC) {
		utfcount = 6;
		unichar = c & 0x01;
	} else /* illegal UTF-8 code */ {
		utfcount = 1;
		unichar = c;
	}

	* length = utfcount;

	utfcount --;
	textpoi ++;
	while (utfcount > 0 && (* textpoi & 0xC0) == 0x80) {
		unichar = (unichar << 6) | (* textpoi & 0x3F);
		utfcount --;
		textpoi ++;
	}
	if (utfcount > 0)
		unichar = 0;

	* unicode = unichar;
}

int
uniwidth (unichar)
  unsigned long unichar;
{
  if (combining_mode == True && iscombining (unichar))
	return 0;
  else if (iswide (unichar))
	return 2;
  else
	return 1;
}

/*
 * inChinese (string, charpoi) determines if charpoi points to the second 
 * byte of a multi-byte character within string
 */
int
inChinese (string, charpoi)
  character * string;
  character * charpoi;
{
  while (string < charpoi) {
	if (multichar (* string)) {
		string ++;
#ifdef murx
		if (* string != '\n' /* would be an error */)
#endif
			string ++;
	} else
		string ++;
  }
  return string > charpoi;
}

/*
   charbegin () determines the first byte of the character pointed to 
   in the given line
 */
char *
charbegin (line, s)
  char * line;
  char * s;
{
  char * utf_search;
  char * utf_prev;

  if (utf8_text == True) {
	utf_search = line;
	utf_prev = utf_search;
	while (utf_search < s) {
		utf_prev = utf_search;
		advance_utf8 (& utf_search);
	}
	if (utf_search > s) {
		return utf_prev;
	} else {
		return s;
	}
  } else if (Chinese == True && inChinese (line, s)) {
	s --;
  }
  return s;
}

/*
 * precede_char () moves the character pointer within line "begin_line" 
 * left by 1 character
 */
void
precede_char (poipoi, begin_line)
  char * * poipoi;
  char * begin_line;
{
  char * utf_search;
  char * utf_prev;

  if (utf8_text == True) {
	utf_search = begin_line;
	utf_prev = utf_search;
	while (utf_search < * poipoi) {
		utf_prev = utf_search;
		advance_utf8 (& utf_search);
	}
	(* poipoi) = utf_prev;
  } else if ((* poipoi) != begin_line) {
	(* poipoi) --;
	if (Chinese == True && inChinese (begin_line, * poipoi)) {
		(* poipoi) --;
	}
  }
}

/*
 * precedingchar () determines the preceding character value
 */
unsigned long
precedingchar (curpoi, begin_line)
  char * curpoi;
  char * begin_line;
{
  char * poi;
  unsigned long unichar;
  int follow;

  if (curpoi == begin_line) {
	return '\n';
  } else {
	poi = curpoi;
	precede_char (& poi, begin_line);
	if (utf8_text == True) {
		utf8_info (poi, & follow, & unichar);
		return unichar;
	} else if (Chinese == True && multichar (* poi)) {
		return ((* poi) << 8) | * (poi + 1);
	} else {
		return * poi;
	}
  }
}

/*
 * Advance pointer and counter to next character.
 * Handle tab characters and Chinese 2-byte characters correctly.
 */
void
advance_utf8_col (poipoi, colpoi)
  char * * poipoi;
  int * colpoi;
{
  unsigned long unichar;
  int follow;

	utf8_info (* poipoi, & follow, & unichar);
	follow --;
	(* poipoi) ++;
	while (follow > 0 && (* * poipoi & 0xC0) == 0x80) {
		(* poipoi) ++;
		follow --;
	}
/*
	(* colpoi) ++;
	if (iswide (unichar))
		(* colpoi) ++;
*/
	(* colpoi) += uniwidth (unichar);
}
void
advance_utf8 (poipoi)
  char * * poipoi;
{
	int follow = UTF_len (* * poipoi) - 1;
	(* poipoi) ++;
	while (follow > 0 && (* * poipoi & 0xC0) == 0x80) {
		(* poipoi) ++;
		follow --;
	}
}
void
advance_char_col (poipoi, colpoi)
  char * * poipoi;
  int * colpoi;
{
  if (is_tab (* * poipoi)) {
	(* poipoi) ++;
	* colpoi = tab (* colpoi);
  }
  else if (utf8_text == True) {
	advance_utf8_col (poipoi, colpoi);
  }
  else if (Chinese == True && multichar (* * poipoi)) {
	(* poipoi) ++;
	if (* * poipoi != '\0') (* poipoi) ++;
	(* colpoi) ++;
	(* colpoi) ++;
  }
  else {
	(* poipoi) ++;
	(* colpoi) ++;
  }
}
void
advance_char (poipoi)
  char * * poipoi;
{
  if (is_tab (* * poipoi)) {
	(* poipoi) ++;
  }
  else if (utf8_text == True) {
	advance_utf8 (poipoi);
  }
  else if (Chinese == True && multichar (* * poipoi)) {
	(* poipoi) ++;
	if (* * poipoi != '\0') (* poipoi) ++;
  }
  else {
	(* poipoi) ++;
  }
}


/*======================================================================*\
|*			Text buffer routines				*|
\*======================================================================*/

int old_x = 0;	/* previous x position */

/*
 * Find_x () returns the x coordinate belonging to address.
 * (Tabs are expanded).
 */
int
find_x (line, address)
  LINE * line;
  char * address;
{
  char * textp = line->text;
  register int x_left = get_shift (line->shift_count) * - SHIFT_SIZE;
  int x_in_line = 0;	/* must start from 0 to calculate correct 
			tab positions (since SHIFT_SIZE is not guaranteed 
			to be a multiple of tabsize (usually 8)) */
	/* Alright, SHIFT_SIZE is now guaranteed to be a multiple of 8 
	   due to lots of display problems related to that matter.
	   Leave this code anyway. */
  int utfcount;
  unsigned long unichar;
  int x_before = 0;
  int prevx;

  while (textp < address && * textp != '\0') {
	/* use "<" rather than "!=" because address may point amidst 
	   a UTF-8 character sequence after mode switching */
	prevx = x_in_line;
	advance_char_col (& textp, & x_in_line);
	/* determine screen pos. to stay on if moving on a combining char. */
	if (x_in_line > prevx)
		x_before = prevx;
  }
  if (utf8_text == True && combining_mode == True) {
	utf8_info (textp, & utfcount, & unichar);
	if (iscombining (unichar)) {
		x_in_line = x_before;
	}
  }
  return x_in_line + x_left;
}

/*
 * Find_address () returns the pointer in the line with given offset.
 * (Tabs are expanded).
 * find_address is only called by move_it ()
get_shift (cnt)		is	((cnt) & DUMMY_MASK) ; DUMMY_MASK is 0x7F
tab (cnt)		is	(((cnt) + tabsize) & ~07)
is_tab (c)		is	((c) == '\t')
 */
char *
find_address (line, new_x, cur_x)
  LINE * line;
  int new_x;
  int * cur_x;
{
  char * textp = line->text;
  int tx = get_shift (line->shift_count) * - SHIFT_SIZE;
  char * prev_poi;
  int prev_x;
  unsigned long unichar;
  int follow;

  while (tx < new_x && * textp != '\n') {
	if (is_tab (* textp)) {
		if (new_x == old_x /* (* cur_x) */ - 1 && tab (tx) > new_x)
			break;	/* Moving left over tab */
		else if (tab_left == True && new_x == old_x && tab (tx) > new_x)
			break;	/* Moving up/down on tab */
		else
			tx = tab (tx);
		textp ++;
	} else {
		prev_poi = textp;
		prev_x = tx;

		advance_char_col (& textp, & tx);
		if (utf8_text == True && combining_mode == True) {
			utf8_info (textp, & follow, & unichar);
			while (iscombining (unichar)) {
				advance_char_col (& textp, & tx);
				utf8_info (textp, & follow, & unichar);
			}
		}

		if (tx > new_x && new_x < * cur_x) {
		/* moving left into multi-width character, stay before it */
			textp = prev_poi;
			tx = prev_x;
			new_x = tx;
		}
	}
  }
  * cur_x = tx;
  return textp;
}

/*
 * move_to: move to given coordinates on screen.
 * move_y: move to given line on screen, staying in last explicit column.
 * move_address: move to given line at given text position.
 * The caller must check that scrolling is not needed.
 * If new x-position is < 0 or > XBREAK, move_it () will check if
 * the line can be shifted. If it can it sets (or resets) the shift_count
 * field of the current line accordingly. By this mechanism, the
 * pseudo-x-positions LINE_START / LINE_END (a very small / big value)
 * perform the appropriate positioning actions.
 * Move also sets cur_text to the right char.
 * "If we're moving to the same x coordinate, try to move to the x-coordinate
 * used on the other previous call." -- This worked erroneously and was 
 * replaced by an explicit old_x variable and move_y call.
 * move_address is directly called by move_next/previous_word(), 
 * re_search(), RDwin(), load_file(), justi(), JUSandreturn()
 */
void
move_it (new_x, new_address, new_y)
  register int new_x;
  int new_y;
  char * new_address;
{
  register LINE * line = cur_line;	/* For building new cur_line */
  register int lineno = line_number;
  int shift = 0;			/* How many shifts to make */
  char * utf_search;
  char * utf_prev;

/*  static int rel_x = 0;	*/	/* Remember relative x position */
/*	This was used as a trick to stay virtually in the previous column 
	even when moving across shorter lines; but it had >= 2 errors.
	Renamed to old_x, made globally accessible and explicitly used 
	by appropriate calls to avoid these problems. TW */
  int tx = x;

/* Check for illegal values */
  if (new_y < 0 || new_y > last_y)
	return;

/* Adjust y-coordinate and cur_line */
  if (new_y < y)
	while (y != new_y) {
		y --;
		line = line->prev;
		lineno --;
	}
  else
	while (y != new_y) {
		y ++;
		line = line->next;
		lineno ++;
	}

/* Set or unset relative x-coordinate */
  if (new_address == NIL_PTR) {
	new_address = find_address (line, new_x, & tx);
	new_x = tx;
  }
  else
	/* rel_x = */ new_x = find_x (line, new_address);

/* Adjust on character boundary */
  if (Chinese == True) {
	if (inChinese (line->text, new_address)) {
	/* adjust position which is currently within a multi-byte character */
		if (new_x >= x) {
			new_x ++;	/* Chinese char has double width */
			new_address ++;
		} else {
			new_x --;	/* Chinese char has double width */
			new_address --;
		}
	}
  } else if (utf8_text == True) {
	if ((* new_address & 0xC0) == 0x80) {	/* UTF-8 sequence byte? */
	/* adjust position which is currently within a UTF-8 character */
	/*  adjust new_x for multi-width characters ? */
		utf_search = line->text;
		utf_prev = utf_search;
		while (utf_search < new_address) {
			utf_prev = utf_search;
			advance_utf8 (& utf_search);
		}
		if (utf_search != new_address) {
			if (new_x >= x)
				new_address = utf_search;
			else
				new_address = utf_prev;
		}
	}
  }

/* Adjust shift_count if new_x lower than 0 or higher than XBREAK */
/* Allow adjustment also if new_x == 0 to enable left shift mark */
  if (new_x <= 0 || new_x >= XBREAK) {
	if (new_x > XBREAK || (new_x == XBREAK && * new_address != '\n'))
		shift = (new_x - XBREAK) / SHIFT_SIZE + 1;
	else {
		shift = new_x / SHIFT_SIZE;
		if (new_x % SHIFT_SIZE)
			shift --;
		if (new_x == 0 && line->shift_count != 0 && SHIFT_BEG != '\0')
			shift --;
	}

	if (shift != 0) {
		line->shift_count += shift;
		new_x = find_x (line, new_address);
		if (new_x == 0 && line->shift_count != 0 && SHIFT_BEG != '\0')
		{	line->shift_count --;
			new_x = find_x (line, new_address);
		}
		set_cursor (0, y);
		line_print (line);
		/* rel_x = new_x; */
	}
  }

/* Assign and position cursor */
  x = new_x;
  cur_text = new_address;
  cur_line = line;
  calc_line_no ();
  line_number = lineno;
  set_cursor_xy ();
}

void
move_y (ny)
  register int ny;
{
  move_it (old_x, NIL_PTR, ny);
}

void
move_to (nx, ny)
  register int nx;
  register int ny;
{
  old_x = x;
  move_it (nx, NIL_PTR, ny);
  old_x = x;
}

void
move_address (nadd, ny)
  register char * nadd;
  register int ny;
{
  old_x = x;
  move_it (0, nadd, ny);
  old_x = x;
}

void
move_address_w_o_RD (nadd, ny)
  register char * nadd;
  register int ny;
{
  old_x = x;
  move_it (0, nadd, ny);
  old_x = x;
}

/*
 * Initialize is called when a another file is edited. It free's the allocated 
 * space and sets modified back to False and fixes the header/tail pointer.
 */
void
initialize ()
{
  register LINE * line;
  register LINE * next_line;

/* Delete the whole list */
  for (line = header->next; line != tail; line = next_line) {
	next_line = line->next;
	free_space (line->text);
	free_header (line);
  }

/* header and tail should point to itself */
  line->next = line->prev = line;
  x = y = 0;
  rpipe = modified = False;
}

/*
 * Get_line reads one line from filedescriptor fd. If EOF is reached on fd,
 * get_line () returns ERRORS, else it returns the length of the string.
 */
char * get_l_err_u1;
char * get_l_err_u2;
char * get_l_err1;
char * get_l_err2;
char last_char = '\0';
long count_ascii;	/* count ASCII replacement candidate chars */
long count_nonascii;	/* count non-ASCII replacement corresponding chars */
long count_good_utf;	/* count good UTF-8 sequences */
long count_bad_utf;	/* count bad UTF-8 sequences */
long count_read_op;	/* count 1st read operation by get_line */
long count_lineend_LF;	/* count Unix lines */
long count_lineend_CRLF; /* count MSDOS lines */
FLAG UTF_BOM;		/* UTF-8 BOM found at beginning of file? */
FLAG Unicode_BOM;	/* Unicode BOM found at beginning of file? */
FLAG Unicode_little_endian;	/* Unicode words reversed? */

void
reset_get_line ()
{
  get_l_err1 = NIL_PTR;
  get_l_err2 = NIL_PTR;
  get_l_err_u1 = NIL_PTR;
  get_l_err_u2 = NIL_PTR;
  count_ascii = 0;
  count_nonascii = 0;
  count_good_utf = 0;
  count_bad_utf = 0;
  count_read_op = 0;
  count_lineend_LF = 0;
  count_lineend_CRLF = 0;
  UTF_BOM = False;
  Unicode_BOM = False;
}

void
show_get_l_error (get_l_err)
  char * get_l_err;
{
  if (get_l_err != NIL_PTR) {
	ring_bell ();
	error (get_l_err, " -- type a blank");
	while (readchar () != ' ' && quit == False) {
		ring_bell ();
		flush ();
	}
  }
}
void
show_get_l_errors ()
{
  show_get_l_error (get_l_err1);
  show_get_l_error (get_l_err2);
  show_get_l_error (get_l_err_u1);
  show_get_l_error (get_l_err_u2);
  clear_status ();
}

static char * UTF16buf = NIL_PTR;
static char * next_byte;
static char * fini_byte;
static unsigned int read_bytes;
static unsigned int surrogate = 0;

int
UTF_transform (UTF8buf, maxbufl, next_byte_poi, fini_byte)
  char * UTF8buf;
  int maxbufl;
  character * * next_byte_poi;
  character * fini_byte;
{
  register char * ptr = UTF8buf;
  int read_chars = 0;
  unsigned long unichar;

  while (read_chars + 4 < maxbufl && * next_byte_poi < fini_byte) {
	unichar = * * next_byte_poi;
	(* next_byte_poi) ++;
	if (* next_byte_poi < fini_byte) {
		if (Unicode_little_endian == True) {
			unichar |= (* * next_byte_poi) << 8;
		} else {
			unichar = (unichar << 8) | (* * next_byte_poi);
		}
		(* next_byte_poi) ++;
	} else if (Unicode_little_endian == False) {
		unichar = 0;
	}

	if ((unichar & 0xFC00) == 0xD800) {
	/* high surrogates */
		surrogate = (unichar - 0xD7C0) << 10;
	} else if ((unichar & 0xFC00) == 0xDC00) {
	/* low surrogates */
		unichar = surrogate | (unichar & 0x03FF);
		surrogate = 0;
		* ptr ++ = 0xF0 | (unichar >> 18);
		* ptr ++ = 0x80 | ((unichar >> 12) & 0x3F);
		* ptr ++ = 0x80 | ((unichar >> 6) & 0x3F);
		* ptr ++ = 0x80 | (unichar & 0x3F);
		read_chars += 4;
	} else if (unichar < 0x80) {
		* ptr ++ = unichar;
		read_chars ++;
	} else if (unichar < 0x800) {
		* ptr ++ = 0xC0 | (unichar >> 6);
		* ptr ++ = 0x80 | (unichar & 0x3F);
		read_chars += 2;
	} else {
		* ptr ++ = 0xE0 | (unichar >> 12);
		* ptr ++ = 0x80 | ((unichar >> 6) & 0x3F);
		* ptr ++ = 0x80 | (unichar & 0x3F);
		read_chars += 3;
	}
  }

  return read_chars;
}

#define debug_readx
#ifdef debug_read
#define trace_read(params)	printf params
#else
#define trace_read(params)	
#endif

int
get_line (fd, buffer, len)
  int fd;
  register char buffer [MAX_CHARS];
  int * len;
{
  static char * last = NIL_PTR;
  static char * current = NIL_PTR;
  static int read_chars;
  register char * cur_pos = current;
  char * begin = buffer;
  char * fini = (char *) (buffer + MAX_CHARS - 2) /* leave space for '\n\0' */;
  register FLAG ignore1char;
  int ret = 0;
  char char_copied;
  int Ulineend_state = 0;

#define debug_overlongx
#ifdef debug_overlong
  fini = (char *) (buffer + 20) /* debug overlong line input */;
#endif

  /* read one line */
  do {	/* read one char */
	do {	/* read one effective (not to be filtered out) char */
	    ignore1char = False;
	    if (cur_pos == last) {
		if (Unicode_BOM == True) {
		    if (next_byte >= fini_byte) {
			do {
				interrupted = False;
				read_bytes = read (fd, UTF16buf, screen_BUFL);
				trace_read (("read %d\n", read_bytes));
			} while (interrupted == True ||
				 (read_chars == -1 && geterrno () == EINTR));
			if (read_bytes <= 0) {
				read_chars = read_bytes;
				break;
			}
			next_byte = UTF16buf;
			fini_byte = & UTF16buf [read_bytes];
		    }
		    read_chars = UTF_transform (screen, screen_BUFL, & next_byte, fini_byte);
		} else {
		    do {
			interrupted = False;
			read_chars = read (fd, screen, screen_BUFL);
			trace_read (("read %d\n", read_chars));
		    } while (interrupted == True ||
			     (read_chars == -1 && geterrno () == EINTR));
		    if (read_chars <= 0)
			break;
		}
		last = & screen [read_chars];
		cur_pos = screen;

		if (count_read_op == 0) {
			if (strncmp (screen, "﻿", 3) == 0) {
				/* UTF-8 BOM */
				UTF_BOM = True;
			} else if (strncmp (screen, "\376\377", 2) == 0
				   && auto_UTF_detect == True) {
				/* UTF-16 BOM */
				Unicode_BOM = True;
				Unicode_little_endian = False;
				UTF_BOM = True;
			} else if (strncmp (screen, "\377\376", 2) == 0
				   && auto_UTF_detect == True) {
				/* little-endian reversed UTF-16 BOM */
				Unicode_BOM = True;
				Unicode_little_endian = True;
				UTF_BOM = True;
			}
			if (Unicode_BOM == True) {
				get_l_err_u1 = "Converted UTF-16 input";

			/* move UTF-16 input to UTF-16 buffer */
				if (UTF16buf == NIL_PTR) {
					UTF16buf = alloc (screen_BUFL);
					if (UTF16buf == NIL_PTR) {
						get_l_err_u2 = "Cannot allocate memory for UTF-16 buffer";
						viewonly = True;
						modified = False;
						return ERRORS;
					}
				}
				memcpy (UTF16buf, screen, (unsigned int) read_chars);
				read_bytes = read_chars;
				next_byte = UTF16buf;
				fini_byte = & UTF16buf [read_bytes];

			/* transform to UTF-8 */
				read_chars = UTF_transform (screen, screen_BUFL, & next_byte, fini_byte);
				last = & screen [read_chars];

				modified = True;
			}
			count_read_op = 1;
		}
	    }

	/* transdeutsch auto detection */
	    switch (* cur_pos) {
		case '{':
		case '|':
		case '}':
		case '[':
		case '\\':
		case ']':
		case '~':
			count_ascii ++;
			break;
		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
			count_nonascii ++;
			break;
	    }

	/* UTF-8 auto-detection */
	    if ((* cur_pos & 0xC0) == 0x80) {
		if ((last_char & 0xC0) == 0xC0) {
			count_good_utf ++;
		} else if ((last_char & 0x80) == 0x00) {
			count_bad_utf ++;
		}
	    } else if ((last_char & 0xC0) == 0xC0) {
		count_bad_utf ++;
	    }

#ifdef msdos_with_auto_crlf
	    if (* cur_pos == '\n' && last_char != '\r') {
		set_modified ();
	    }
#endif

	/* line-end transformation */
	    if (RET_opt == 'R' && * cur_pos == '\n' && last_char == '\r') {
		/* handling DOS line-ends in Mac mode */
		last_char = * cur_pos;
		cur_pos ++;
		ignore1char = True;
		set_modified ();
	    }
	    else
		last_char = * cur_pos;

	    if (* cur_pos == '\r' ) {
		if (RET_opt == 'R') {
		/* convert Mac line-ends */
			* cur_pos = '\n';
			set_modified ();
		}
		else if (RET_opt == 'r') {
		/* convert DOS line-ends */
			cur_pos ++;
			if (* cur_pos == '\n') {
				ignore1char = True;
#ifndef pc
				set_modified ();
#endif
			} else {
				cur_pos --;	/* leave '^M' within line */
			}
		}
	    }

	/* NUL character handling */
#define new_nul_handling
	    if (* cur_pos == '\0') {
#ifdef new_nul_handling
		* cur_pos = '\n';
		ret = NUL_LINE;
#else
		get_l_err1 = "File contains NUL characters (^@) - changed to DEL characters";
		* cur_pos = '\177';
#endif
	    }

	} while (ignore1char == True);

	/* detect if no more lines available */
	if (cur_pos == last) {
		break;
	}

	/* handle if line buffer full */
	if (buffer > fini - 6 && * cur_pos != '\n') {
	    /* try not to split within a UTF-8 sequence */
	    if (Chinese == True
		? buffer == fini
		  || (buffer == fini - 1 && ! inChinese (begin, buffer))
		: buffer == fini	/* last chance to split! */
		  || (* cur_pos & 0xC0) != 0x80
		)
	    {
		get_l_err2 = "Line(s) too long - split";
		* buffer ++ = '\n';

		/* indicate split line */
		if (linesplitmarker == '\0') {
			/* default: new line splitting */
			ret = SPLIT_LINE;
		} else {
			cur_pos --;
			* cur_pos = linesplitmarker;
			/* To provide safe space for this operation, the 
			   screen array (which cur_pos points into) is 
			   preceded by the char variable linesplitmarker, 
			   which also holds the marker character */
		}

		break;
	    }
	}

	/* copy char from input buffer into line buffer */
	char_copied = * buffer ++ = * cur_pos ++;
	if (utf8_lineends == True && Ulineend_state == 0 && char_copied == '')
		Ulineend_state = 1;
	else if (Ulineend_state > 0) {
		if (Ulineend_state == 1 && char_copied == '')
			Ulineend_state = 2;
		else if (Ulineend_state == 2 && char_copied == '') {
			Ulineend_state = 3;
			/* LS   detected */
			char_copied = '\n';
		} else if (Ulineend_state == 2 && char_copied == '') {
			Ulineend_state = 3;
			/* PS   detected */
			char_copied = '\n';
		} else
			Ulineend_state = 0;
	}
  } while (char_copied != '\n' &&
		(char_copied != '\r' || RET_opt != 'M')
	  );

  current = cur_pos;
  if (read_chars <= 0) {
	if (buffer == begin)
		return ERRORS;
	if (* (buffer - 1) != '\n' &&
		(* (buffer - 1) != '\r' || RET_opt != 'M')
	) {
		if (loading == True) {
		/* Add '\n' to last line of file, for internal handling */
			* buffer ++ = '\n';
			* buffer = '\0';
		/* but inform load_file of missing line end */
			ret = NO_LINE;
		} else {
			* buffer = '\0';
			ret = NO_LINE;
		}
	}
  } else if (Chinese == True) {
	if (inChinese (begin, buffer - 1)) {
		buffer --;
		* buffer ++ = ' ';
		* buffer ++ = '\n';
		get_l_err1 = "Corrupted 2-byte character (missing half at line end)";
		modified = True;
	}
  }

  * buffer = '\0';
  * len = (int) (buffer - begin);
  return ret;
}

/*
 * get_open_pos and save_open_pos look up and save the current file 
 * position in @mined.mar . For save_open_pos, line_number must be up-to-date
 */
void
get_open_pos (fn)
  char * fn;
{
  int mark_fd = open (mark_file, O_RDONLY | O_BINARY, 0);
  char * post;
  FLAG modif = modified;
  int dumlen;
  int scanned;
  int v4, v5, v6, v7 = -1;

  open_linum = 0;
  open_col = 0;
  if (mark_fd >= 0) {
	while ((get_line (mark_fd, text_buffer, & dumlen)) != ERRORS) {
	    if (strncmp (fn, text_buffer, strlen (fn)) == 0) {
		post = text_buffer + strlen (fn);
		if (* post == ' ') {
			lines_per_page = 0;
			scanned = sscanf (post, "%d%d%d%d%d%d%d", 
			   & open_linum, & open_col, & lines_per_page, 
			   & v4, & v5, & v6, & v7);
			if (scanned >= 6 && v4 >= 0) {
				JUSlevel = 1;
				first_left_margin = v4;
				next_left_margin = v5;
				right_margin = v6;
				set_quote_type (v7);
			} else {
				JUSlevel = 0;
				set_quote_type (v5);
			}
		}
	    }
	}
	(void) close (mark_fd);
	clear_buffer ();
  }
  modified = modif; /* don't let the line number file affect the modified flag */
}

int
get_cur_col ()
{
  int cur_column = 0;
  char * cpoi = cur_line->text;

  while (* cpoi != '\0' && cpoi != cur_text) {
	advance_char_col (& cpoi, & cur_column);
  }
  return cur_column;
}

void
save_open_pos (fn, force)
  char * fn;
  int force;
{
  int mark_fd;
  char marktext [maxLINE_LEN];
  int cur_column;

  if (fn [0] != '\0') {
    if (force > 0)
	mark_fd = open (mark_file, O_WRONLY | O_APPEND | O_CREAT, fprot);
    else
	mark_fd = open (mark_file, O_WRONLY | O_APPEND, fprot);
    if (mark_fd >= 0) {
	cur_column = get_cur_col ();
	if (JUSlevel > 0)
	    build_string (marktext, "%s %d %d %d %d %d %d %d\n",
			  fn, line_number, cur_column, lines_per_page, 
			  first_left_margin, next_left_margin, right_margin, 
			  quote_type);
	else
	    build_string (marktext, "%s %d %d %d -3 %d\n",
			  fn, line_number, cur_column, lines_per_page, 
			  quote_type);
	write (mark_fd, marktext, strlen (marktext));
	(void) close (mark_fd);
    }
  }
}

/*
 * extract_lineend_type extracts the type of line end from the string, 
 * removes it from the string and returns it
 */
lineend_type
extract_lineend_type (text)
  char * text;
{
  int length = length_of (text);
  char * end = text + length;
  end --;

  if (* end == '\n') {
	end --;
	if (length > 1 && * end == '\r') {
		* end = * (end + 1);
		* (end + 1) = '\0';
		count_lineend_CRLF ++;	/* count MSDOS lines */
		return lineend_CRLF;
	} else {
		count_lineend_LF ++;	/* count Unix lines */
		return lineend_LF;
	}
  } else if (* end == '\r') {
	* end = '\n';
	return lineend_CR;
  } else if (utf8_lineends == True && length >= 3 && strncmp (end - 2, " ", 3) == 0) {
	end --;
	end --;
	* end ++ = '\n';
	* end = '\0';
	return lineend_LS;	/* Unicode line separator 2028 */
  } else if (utf8_lineends == True && length >= 3 && strncmp (end - 2, " ", 3) == 0) {
	end --;
	end --;
	* end ++ = '\n';
	* end = '\0';
	return lineend_PS;	/* Unicode paragraph separator 2029 */
  } else {
	return lineend_NONE;
  }
}

/*
 * Load_file loads the file with given name or the input pipe into memory.
 * If the file couldn't be opened, just an empty line is installed.
 * Buffer pointers are initialized.
 * The current position is set to the last saved position.
 *
 * Load_file_position loads and positions as follows:
	if open_linum > 0, the given position is used
	if open_linum == 0, the last saved position is used
	if open_linum < 0, no position is set (stays at beginning)
 */
void
load_file_position (file, with_display, to_open_linum, to_open_col)
  char * file;
  FLAG with_display;
  int to_open_linum;
  int to_open_col;
{
  register LINE * line = header;
  int len;
  int ret;
  long nr_of_bytes = 0L;
  long nr_of_chars = 0L;
  long nr_of_utfchars = 0L;
  int fd = -1;		/* Filedescriptor for file */
  FLAG locked = False;	/* lockf performed ? */
  lineend_type new_return_type;
  FLAG utf8_text_context;	/* save state of utf8_text */

  total_lines = 0;	/* Zero lines to start with */

  open_linum = to_open_linum;
  open_col = to_open_col;

  clearscreen ();
  flush ();
  overwriteOK = False;
  writable = True;
  if (file == NIL_PTR) {
	if (rpipe == False)
		status_msg ("No file");
	else {
		fd = 0;
		file = "standard input";
	}
	file_name [0] = '\0';
  }
  else {
	copy_string (file_name, file);	/* Save file name */
	status_line ("Accessing ", file);
	if (access (file, 0 /* F_OK */) < 0) {	/* Cannot access file */
		status_line ("New file ", file);
		overwriteOK = False;
	}
	else if ((fd = open (file, O_RDWR | O_BINARY, 0)) >= 0) {
		overwriteOK = True;
		writable = True;
/*
		if (lockf (fd, F_LOCK, 0) < 0) {
			status_line ("No lock for ", file);
		}
		else	locked = True;
*/
		if (open_linum == 0)
			get_open_pos (file_name);
	}
	else if ((fd = open (file, O_RDONLY | O_BINARY, 0)) >= 0) {
		overwriteOK = True;
		writable = False;
		if (open_linum == 0)
			get_open_pos (file_name);
	} else {
		error ("Cannot open: " /*, file */, serror ());
	}
  }
  set_file_type_flags ();

/* Read file */
  loading = True;		/* Loading file, so set flag */
  reset_get_line ();

  if (fd >= 0) {
	status_line ("Reading ", file);
	utf8_text_context = utf8_text;
	while (line != NIL_LINE
		&& (ret = get_line (fd, text_buffer, & len)) != ERRORS)
	{
		if (ret == NO_LINE || ret == SPLIT_LINE) {
			new_return_type = lineend_NONE;
			len = length_of (text_buffer);
		} else if (ret == NUL_LINE) {
			new_return_type = lineend_NUL;
			len = length_of (text_buffer);
		} else {
			new_return_type = extract_lineend_type (text_buffer);
		}
		line = line_insert (line, text_buffer, len, new_return_type);

		nr_of_bytes += (long) len;
		utf8_text = False;
		nr_of_chars += char_count (text_buffer);
		if (Chinese == False) {
			utf8_text = True;
			nr_of_utfchars += char_count (text_buffer);
		}
		if (new_return_type == lineend_NONE) {
			nr_of_chars --;
			nr_of_utfchars --;
			nr_of_bytes --;
		}
	}
	utf8_text = utf8_text_context;

	if (total_lines == 0 && line != NIL_LINE)   /* The file was empty! */
		line = line_insert (line, "\n", 1, lineend_NONE);
	clear_buffer (); /* Clear output buffer: out_count = 0; */
	cur_line = header->next;
	line_number = 0;
	if (locked == False)
		(void) close (fd);		/* Close file */

	if (auto_transdeutsch == True) {
		if (count_nonascii > count_ascii)
			transdeutsch = True;
		else	transdeutsch = False;
	}

	if (count_lineend_CRLF > count_lineend_LF)
		default_lineend = lineend_CRLF;
  }
  else				/* Just install a "\n" */
	line = line_insert (line, "\n", 1, default_lineend);

  /* Unicode encoding detection */
  if (auto_UTF_detect == True) {
	if (UTF_BOM == True || 
		utf8_screen == True ? count_good_utf >= count_bad_utf
				: count_good_utf > count_bad_utf)
	{
		nr_of_chars = nr_of_utfchars;
		utf8_text = True;
	} else {
		utf8_text = False;
	}
  }

  if (fd >= 0)
	if (line != NIL_LINE) {
	   show_get_l_errors ();
	   fstatus ("Read", nr_of_bytes, nr_of_chars);
	}

  if (line == NIL_LINE) {
	sleep (2) /* give time to read allocation error msg */;
	viewonly = True;
	modified = False;
  }

  reset (header->next, 0);	/* Initialize pointers */
  move_to (0, 0);
  loading = False;		/* Stop loading, reset flag */

  if (open_linum > 0) {
	open_line = proceed (header->next, open_linum - 1);
	move_to (0, find_y_w_o_RD (open_line));
	{	/* ??? */
		int cur_column = 0;
		char * cpoi = cur_line->text;
		while (* cpoi != '\n' && cur_column < open_col) {
			advance_char_col (& cpoi, & cur_column);
		}
		move_address (cpoi, y);
		open_col = 0;
	}
	if (with_display == True)
		display (0, top_line, last_y, 0);
	fstatus ("Read", nr_of_bytes, nr_of_chars);
	while (open_col > 0) {
		MRT ();
		open_col --;
	}
	Markn (-1);
  } else if (with_display == True) {
	display (0, top_line, last_y, 0);
	move_to (0, 0);
  }

#ifdef unix
  RD_window_title ();
#endif
}

void
load_file (file, with_display)
  char * file;
  FLAG with_display;
{
  load_file_position (file, with_display, 0, 0);
}


#ifdef msdos
# ifdef __TURBOC__
#include "mineddir.h"
# else
#include "dir.h"
# endif
struct ffblk ffblock;
FLAG ffblockvalid = False;
char wildfile [maxLINE_LEN];
char drive [9], dir [maxLINE_LEN], name [13], ext [5];
#endif

void
load_wild_file (file, with_display)
  char * file;
  FLAG with_display;
{
#ifdef msdos
  char dumdrive [9];
  char dumdir [maxLINE_LEN];
#endif

#ifdef msdos
  ffblockvalid = False;
#endif
  if (file == NIL_PTR) load_file (file, with_display);
  else {
#ifdef msdos
/*	(void) fnsplit (fnamv [fnami], drive, dir, name, ext);
	if (findfirst (fnamv [fnami], & ffblock, 0) == 0) {
*/
	(void) fnsplit (file, drive, dir, name, ext);
	if (findfirst (file, & ffblock, 0) == 0) {
		ffblockvalid = True;
		(void) fnsplit (ffblock.ff_name, dumdrive, dumdir, name, ext);
		fnmerge (wildfile, drive, dir, name, ext);
		load_file (wildfile, with_display);
	}
	else
#endif
	load_file (file, with_display);
  }
}

/*-------------------------------------------------------------------------*/

/*
 * Ask user if named file should be overwritten.
 */
FLAG
checkoverwrite (name)
char * name;
{
  character c;

  status_line ("Checking ", name);
  if (access (name, 0 /* F_OK */) < 0)	/* Cannot access file */
	return True;	/* thus no danger of unwanted damage */

  status_line (name [0] ? name : "[buffer]" ,
			": OK to overwrite? (y/n)");
  c = promptyn ();
  clear_status ();
  if (c == 'y')
	return True;
  else if (c == 'n')
	return False;
  else {
/*	quit = False;	abort character has been given */
	return False;
  }
}

/*
 * Attach new file name to buffer
 */
void
NN ()
{
  char file [maxLINE_LEN];	/* Buffer for new file name */

  if (restricted == True)
	{restrictederr (); return;}
  if (get_file ("Enter new file name:", file) == ERRORS)
	return;

  overwriteOK = False;
  writable = True;
  copy_string (file_name, file);	/* Save new file name */
#ifdef unix
  if (modified == True) {
	RD_window_title ();
  }
#endif
  set_modified ();	/* cf. CHDI command */
  clear_status ();
}


/*
 * writechar does a buffered output to file.
 */
int
writechar (fd, c)
  int fd;
  char c;
{
  screen [out_count ++] = c;
  if (out_count == screen_BUFL) {
	return flush_buffer (fd);
  }
  return FINE;
}

int
write_lineend (fd, return_type)
  register int fd;
  lineend_type return_type;
{
  switch (return_type) {
    case lineend_NONE:	return 0;
    case lineend_NUL:	if (writechar (fd, '\0') == ERRORS)
				return ERRORS;
			return 1;
    case lineend_LF:	if (writechar (fd, '\n') == ERRORS)
				return ERRORS;
			return 1;
    case lineend_CRLF:	if (writechar (fd, '\r') == ERRORS)
				return ERRORS;
			if (writechar (fd, '\n') == ERRORS)
				return ERRORS;
			return 2;
    case lineend_CR:	if (writechar (fd, '\r') == ERRORS)
				return ERRORS;
			return 1;
    case lineend_LS:	/* Unicode line separator 2028:   */
			if (writechar (fd, '\342') == ERRORS)
				return ERRORS;
			if (writechar (fd, '\200') == ERRORS)
				return ERRORS;
			if (writechar (fd, '\250') == ERRORS)
				return ERRORS;
			return 3;
    case lineend_PS:	/* Unicode paragraph separator 2029:   */
			if (writechar (fd, '\342') == ERRORS)
				return ERRORS;
			if (writechar (fd, '\200') == ERRORS)
				return ERRORS;
			if (writechar (fd, '\251') == ERRORS)
				return ERRORS;
			return 3;
    default:		if (writechar (fd, '\n') == ERRORS)
				return ERRORS;
			return 1;
  }
}

/*
 * Writestring writes the given string on the given filedescriptor.
 * (buffered via writechar via misused screen buffer!)
 */
int
writestring (fd, text, return_type)
  register int fd;
  register char * text;
  lineend_type return_type;
{
  char c;
  int len;
  int ccount = 0;

  while (* text != '\0') {
	c = * text ++;
	if (c == '\n') {
	/* handle different line ends */
		len = write_lineend (fd, return_type);
		if (len == ERRORS)
			return ERRORS;
		ccount += len;
	} else {
		if (writechar (fd, c) == ERRORS)
			return ERRORS;
		ccount ++;
	}
  }
  return ccount;
}

/* Call graph for writing functions:
	panic --\------> panicwrite --------------------------\
		 > QUED ----------\			       \
	ESC q --/		   > ask_save --\		> write_file
	ESC e ---> EDIT --> edit_file -/	 \	       /
	ESC v ---> VIEW -/			  \	      /
	ESC w -------------------> WT -------------> write_text
	ESC W -------------------> WTU -----------/
	ESC z -------------------> SUSP ---------/
	ESC ESC -> EXED ---------> EXFILE ------/
			\--------> EXMINED ----/
	ESC t -------------------> Stag ------/
*/
long write_count;	/* number of bytes written */
long chars_written;	/* number of chars written */

/*
 * Write text in memory to file.
 */
void
write_file (fd)
  int fd;
{
  register LINE * line;
  int ret;

  write_count = 0L;
  chars_written = 0L;
  clear_buffer (); /* out_count = 0; */
  for (line = header->next; line != tail; line = line->next) {
	if (line->shift_count & DUMMY) {
		if (line->next == tail && line->text [0] == '\n')
			continue;
	}
	ret = writestring (fd, line->text, line->return_type);
	if (ret == ERRORS) {
		write_count = -1L;
		chars_written = -1L;
		break;
	}
	write_count += (long) ret;
	chars_written += (long) char_count (line->text);
	if (line->return_type == lineend_NONE) {
		chars_written --;
	}
  }

  if (write_count > 0L && flush_buffer (fd) == ERRORS) {
	write_count = -1L;
	chars_written = -1L;
  }

  ret = close (fd);
  if (ret == -1) {
	error ("Write failed (File incomplete): ", serror ());
	write_count = -1L;
	chars_written = -1L;
  }
}

int
write_text (forceflag)
  FLAG forceflag;
{
  char file [maxLINE_LEN];	/* Buffer for new file name */
  int fd;			/* Filedescriptor of file */
  int ret;

  if (wpipe == True) {
    fd = STD_OUT;
    status_line ("Writing ", "to standard output");
    wpipe = False; /* no further write to same stream possible */
  }
  else {
    if (forceflag == False && modified == False) {
	if (file_name [0] != '\0') {
		fstatus ("(Write not necessary)", -1L, -1L);
		save_open_pos (file_name, hop_flag);
	}
	else	status_msg ("Write not necessary.");
	return FINE;
    }

    /* Check if file_name is valid and if file can be written */
    if (file_name [0] == '\0' || writable == False) {
	overwriteOK = False;
	if ((ret = get_file ("Saving edited text - enter file name:", file))
		!= FINE)
	{
		return ret;
	}
	copy_string (file_name, file);		/* Save file name */
    }
    if (overwriteOK == False) {
	if (checkoverwrite (file_name) == True) {
		overwriteOK = True;
	} else {
		if (quit == False)
			writable = False;
		return ERRORS;
	}
    }
    status_line ("Opening ", file_name);
/*    fd = creat (file_name, fprot | xprot);	*/
    fd = open (file_name, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, fprot | xprot);
    if (fd < 0) {	/* Empty file */
	error ("Cannot create or write: " /*, file_name */, serror ());
	writable = False;
	return ERRORS;
    } else {
	writable = True;
    }

    status_line ("Writing ", file_name);
  }

  write_file (fd);

  if (write_count == -1L)
	return ERRORS;

  modified = False;
#ifdef unix
  RD_window_title ();
#endif
  rpipe = False;	/* File name is now assigned */

/* Display how many chars (and lines) were written */
  fstatus ("Wrote", write_count, chars_written);
/*  fstatus ("Wrote", -1L); */
  save_open_pos (file_name, hop_flag);
  return FINE;
}

void
WT ()
{
  (void) write_text (False);
}

void
WTU ()
{
  (void) write_text (True);
}

/*
 * Save current file pos
 */
void
SAVPOS ()
{
  fstatus ("remembering file position", -1L, -1L);
  save_open_pos (file_name, 1);
}

/*
 * Ask the user if he wants to save the file or not.
 */
int
ask_save ()
{
  register character c;

  status_line (file_name [0] ? file_name : "[buffer]" ,
			" has been modified. Save? (y/n)");
  c = promptyn ();
  clear_status ();
  if (c == 'y') {
	return write_text (False);
  } else if (c == 'n') {
	return FINE;
  } else {
	quit = False;	/* abort character has been given */
	return ERRORS;
  }
}

int
panicwrite ()
{
  int fd;
/*  fd = creat (panic_file, fprot);	*/
  fd = open (panic_file, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, bufprot);
  write_file (fd);
  if (write_count == -1L)
	return ERRORS;
  else	return FINE;
}

/*
 * Edit/view another file. If the current file has been modified, 
 * ask whether the user wants to save it.
 * (We could allow to switch between edit and view mode without changing 
 * the file, but we would have to consider carefully the relationship 
 * between viewonly and modified.)
 */
void
edit_file (prompt, vomode)
  char * prompt;
  FLAG vomode;
{
  char new_file [maxLINE_LEN];	/* Buffer to hold new file name */

  if (modified == True && viewonly == False && ask_save () != FINE)
	return;

/* Get new file name */
  if (get_file (prompt, new_file) == ERRORS)
	return;

  viewonly = vomode;

/* Free old linked list, initialize global variables and load new file */
  initialize ();
  clearscreen ();
  load_wild_file (new_file [0] == '\0' ? NIL_PTR : new_file, True);
}

void
edit_this_file (new_file, vomode)
  char * new_file;
  FLAG vomode;
{
  if (modified == True && viewonly == False && ask_save () != FINE)
	return;

  viewonly = vomode;

/* Free old linked list, initialize global variables and load new file */
  initialize ();
  clearscreen ();
  load_wild_file (new_file [0] == '\0' ? NIL_PTR : new_file, True);
}

void
EDIT ()
{
  if (restricted == True)
	{restrictederr (); return;}
  edit_file ("Edit file:", False);
}

void
VIEW ()
{
  if (restricted == True)
	{restrictederr (); return;}
  edit_file ("View file:", True);
}

void
EDITmode ()
{
  viewonly = False;
  FSTAT ();
  flags_changed = True;
}

void
VIEWmode ()
{
  if (modified == False) {
	viewonly = True;
	FSTAT ();
	flags_changed = True;
  } else error ("Cannot view only - already modified", NIL_PTR);
}

extern FLAG viewing_help;
static void end_view_help ();

void
quit_mined ()
{
  if (viewing_help == True) {
	end_view_help ();
	return;
  }

  delete_yank_file ();
  set_cursor (0, YMAX);
  putchar ('\n');
#ifdef unix
  clear_window_title ();
#endif
  raw_mode (OFF);
  exit (0);
}

void
edit_nth_file (n)
  int n;
{
  int number;
  int index;

  if (modified == True && viewonly == False && ask_save () != FINE)
	return;

  if (n == -1) {
	status_msg ("Edit which file (enter number) or # to reload current file");
	index = readchar ();
	if (quit == True) return;
	if (index == '#')
		n = fnami;
	else {
		index = get_number ("Edit which file (enter number) ...", index, & number);
		if (index == ERRORS) return;
		n = number - 1 + fnami_min;
	}
  }
  if (n < fnami_min) n = fnami_min;
  if (n > fnami_max) n = fnami_max;

/* Free old linked list, initialize global variables and load new file */
  initialize ();
  clearscreen ();

  fnami = n;
  if (fnami < fnami_min) load_wild_file (NIL_PTR, True);
  else	load_wild_file (fnamv [fnami], True);
}

void
nextfile (exitiflast)
	FLAG exitiflast;
{
#ifdef msdos
  char dumdrive [9];
  char dumdir [maxLINE_LEN];
#endif

  if (hop_flag > 0) edit_nth_file (fnami_max);
  else
#ifdef msdos
	if (modified == True && viewonly == False && ask_save () != FINE)
		return;

	if ((ffblockvalid == True) && (findnext (& ffblock)) == 0) {
	/* Free old linked list, initialize global variables and load new file */
	  initialize ();
	  clearscreen ();
	  (void) fnsplit (ffblock.ff_name, dumdrive, dumdir, name, ext);
	  fnmerge (wildfile, drive, dir, name, ext);
	  load_file (wildfile, True);
	  return;
	}
	ffblockvalid = False;
#endif
	if (fnami >= fnami_max || fnami == 0) {
		if (exitiflast == True) quit_mined ();
		else error ("Already at last file", NIL_PTR);
	}
	else edit_nth_file (fnami + 1);
}

void
NXTFILE ()
{
  nextfile (False);
}

void
PRVFILE ()
{
  if (hop_flag > 0) edit_nth_file (fnami_min);
  else	if (fnami <= fnami_min) {
#ifdef msdos
		error ("Already at first file name", NIL_PTR);
#else
		error ("Already at first file", NIL_PTR);
#endif
		return;
	}
	else edit_nth_file (fnami - 1);
}

void
NTHFILE ()
{
  edit_nth_file (-1);
}

/*
 * Leave editor. If the file has changed, ask if the user wants to save it.
 */
void
QUED ()
{
  if (modified == True && viewonly == False && ask_save () != FINE)
	return;

  quit_mined ();
}

/*
 * Exit editing current file. If the file has changed, save it.
 * Edit next file if there is one.
 */
void
EXFILE ()
{
  if (modified == True) {
	if (write_text (False) != FINE) {
		return;
	}
  }

  if (hop_flag == 0) nextfile (True);
  else quit_mined ();
}

/*
 * Exit editor. If the file has changed, save it.
 */
void
EXMINED ()
{
  if (modified == True) {
	if (write_text (False) != FINE) {
		return;
	}
  }

  quit_mined ();
}

/*
 * Exit editing current file. Exit editor if multiexit flag set.
 */
void
EXED ()
{
  if (multiexit == True) {
	EXFILE ();
  } else {
	EXMINED ();
  }
}


int
get_tagline (idf, filename, search)
  char * idf;
  char * filename;
  char * search;
{
  int tags_fd = open ("tags", O_RDONLY | O_BINARY, 0);
  FLAG modif = modified;
  unsigned int len = strlen (idf);
  int dumlen;
  char * poi;
  char * outpoi;
  FLAG found;

  flush ();	/* clear the shared screen/get_line buffer! */

  if (tags_fd >= 0) {
	found = False;
	while (found == False && (get_line (tags_fd, text_buffer, & dumlen)) != ERRORS) {
	    if (strncmp (idf, text_buffer, len) == 0 && text_buffer [len] == '\t') {
		found = True;
		poi = text_buffer + len + 1;

		outpoi = filename;
		while (* poi != '\0' && * poi != '\t') {
			* outpoi ++ = * poi ++;
		}
		* outpoi = '\0';

		outpoi = search;
		poi ++;
		if (* poi == '/') {
			poi ++;
		}
		while (* poi != '\0' && * poi != '\t' && * poi != ';' && * poi != '/') {
			if (* poi == '[' || * poi == ']') {
				* outpoi ++ = '\\';
			}
			* outpoi ++ = * poi ++;
		}
		* outpoi = '\0';
	    }
	}
	(void) close (tags_fd);
	clear_buffer ();
	modified = modif; /* don't let the tags file affect the modified flag */
	if (found == False) {
		error ("Identifier not found in tags file", NIL_PTR);
		return ERRORS;
	} else {
		return FINE;
	}
  } else {
	error ("No tags file present; use the ctags command", NIL_PTR);
	return ERRORS;
  }
}

/*
 * Stag () opens file and moves to idf, using tags file
 */
void
Stag ()
{
  char idf_buf [MAX_CHARS];	/* identifier to search for */
  char new_file [maxLINE_LEN];	/* Buffer to hold new file name */
  char search [MAX_CHARS];	/* search expression */
  int lineno;

  if (hop_flag > 0) {
	if (get_string ("Enter identifier (to locate definition):", idf_buf, True, "") != FINE)
		return;
  } else {
	if (get_idf (idf_buf, cur_text, cur_line->text) == ERRORS)
		return;
  }

  if (get_tagline (idf_buf, new_file, search) == ERRORS) {
	return;
  }

  if (strcmp (new_file, file_name) != 0) {
/*
	if (modified == True && viewonly == False && ask_save () != FINE) {
		return;
	}
*/
	if (modified == True) {
		if (write_text (False) == ERRORS) {
			return;
		}
	}

	/* Free old linked list, initialize global variables and load new file */
	initialize ();
	clearscreen ();
	load_file (new_file, True);
  }

  if (* search >= '0' && * search <= '9') {
	make_number (& lineno, search);
	goline (lineno);
  } else {
	search_for (search, FORWARD);
  }
}


/*
 * Count_chars () count the number of chars that the line would occupy on the
 * screen. Counting starts at the real x-coordinate of the line.
 * Was only called by delete_text () in a buggy part.
 * !!! Would have to be fixed for Chinese/UTF8 if re-activated!
 */
#ifdef UNUSED
int
count_chars (line)
  LINE * line;
{
  register int cnt = get_shift (line->shift_count) * - SHIFT_SIZE;
  register char * textp = line->text;

/* Find begin of line on screen */
  while (cnt < 0) {
	if (is_tab (* textp ++))
		cnt = tab (cnt);
	else
		cnt ++;
  }

/* Count number of chars left */
  cnt = 0;
  while (* textp != '\n') {
	if (is_tab (* textp ++))
		 cnt = tab (cnt);
	else
		cnt ++;
  }
  return cnt;
}
#endif /* UNUSED */


/*======================================================================*\
|*			Web marker insertion				*|
\*======================================================================*/

/*
   Strip string from first blank.
 */
void
strip (s)
  char * s;
{
  while (* s != '\0' && * s != ' ') {
	s ++;
  }
  * s = '\0';
}

void
embed_HTML ()
{
  char marker [maxLINE_LEN];
  char tag [maxLINE_LEN];

  if (viewonly == True) {
	viewonlyerr ();
	return;
  }
  if (get_string_nokeymap ("Embed text in HTML marker:", marker, True, "") != FINE) {
	return;
  }

  yank_HTML (DELETE);
  S ('<');
  S ('/');
  S ('>');
  MLF ();
  strcpy (tag, marker);
  strip (tag);
  if (insert (cur_line, cur_text, tag) == ERRORS) {
	return;
  }
  MLF ();
  MLF ();
  paste_HTML ();
  S ('<');
  S ('>');
  MLF ();
  if ((marker [0] == 'A' || marker [0] == 'a') && marker [1] == '\0') {
	S (marker [0]);
	S (' ');
	S ('h');
	S ('r');
	S ('e');
	S ('f');
	S ('=');
  } else {
	(void) insert (cur_line, cur_text, marker);
  }
  RD ();
}

static char HTMLmarker [maxLINE_LEN];
static FLAG HTMLmarking = False;

void
HTML ()
{
  char * htmlpoi;

  if (viewonly == True) {
	viewonlyerr ();
	return;
  }

  if (hop_flag > 0) {
	hop_flag = 0;
	embed_HTML ();
  } else {
	if (HTMLmarking == False) {
		if (get_string_nokeymap ("Begin HTML marker:", HTMLmarker, True, "") != FINE)
			return;
		clear_status ();
		S ('<');
		S ('>');
		MLF ();
		if (insert (cur_line, cur_text, HTMLmarker) == ERRORS) {
			return;
		}
		HTMLmarking = True;
	} else {
		S ('<');
		S ('/');
		S ('>');
		MLF ();
		strip (HTMLmarker);
		if (insert (cur_line, cur_text, HTMLmarker) == ERRORS) {
			return;
		}
		HTMLmarking = False;
	}
/*
	set_cursor (0, y);
	line_print (cur_line);
*/
	RD ();
	if (* cur_text != '>') {
		htmlpoi = cur_text;
		do {
			advance_char (& htmlpoi);
		} while (* htmlpoi != '>' && * htmlpoi != '\0');
		if (* htmlpoi == '>') {
			htmlpoi ++;
		}
		move_address (htmlpoi, y);
	}
  }
}


/*======================================================================*\
|*			Line wrap around				*|
\*======================================================================*/

#define dont_debug_jc
#ifdef debug_jc
#define trace_jc(params)	printf params
#else
#define trace_jc(params)	
#endif

#define dont_debug_justi
#ifdef debug_justi
#define trace_justi(params)	printf params
#else
#define trace_justi(params)	
#endif

/*
 * Return string without '\n' for tracing
 */
static char tracebuf [maxLINE_LEN];
char *
trcstr (text, poi)
  char * text;
  char * poi;
{
	char * p = text;
	char * outp = tracebuf;

	while (* p != '\0') {
		if (p == poi) {
			* outp ++ = '^';
		}
		if (p == cur_text) {
			* outp ++ = '!';
		}
		if (* p != '\n') {
			* outp ++ = * p;
		}
		p ++;
	}
	* outp = '\0';
	return tracebuf;
}


/*
 * determine if the next line contains visible text
 */
int
nonblank_line_follows ()
{
  LINE * next_line = cur_line->next;
  char * next_char;

  if (next_line == NIL_LINE) return 0;

  next_char = next_line->text;
  if (next_char == NIL_PTR) return 0;

  while (white_space (* next_char)) {
	next_char ++;
  }
  return * next_char != '\n';
}

/*
 * JUS justifies the current line according to the current margins
 */
void
justi_line (left_margin, jushop, first_line, justi_tabs, space_entered, auto_jus)
	int left_margin;
	int jushop;
	FLAG first_line;
	FLAG justi_tabs;
	int space_entered;	/* was last appended character a space? */
	int auto_jus;		/* is this a case of justify while typing? */
{
  char * poi;
  char * last_blank;
  char * very_last_blank;
  int column;

  poi = cur_line->text;
  column = 0;
/*
old:
  while (column < left_margin && (poi < cur_text || white_space (* poi)))
try:
  while (column < left_margin && (poi != cur_text && white_space (* poi)))
*/
  while (column < left_margin && 
		(poi < cur_text ||
		 (first_line == True && white_space (* poi))
		)
	)
  {
	if (* poi == '\t') {
		justi_tabs = True;
	}
	advance_char_col (& poi, & column);
  }
  trace_justi (("justi (%d, %d) advanced to %d (%s)\n", left_margin, jushop, column, trcstr (cur_line->text, poi)));
  if (column < left_margin) {
	move_address (poi, y);
	while (column < left_margin) {
		if (justi_tabs == True && tab (column) <= left_margin) {
			S0 ('\t');
			column = tab (column);
		} else {
			S0 (' ');
			column ++;
		}
	}
	poi = cur_line->text;	/* old text pointer may be invalid */
	column = 0;		/* so start again */
  }
  trace_justi ((" -> advanced to %d (%s)\n", column, trcstr (cur_line->text, poi)));
  last_blank = NIL_PTR;
  very_last_blank = NIL_PTR;
  while (column < right_margin && * poi != '\n') {
	if (column > left_margin && white_space (* poi))
	{
		trace_justi (("  last blank (%02X) at %d\n", * poi, column));
		if (space_entered) {
			last_blank = very_last_blank;
		} else {
			last_blank = poi;
		}
		very_last_blank = poi;
	}
	advance_char_col (& poi, & column);
  }
  if (! auto_jus) {
	last_blank = very_last_blank;
  }

  trace_justi ((" -> advanced to %d (%s)\n", column, trcstr (cur_line->text, poi)));
  if (* poi != '\n') {
	if (very_last_blank != NIL_PTR) {
		poi = very_last_blank;
		advance_char (& poi);
		move_address (poi, y);
		trace_justi ((" entering newline after last blank\n"));
		S0 ('\n');
		justi_line (next_left_margin, jushop, False, justi_tabs, space_entered, auto_jus);
	}
	else {	/* no wrapping point found */
		while (! white_space (* poi) && * poi != '\n') {
			advance_char_col (& poi, & column);
				/* to handle Chinese */
		}
		if (* poi == '\n') {
			move_address (poi, y);
			trace_justi ((" moving right beyond newline\n"));
			MRT ();
		}
		else {
			advance_char (& poi);
			move_address (poi, y);
			if (* poi != '\n') {
				trace_justi ((" entering newline\n"));
				S0 ('\n');
			} else {
				trace_justi ((" moving right\n"));
				MRT ();
			}
			justi_line (next_left_margin, jushop, False, justi_tabs, space_entered, auto_jus);
		}
	}
  }
  else if (poi - last_blank == 1 ||
	   (jushop > 0 && nonblank_line_follows ()))
  {
  /* continue with next line */
	move_address (poi, y);
	if (poi - last_blank != 1) {
		trace_justi ((" inserting a blank\n"));
		S0 (' ');
	}
	trace_justi ((" concatenating lines\n"));
	/* concatenate lines */
	if (cur_line->next != tail) {
		DCC0 ();
	}
	while (white_space (* cur_text)) {
		/* remove leading space */
		DCC0 ();
	}

	if (* cur_text != '\n') {
		justi_line (next_left_margin, jushop, False, justi_tabs, space_entered, auto_jus);
	} else {
		trace_justi ((" removing trailing space\n"));
		poi = cur_text;
		precede_char (& poi, cur_line->text);
		while (white_space (* poi)) {
			DPC0 ();
			poi = cur_text;
			precede_char (& poi, cur_line->text);
		}
		MRT ();
	}
  }
  else {
	trace_justi ((" moving right finally\n"));
	move_address (poi, y);
	MRT ();
  }
}

void
justi (left_margin, jushop, space_entered, auto_jus)
	int left_margin;
	int jushop;
	int space_entered;	/* was last appended character a space? */
	int auto_jus;		/* is this a case of justify while typing? */
{
  justi_line (left_margin, jushop, True, False, space_entered, auto_jus);
}

void
JUS ()
{
  int space_entered = white_space (* (cur_text - 1));

  trace_justi (("\nJUS ------------------------\n"));
  if (hop_flag > 0) {
	hop_flag = 0;
	justi (first_left_margin, 1 - JUSmode, space_entered, 0);
  } else
	justi (first_left_margin, JUSmode, space_entered, 0);
}

void
JUSclever ()
{
  char * poi = cur_line->text;
  int column = 0;
  char * ppoi;
  char lastc = '\n';
  int save_first = first_left_margin;
  int save_next = next_left_margin;
  FLAG javadoc = False;
  int utfcount;
  unsigned long unichar;

  /* set left margins to first non-blank in current line */
  while (white_space (* poi)) {
	advance_char_col (& poi, & column);
  }
  first_left_margin = column;
  next_left_margin = first_left_margin;
  if (first_left_margin > right_margin - 2) {
	right_margin = first_left_margin + 2;
  }

  /* check end of previous line for blank to see if current line 
     begins a paragraph */
  ppoi = cur_line->prev->text;
  if (ppoi != NIL_PTR) {
	lastc = * ppoi;
	if (lastc != '\n') {
		while (* ppoi != '\n') {
			lastc = * ppoi;
			advance_char (& ppoi);
		}
	}
  }
  if (lastc != ' ') {
	/* current line begins a paragraph */
	/* !"#$%&'()*+,-./0123456789:;<=>?
	   @ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_
	   `abcdefghijklmnopqrstuvwxyz{|}~ */
trace_jc (("skip letters from %d\n", column));
	if (utf8_text == True) {
		utf8_info (poi, & utfcount, & unichar);
	} else if (Chinese == True && multichar (* poi)) {
		unichar = '\0';
	} else {
		unichar = * poi;
	}
	if (  unichar == '@'
	   || (unichar == '/' && * (poi + 1) == '*')
	   || unichar == '*'
	   || unichar == (character) ''
	   )
	{
		javadoc = True;
	}
	while   (  (* poi < '#' && * poi > ' ')
		|| * poi == '*'
		|| (* poi >= '/' && * poi <= '9')
		|| * poi > '@'
		)
	{
		/* skip letters ... */
		advance_char_col (& poi, & column);
	}
trace_jc (("check marker at %d\n", column));
	if (javadoc == False && * poi != '.' && * poi != ')') {
		/* normal text */
		next_left_margin = first_left_margin;
	} else {
trace_jc (("skip non-space from %d\n", column));
		/* numbered text / listing */
		while (! white_space (* poi)) {
			advance_char_col (& poi, & column);
		}
trace_jc (("skip space from %d\n", column));
		while (white_space (* poi)) {
			advance_char_col (& poi, & column);
		}
trace_jc (("arrive at %d\n", column));
		next_left_margin = column;
		if (next_left_margin > right_margin - 2) {
			right_margin = next_left_margin + 2;
		}
	}
  }
trace_jc (("margins %d-%d.\n", first_left_margin, next_left_margin));

  JUS ();

  first_left_margin = save_first;
  next_left_margin = save_next;
}

/* JUSandreturn is called by character insertion S1 () */
void
JUSandreturn ()
{
  LINE * prevline = cur_line->prev;
  int prev_offset;
  int i;
#ifdef garble
  int not_at_lineend = * cur_text != '\n';
#endif

  char * poi;
  int column;	/* effective visual column of current position */
  int space_entered = white_space (* (cur_text - 1));

  trace_justi (("\nJUSandreturn ------------------------\n"));
  poi = cur_line->text;
  column = 0;
  prev_offset = 0;
  while (poi < cur_text) {
	advance_char_col (& poi, & column);
	prev_offset ++;
  }
  if (JUSlevel < 2) {
	if (column <= right_margin) {
		return;
	}
  }

  justi (first_left_margin, JUSmode, space_entered, 1);
  move_address (prevline->next->text, find_y (prevline->next));
  for (i = 0; i < prev_offset; i ++) {
	if (* cur_text == '\n') {
		MRT ();
		/* skip indentation */
		column = 0;
		poi = cur_text;
		while (white_space (* poi) && column < next_left_margin) {
			advance_char_col (& poi, & column);
		}
		if (column > 0) {
			move_address (poi, y);
		}
	}
	MRT ();
  }
#ifdef garble
  if (not_at_lineend && * cur_text == '\n') {
printf (" moving right 1 more\n");
	MRT ();
	while (white_space (* cur_text)) {
printf (" moving right 1 more\n");
		MRT ();
	}
  }
#endif
}

void
modify_int (name, var, min, max)
  char * name;
  int * var;
  int min;
  int max;
{
  int number;

  build_string (text_buffer, "%s (%d), new value:", name, * var);
  if (get_number (text_buffer, '\0', & number) == ERRORS) {
	return;
  }
  if (number < min) {
	error ("Value too small", NIL_PTR);
	return;
  }
  if (number > max) {
	error ("Value too large", NIL_PTR);
	return;
  }
  * var = number;
}

void
modify_col (name, var, min, max)
  char * name;
  int * var;
  int min;
  int max;
{
  int number;

  build_string (text_buffer, "%s (%d), new value (Enter for current column):", name, * var);
  if (get_number (text_buffer, '0', & number) == ERRORS) {
	return;
  }
  if (number == 0) {
	number = x + 1;
  }
  if (number < min) {
	error ("Value too small", NIL_PTR);
	return;
  }
  if (number > max) {
	error ("Value too large", NIL_PTR);
	return;
  }
  * var = number;
}

void
ADJLM ()
{	next_left_margin ++;
	modify_col ("both left margins", & next_left_margin, 1, right_margin - 2);
	next_left_margin --;
	first_left_margin = next_left_margin;
}

void
ADJFLM ()
{	first_left_margin ++;
	modify_col ("first left margin", & first_left_margin, 1, right_margin - 2);
	first_left_margin --;
}

void
ADJNLM ()
{	next_left_margin ++;
	modify_col ("next left margin", & next_left_margin, 1, right_margin - 2);
	next_left_margin --;
}

void
ADJRM ()
{
  int left = next_left_margin;
  if (first_left_margin > left) {
	left = first_left_margin;
  }

  right_margin --;
  modify_col ("right margin", & right_margin, left + 2, 999);
  right_margin ++;
}

void
ADJPAGELEN ()
{
	modify_int ("assumed lines per page", & lines_per_page, 0, 999);
}


/*======================================================================*\
|*			Miscellaneous					*|
\*======================================================================*/

/*
   Set flags depending on file type.
 */
void
set_file_type_flags ()
{
  char * suffix = strrchr (file_name, '.');
  if (suffix != NIL_PTR) {
	suffix ++;
  } else {
	suffix = "";
  }
  if (streq (suffix, "html")
	|| streq (suffix, "htm")
	|| streq (suffix, "sgml")
	|| streq (suffix, "xml")
	|| streq (suffix, "jsp")
#ifdef msdos
	|| streq (suffix, "HTML")
	|| streq (suffix, "HTM")
	|| streq (suffix, "SGML")
	|| streq (suffix, "XML")
	|| streq (suffix, "JSP")
#endif
     )
  {
	dim_HTML = True;
  } else {
	dim_HTML = False;
  }
}


/*
 * return the mined command associated with the key value
 */
voidfunc
command (c)
  int c;
{
  if (c == (character) FUNcmd) {
	return keyproc;
  } else {
	return key_map [c];
  }
}


#ifdef unix
/*
 * Set window headline and icon text
 */
static
void
build_header (ws, fn, modf)
	char * ws;
	char * fn;
	char * modf;
{
  build_string (ws, window_string_code, fn, modf, fn, modf);
}
void
RD_window_title ()
{
  char window_string [MAX_CHARS];
  char filename_ok [maxLINE_LEN];
  char * filename_dispoi;
  char * filename_poi;
  character mask;

  if (utf8_screen == True) {
	mask = 0xE0;
  } else {
	mask = 0x60;
  }

  filename_poi = file_name;
  filename_dispoi = filename_ok;
  while (* filename_poi != '\0') {
	if ((* filename_poi & mask) == 0) {
		* filename_dispoi = '?';
	} else {
		* filename_dispoi = * filename_poi;
	}
	filename_poi ++;
	filename_dispoi ++;
  }
  * filename_dispoi = '\0';

  if (loading == False) {
	build_header (window_string, file_name [0] == '\0' ? "[buffer]" : filename_ok,
					modified == True ? mined_modf : "");
/*	putstring (window_string);	*/
/*	Mind! As long as screen buffer shared with file buffer:	*/
	write (output_fd, window_string, (unsigned int) length_of (window_string));
  }
}
void
clear_window_title ()
{
  char window_string [MAX_CHARS];

	build_header (window_string, " ", " ");
/*	putstring (window_string);	*/
/*	Mind! As long as screen buffer shared with file buffer:	*/
	write (output_fd, window_string, (unsigned int) length_of (window_string));
}
#endif

/*
 * Set the modified flag
 */
void
set_modified ()
{
  if (modified == False) {
	modified = True;
#ifdef unix
	RD_window_title ();
#endif
  }
}

/*
 * Redraw the screen
 */
void
RD_nobot ()
{
  reverse_off ();
  clearscreen ();

/* display page */
  display (0, top_line, last_y, y);

/* redraw scroll bar */
  if (scrollbar == True) {
	(void) display_scrollbar (False);
  }

/* clear/redraw last line */
  set_cursor (0, YMAX);
  clear_lastline ();
  move_y (y);

#ifdef unix
  RD_window_title ();
#endif
}

void
RD ()
{
  RD_nobot ();
  if (stat_visible == True) {
	rd_bottom_line ();
  }
}

void
RD_y (y_pos)
  int y_pos;
{
  reverse_off ();
  clearscreen ();

/* display page */
  display (0, top_line, last_y, y_pos);

/* clear/redraw last line */
  set_cursor (0, YMAX);
  clear_lastline ();
  if (stat_visible == True) {
	rd_bottom_line ();
  }
}

/*
 * Adjust current window size after WINCH signal
 */
void
RDwin ()
{
  register LINE * current_line;

  winchg = False;
  getwinsize ();

  if (loading == False) {
	current_line = cur_line;
	reset (top_line, y);
/*	move_y (find_y_w_o_RD (current_line)); */
	move_address (cur_text, find_y_w_o_RD (current_line));
	RD_nobot ();
	if (MENU) {
		displaymenuline ();
		set_cursor_xy ();
		redrawmenu ();
	}
  }
  if (stat_visible == True) {
	rd_bottom_line ();
  }
  flush ();
}

void
change_screen_size (sb, keep_columns)
  FLAG sb;
  FLAG keep_columns;
{
  int index;
  int mode1;
#ifdef msdos_screenfunctions
  int mode2;
#endif

/* Experimental area: */
/*	set_screen_mode (mode1);	any available mode number */
/*	set_video_lines (mode1);	0/1/2: 200/350/400 lines */
	/* does not seem to have any effect */
/*	set_textmode_height (mode1);	0/1/2: font height 8/14/16 */
/*	set_grafmode_height (mode1, mode2);
		0/1/2: font height 8/14/16 1/2/3/n: 14/25/43/n lines */
/*	set_fontbank (f);		0..7 */
/**/
  if (hop_flag > 0) {
#ifdef msdos_screenfunctions
    if (keep_columns == True) {
      if (sb == BIGGER) {
	index = get_number ("Switch to font bank (0..7) ", '\0', & mode1);
	if (index == ERRORS) {
		return;
	}
	set_fontbank (mode1);
      } else {
	index = get_number ("Set character height (<= 32 pixels) ", '\0', & mode1);
	if (index == ERRORS) {
		return;
	}
	set_font_height (mode1);
      }
    } else {
      if (sb == BIGGER) {
#endif
	index = get_number ("Select video mode ", '\0', & mode1);
	if (index == ERRORS) {
		return;
	}
	set_screen_mode (mode1);
#ifdef msdos_screenfunctions
      } else {
	index = get_number ("Select graf font (0/1/2: font height 8/14/16) ", '\0', & mode1);
	if (index == ERRORS) {
		return;
	}
	index = get_number ("Select line number (1/2/3/n: 14/25/43/n) ", '\0', & mode2);
	if (index == ERRORS) {
		return;
	}
	set_grafmode_height (mode1, mode2);	/* 0/1/2: font height 8/14/16 */
					/* 1/2/3/n: 14/25/43/n lines */
      }
    }
#endif
  }
  else {
	resize_the_screen (sb, keep_columns);
  }
  RDwin ();
}
void
screenmorelines ()
{
  change_screen_size (BIGGER, True);
}
void
screenlesslines ()
{
  change_screen_size (SMALLER, True);
}
void
screenbigger ()
{
  change_screen_size (BIGGER, False);
}
void
screensmaller ()
{
  change_screen_size (SMALLER, False);
}

void
LNCI ()
{
  switch_textmode_height (True);
  RDwin ();
}

void
LNSW ()
{
  switch_textmode_height (False);
  RDwin ();
}

/*
 * Ignore this keystroke.
 */
void
I ()
{
}

/*
 * Fortifying 'HOP' key.
 */
void
HOP ()
{
  hop_flag = 2;
  if (MENU) {
	displayflags ();
	set_cursor_xy ();
	flush ();
  }
  if (! char_ready_within (500))
	status_msg ("Continue HOP command (next command fortified) ...");
}

/*
 * Cancel prefix function.
 */
void
CANCEL ()
{
  hop_flag = 0;
  clear_status ();
}

/*
 * Call proc associated with function key.
 */
void
FUNKEY ()
{
  (* keyproc) ('\0');
  keyproc = I;
}

/*
 * Toggle insert/overwrite mode.
 */
void
TOGINS ()
{
  if (insert_mode == True) {
	insert_mode = False;
  } else {
	insert_mode = True;
  }
}

/*
 * Toggle input character translation mode.
 */
void
TOGGO ()
{
  if (transdeutsch == True) {
	transdeutsch = False;
  } else {
	transdeutsch = True;
  }
}

#define cmd_char(c)	(c < '\040' ? c + '\100' : (c >= '\140' ? c - '\040' : c))

/*
 * Interpret control-Q commands. Most can be implemented with the Hop function.
 */
void
ctrlQ ()
{
  character c;
  void (* func) ();

  if (! char_ready_within (500)) {
	status_msg ("^Q: blockBegin Find replAce goto<n>mark HOP...");
  }
  if (quit == True) return;

  c = readchar_nokeymap ();
  if (quit == True) return;

  clear_status ();
  if ('0' <= c && c <= '9') 
	{GOMAn ((int) c - (int) '0'); return;}
  if (c == '\033' || c == QUITCHAR) 
	{CANCEL (); return;}
  switch (cmd_char (c)) {
	case 'B' : {GOMA () ; return;}
	case 'K' : { ; return;}		/* not exactly WS function */
	case 'P' : { ; return;}		/* not exactly WS function */
	case 'V' : { ; return;}		/* not exactly WS function */
	case 'W' :			/* not exactly WS function */
	case 'Z' :			/* not exactly WS function */
	case 'Y' :
	case '\177' : {
			func = key_map [c];
			hop_flag = 1;
			(* func) (c);
			return;
		      }
	case 'F' : {if (hop_flag > 0)
			SRV ();
		    else
			SFW ();
		    return;
		   }
	case 'A' : {if (hop_flag > 0)
			REPL ();
		    else
			GR ();
		    return;
		   }
	case 'Q' : {REPT (' '); return;}	/* not exactly WS function */
	case 'L' :			/* not exactly WS function */
/*
^Q: B/K top/bottom block
    P last position
    W/Z continuous scroll
    V last find or block
    Y/DEL delete line right/left
    0-9 marker
    F find
    A replace
    Q repeat next key/command
    L find misspelling
*/
	default : {
		   func = key_map [c];
		   if ((c < ' ') || ((voidfunc) func == (voidfunc) FUNKEY)) {
	/* (voidfunc) is an identity cast here. It seems to be required for
	   the sake of the apparently totally rotten microvax C compiler */
			hop_flag = 1;
			(* func) (c);
		   }
		   else
			BAD (c);
		   return;
		  }
  }
}

/*
 * Interpret control-K commands.
 */
void
ctrlK ()
{
  character c;

  if (! char_ready_within (500))
	status_msg ("^K: Save Done eXit Quit Read Log <n>mark / block: B/K mark Cop Ydel moV Wr...");
  if (quit == True) return;

  c = readchar_nokeymap ();
  if (quit == True) return;

  clear_status ();
  if ('0' <= c && c <= '9') 
	{MARKn ((int) c - (int) '0'); return;}
  if (c == '\033' || c == QUITCHAR) 
	{CANCEL (); return;}
  switch (cmd_char (c)) {
	case 'S' : {WTU (); return;}
	case 'D' : {EXFILE (); return;}
	case 'X' : {EXMINED (); return;}
	case 'Q' : {QUED (); return;}
	case 'B' : {MARK () ; return;}
	case 'K' : {YA () ; return;}	/* not exactly WS function */
	case 'H' : { ; return;}		/* not exactly WS function */
	case 'C' : {PT () ; return;}	/* not exactly WS function */
	case 'Y' : {DT () ; return;}	/* not exactly WS function */
	case 'V' : {PT (); return;}	/* not exactly WS function */
	case 'W' : {WB (); return;}	/* not exactly WS function */
	case 'N' : { ; return;}		/* not exactly WS function */
	case 'R' : {INSFILE (); return;}
	case 'L' : {CHDI (); return;}
/*
^K  0-9 set/hide marker
    B/K block begin/end
    H block hide
    C/Y/V/W block copy/delete/move/write
    N column block
*/
	default : {
		   BAD (c);
		   return;
		  }
  }
}

/*
 * Interpret control-O commands.
 */
void
ctrlO ()
{
  character c;

  if (! char_ready_within (500))
	status_msg ("^O: L/R left/right margins...");
  if (quit == True) return;

  c = readchar_nokeymap ();
  if (quit == True) return;

  clear_status ();
  if ('0' <= c && c <= '9') 
	{return;}
  if (c == '\033' || c == QUITCHAR) 
	{CANCEL (); return;}
  switch (cmd_char (c)) {
	case 'L' : {ADJLM (); return;}
	case 'R' : {ADJRM (); return;}
	case 'G' : {ADJFLM () /* actually paragraph tab */; return;}
/*
^O  L/R/M set left/right margin /release
    I/N set/clear tab
    F ruler from line
    C center line
    S set line spacing
    W toggle word wrap
    T toggle ruler line
    J toggle justify
    V     vari-tabs
    H     hyph-help
    E     soft hyph
    D     print display
    P     page break
*/
	default : {
		   BAD (c);
		   return;
		  }
  }
}

/*
 * Set marker / go to marker.
 */
void
MARKER ()
{
  character c;

/*  if (! char_ready_within (500))*/
  	status_msg ("0..9: set marker / , or blank: default marker");
/*  if (quit == True) return;*/

  c = readchar ();
  if (quit == True) return;

  clear_status ();
  if (c == '\033' || c == QUITCHAR) 
	{CANCEL (); return;}
  if ('0' <= c && c <= '9') 
	{MARKn ((int) c - (int) '0'); return;}
  if (c == ',' || c == '\'' || c == ' ' || c == ']' || c == '\035') 
	{MARK (); return;}
  BAD (c);
  return;
}
void
GOMARKER ()
{
  character c;

/*  if (! char_ready_within (500))*/
  	status_msg ("0..9: go marker / blank: default marker");
/*  if (quit == True) return;*/

  c = readchar ();
  if (quit == True) return;

  clear_status ();
  if (c == '\033' || c == QUITCHAR) 
	{CANCEL (); return;}
  if ('0' <= c && c <= '9') 
	{GOMAn ((int) c - (int) '0'); return;}
  if (c == ',' || c == '.' || c == 'g' || c == 'G' || c == '\'' || 
      c == ' ' || c == ']' || c == '\035') 
	{GOMA (); return;}
  BAD (c);
  return;
}

/*
 * ...MENU () opens the respective menu
 */
void
FILEMENU ()
{
  openmenu (0);
}
void
EDITMENU ()
{
  openmenu (1);
}
void
SEARCHMENU ()
{
  openmenu (2);
}
void
EXTRAMENU ()
{
  openmenu (3);
}
void
PARAMENU ()
{
  openmenu (4);
}
void
QUICKMENU ()
{
  QUICKMEN (x, y);
}

/*
   Toggle TAB width.
 */
void
toggletab ()
{
  if (tabsize == 4) {
	tabsize = 8;
  } else {
	tabsize = 4;
  }
  RDwin ();
}

/*
 * Interpret Escape commands.
 */
void
ESCAPE ()
{
  character c;
  voidfunc func;

  if (! char_ready_within (500)) {
	status_msg ("ESC(exit) TAB,blank(menu) q(uit /,\\(search) (match r(eplace g(oto h(elp ...");
  }
  if (quit == True) {
	return;
  }

  c = readchar_nokeymap ();
  if (quit == True) {
	return;
  }

  clear_status ();

  if ('0' <= c && c <= '9') {
	REPT (c);
	return;
  }

  switch (c) {
	case '\033' : {EXED (); return;}
	case 'q' : {QUED (); return;}
	case '/' : {SFW (); return;}
	case '\\' : {SRV (); return;}
	case 'R' : {LR (); return;}
	case 'r' : {REPL (); return;}
	case 'w' : {WT (); return;}
	case 'W' : {WTU (); return;}
	case 'v' : {VIEW (); return;}
	case 'E' : {EDITmode (); return;}
	case 'V' : {VIEWmode (); return;}
	case 'g' : {GOTO (); return;}
	case 'h' : {HELP (); return;}
	case '?' : {FS (); return;}
	case '.' : {RDwin (); return;}
	case 'm' : {MARKER (); return;}
	case '\'' : {GOMARKER (); return;}
	case 'i' : {INSFILE (); return;}
	case 'b' : {WB (); return;}
	case '=' : {REPT (' '); return;}
	case 'z' : {SUSP (); return;}
	case 'd' : {CHDI (); return;}
	case '!' : {SH (); return;}
	case '^' : {MARK (); return;}
	case ']' : {GOMA (); return;}
	case 'n' : {NN (); return;}
	case 'c' : {CMD (); return;}
	case 'u' : {display_code (); return;}
	case 'U' : {changeuni (); return;}
	case 'X' : {changehex (); return;}
	case 'A' : {changeoct (); return;}
	case 'D' : {changedec (); return;}
	case '+' : {NXTFILE (); return;}
	case '-' : {PRVFILE (); return;}
	case '#' : {NTHFILE (); return;}
	case '%' : {screensmaller (); return;}
	case '&' : {screenbigger (); return;}
	case 'l' : {screenlesslines (); return;}
	case 'L' : {screenmorelines (); return;}
	case 'J' : {JUS (); return;}
	case 'j' : {JUSclever (); return;}
	case '<' : {ADJLM (); return;}
	case ';' : {ADJFLM (); return;}
	case ':' : {ADJNLM (); return;}
	case '>' : {ADJRM (); return;}
	case 'P' : {ADJPAGELEN (); return;}
	case 'T' : {toggletab (); return;}
	case 'H' : {HTML (); return;}
	case '_' : {UML (); return;}
	case 'C' : {LOWCAP (); return;}
	case '(' : {SCORR (); return;}
	case ')' : {SCORR (); return;}
	case '{' : {SCORR (); return;}
	case '}' : {SCORR (); return;}
	case 't' : {Stag (); return;}
	case 'a' : {toggleappend (); return;}
	case 'k' : {toggleKEYMAP (); return;}
	case 'K' : {setupKEYMAP (); return;}
	case 'Q' : if (utf8_text == True) {
			if (hop_flag > 0) {
				quote_type_up ();
			} else {
				handleQuotemenu (0);
			}
			displayflags ();
		   } else {
			error ("Smart quotes not active", NIL_PTR);
		   }
		   return;
/*	case 'o' : {TOGINS (); return;}	*/
/*	case '\t' : {FILEMENU (); return;}	*/
	case ' ' : {QUICKMENU (); return;}
	case 'f' : {FILEMENU (); return;}
	case 'e' : {EDITMENU (); return;}
	case 's' : {SEARCHMENU (); return;}
	case 'x' : {EXTRAMENU (); return;}
	case 'p' : {PARAMENU (); return;}
	case ',' : {GR (); return;}
/*
	case 'e' : {EDIT (); return;}
	case 's' : {GR (); return;}
	case 'p' : {PBUF (); return;}
*/
	case QUITCHAR : {CANCEL (); return;}
	default : {
		   func = key_map [c];
		   if ((c < ' ') || ((voidfunc) func == (voidfunc) FUNKEY)) {
	/* (voidfunc) is an identity cast here. It seems to be required for
	   the sake of the apparently totally rotten microvax C compiler */
			hop_flag = 1;
			(* func) (c);
		   } else {
			BAD (c);
		   }
		   return;
		  }
  }
}

/*
 * MOUSEescape () is invoked after a mouse escape sequence with 
 * coordinates and button info already stored in 
 * mouse_xpos, mouse_ypos, and mouse_button.
 * Then it performs a menu or other mouse controlled function.
 */
void
MOUSEescape ()
{
  if (mouse_button == movebutton) {
	/* ignore */
  } else if (mouse_ypos == -1) {	/* menu stuff */
	openmenuat (mouse_xpos);
  } else if (mouse_ypos == YMAX) {
	if (mouse_button == leftbutton) {
		PD ();
	} else if (mouse_button == middlebutton) {
		FS ();
	} else if (mouse_button == rightbutton) {
		PU ();
	}
  } else if (scrollbar == True && mouse_xpos == XMAX) {
	if (mouse_button == leftbutton) {
		PD ();
	} else if (mouse_button == rightbutton) {
		PU ();
	} else if (mouse_button == middlebutton
		   || (mouse_button == releasebutton
		       && mouse_lastbutton == middlebutton)) {
		goproz (mouse_ypos * 100 / YMAX);
	}
  } else if (quickmenu == True) {
	if (mouse_ypos > last_y) {
		mouse_ypos = last_y;
	}
	if (mouse_button == leftbutton) {
		move_to (mouse_xpos, mouse_ypos);
		report_release = True;
	} else if (mouse_button == middlebutton) {
		FS ();
	} else if (mouse_button == rightbutton) {
		move_to (mouse_xpos, mouse_ypos);
		QUICKMEN (mouse_xpos, mouse_ypos);
	} else if (mouse_button == releasebutton 
		&& mouse_lastbutton == leftbutton 
		&& (mouse_xpos != mouse_lastxpos 
		    || mouse_ypos != mouse_lastypos)) {
		MARK ();
		move_to (mouse_xpos, mouse_ypos);
		YA ();
	}
  } else {
	if (mouse_ypos > last_y) {
		mouse_ypos = last_y;
	}
	move_to (mouse_xpos, mouse_ypos);
	if (mouse_button == leftbutton) {
		MARK ();
	} else if (mouse_button == middlebutton) {
		PT ();
	} else if (mouse_button == rightbutton) {
		YA ();
	}
  }
}

/*
 * REPT () prompts for a count and wants a command after that. It repeats the
 * command count times. If a ^\ is given during repeating, stop looping and
 * return to main loop.
 */
void
REPT (firstdigit)
  char firstdigit;
{
  register int count;
  register void (* func) ();
  int index;
  int number;

  hop_flag = 0;
  if (firstdigit >= '0' && firstdigit <= '9') {
     index = get_number ("Please continue repeat count...", firstdigit, & number);
     if (firstdigit != '0' && number < 10) {
	error ("Invalid repeat count after ESC <digit>", NIL_PTR);
	return;
     }
  } else {
     index = get_number ("Please enter repeat count...", '\0', & number);
  }
  if (index == ERRORS) {
	return;
  }

  func = key_map [index];
  if ((voidfunc) func == (voidfunc) I) {	/* Function assigned? */
	/* (voidfunc) is an identity cast here. It seems to be required for
	   the sake of the apparently totally rotten microvax C compiler */
	clear_status ();
	return;
  }
  if ((voidfunc) func == (voidfunc) FUNKEY) {
	/* (voidfunc) is an identity cast here. It seems to be required for
	   the sake of the apparently totally rotten microvax C compiler */
/*	func = * keyproc;	*/
	func = keyproc;
	keyproc = I;
	index = '\0';
  }

  count = number;
  while (count -- > 0 && quit == False) {
	if (stat_visible == True) {
		clear_status ();
	}
	(* func) (index);
	flush ();
  }

  if (quit == True)		/* Abort has been given */
	error ("Repeat aborted", NIL_PTR);
  else
	clear_status ();
}

/*
 * Complains to illegal commands and eats up illegal escape sequences.
 */
void
BAD (c)
  character c;
{
  static char message2 [] = "'**' - type a blank";

  if (c < ' ')
	{ message2 [1] = '^'; message2 [2] = c + '@'; }
  else
	{ message2 [1] = ' '; message2 [2] = c; }
  error ("Unknown command ", message2);
  while (readchar () != ' ' && quit == False) {
	ring_bell ();
	flush ();
  }
  clear_status ();
}

/*
 * Change working directory.
 */
void
CHDI ()
{
  char new_dir [maxLINE_LEN];	/* Buffer to hold new dir. name */

  if (restricted == True)
	{restrictederr (); return;}

#ifdef pc
  build_string (text_buffer, "Drive/Directory: %s, change to:", unnull (getcwd (new_dir, maxLINE_LEN)));
#else
  build_string (text_buffer, "Directory: %s, change to:", unnull (getcwd (new_dir, maxLINE_LEN)));
#endif

  if (get_file (text_buffer, new_dir) != FINE) {
	return;
  }
#ifdef msdos
  if (new_dir [0] != '\0' && new_dir [1] == ':')
	if (new_dir [2] == '\0') {
		new_dir [2] = '.';	/* change to current dir. of drive */
		new_dir [3] = '\0';
	}
#endif
  if (chdir (new_dir) == 0) {
#ifdef msdos
	if (new_dir [0] != '\0' && new_dir [1] == ':')
		setdisk (((int) new_dir [0] & (int) '\137') - (int) 'A');
	RD ();	/* disk error dialog may be on screen after chdir */
#endif
	clear_status ();
	overwriteOK = False;	/* Same file base name ... */
	writable = True;
/*	if (viewmode == False)	*/
	    set_modified ();	/* would mean different file now */
  } else {
#ifdef msdos
	RD ();	/* disk error dialog may be on screen */
#endif
	error ("Could not change work dir: ", serror ());
  }
}

/*
 * Print file status.
 */
void
FSTAT ()
{
  fstatus (file_name [0] ? "" : "[buffer]", -1L, -1L);
}
void
FS ()
{
  if (hop_flag > 0) {
	if (fstat_always == False)
		fstat_always = True;
	else
		fstat_always = False;
  } else
	FSTAT ();
}


/*
 * Checkout (from version managing system).
 */
void
checkout ()
{
  int save_cur_column;
  int save_cur_line;
  char syscommand [maxLINE_LEN];	/* Buffer for full system command */
  int sysres;

	if (modified == True) {
		if (write_text (False) != FINE) {
			return;
		}
	}

	/* save current position */
	save_cur_line = line_number;
	save_cur_column = get_cur_col ();

	/* try to check out */
	raw_mode (OFF);
	build_string (syscommand, "co %s", file_name);
	sysres = system (syscommand);
	sleep (1);
	raw_mode (ON);
	if (sysres != 0) {
		error ("Checkout failed", NIL_PTR);
	}

	/* reload file */
	initialize ();
	clearscreen ();
	load_file_position (file_name, True, save_cur_line, save_cur_column);
}

/*
 * Checkin (to version managing system).
 */
void
checkin ()
{
  char syscommand [maxLINE_LEN];	/* Buffer for full system command */
  int sysres;

	if (modified == True) {
		if (write_text (False) != FINE) {
			return;
		}
	}

	/* try to check in */
	raw_mode (OFF);
	build_string (syscommand, "ci %s", file_name);
	sysres = system (syscommand);
	sleep (1);
	raw_mode (ON);
	RDwin ();
	if (sysres != 0) {
		error ("Checkin failed", NIL_PTR);
	}
}


static FLAG save_viewonly;
static FLAG save_restricted;
static char save_file_name [maxLINE_LEN];
static FLAG viewing_help = False;
static int save_cur_column;
static int save_cur_line;

static
void
view_help (helpfile, item)
  char * helpfile;
  char * item;
{
  char searchstring [maxLINE_LEN];

  /* unless already viewing help, save edited text */
  if (viewing_help == False) {
	if (modified == True) {
		if (write_text (False) != FINE) {
			return;
		}
	}

	/* save current position */
	save_cur_line = line_number;
	save_cur_column = get_cur_col ();

	/* save editing mode and file name */
	save_viewonly = viewonly;
	save_restricted = restricted;
	copy_string (save_file_name, file_name);

	/* set mode appropriate for viewing online help */
	viewonly = True;
	restricted = True;
	viewing_help = True;

	/* load online help file */
	initialize ();
	clearscreen ();
	load_file_position (helpfile, True, -1, 0);
  }

  /* position to selected help topic */
  BFILE ();
  build_string (searchstring, "mined help topic '%s'", item);
  search_for (searchstring, FORWARD);
}

static
void
end_view_help ()
{
  viewonly = save_viewonly;
  restricted = save_restricted;
  viewing_help = False;

  initialize ();
  clearscreen ();
  load_file_position (save_file_name, True, save_cur_line, save_cur_column);
}

void
show_help (item)
  char * item;
{
#ifndef pc
  char syscommand [maxLINE_LEN];	/* Buffer for full system command */
  int sysres;
#endif
  char * helpfile;
  int hf = -1;

  helpfile = (char *) getenv ("MINEDHELPFILE");
  if (helpfile != NIL_PTR && * helpfile != '\0') {
	hf = open (helpfile, O_RDONLY);
  }
  if (hf == -1) {
	helpfile = "/usr/local/share/help/mined.hlp";
	hf = open (helpfile, O_RDONLY);
  }
  if (hf == -1) {
	helpfile = "/usr/local/info/mined.hlp";
	hf = open (helpfile, O_RDONLY);
  }
  if (hf == -1) {
	helpfile = "/usr/share/doc/packages/mined/mined.hlp";
	hf = open (helpfile, O_RDONLY);
  }
  if (hf == -1) {
	helpfile = "/opt/gnu/share/help/mined.hlp";
	hf = open (helpfile, O_RDONLY);
  }
  if (hf == -1) {
	helpfile = "/opt/gnu/info/mined.hlp";
	hf = open (helpfile, O_RDONLY);
  }

  if (hf == -1) {
	status_msg ("Help file not found; configure MINEDHELPFILE in environment!");
	return;
#ifndef pc
  } else if (! (hop_flag > 0 || viewing_help == True)) {
	(void) close (hf);
/*
	build_string (syscommand, "unzip -p %s %s | LESS=\"QMPMMined help topic '%s'?e (end):?pb (%%pb\\%%).. - type q to return to mined $ +Gg\" ${PAGER-less}", 
			helpfile, item, item);
*/
	build_string (syscommand, 
		"less '-QMPMMined help?e (end):?pb (%%pb\\%%).. - type \"'\\''h\" for help topic \"%s\", q to return to mined '"
		" +\"/mined help topic '%s'\nkmhG'h\" %s",
			item, item, helpfile);

	clear_status ();
	set_cursor (0, YMAX);
#ifdef unix
	clear_window_title ();
#endif
	raw_mode (OFF);

	sysres = system (syscommand);

	raw_mode (ON);
	RDwin ();
	if (sysres != 0) {
		error ("Help item not found in help file", NIL_PTR);
	}
#endif
  } else {
	(void) close (hf);
	view_help (helpfile, item);
  }
}

/*
 * Interactive Help
 */
void
HELP ()
{
  character c;

  status_msg ("Help on: i(ntroduction k(eypad f(unction-keys c(ommands m(enu");
  if (quit == True) return;

  c = readchar_nokeymap ();
  if (quit == True) return;

  clear_status ();
  switch (c) {
	case '\033': return;
	case QUITCHAR: return;
	case 'i': show_help ("introduction"); return;
	case 'k': show_help ("keypad"); return;
	case 'f': show_help ("function-keys"); return;
	case 'c': show_help ("commands"); return;
	case 'm': show_help ("menu"); return;
	default: status_msg ("No such help available"); return;
  }
}

/*
 * Print buffer
 */
void
PBUF ()
{
  int fd;
  char cmd [maxLINE_LEN];	/* Buffer for print command */

  if ((fd = yankfile (READ, False)) == ERRORS) {
	error ("Buffer is empty", NIL_PTR);
	return;
  }
  (void) close (fd);
  build_string (cmd, getenv ("MINEDPRINT") ?
		(char *) getenv ("MINEDPRINT") : printcommand, yank_file);
/* Turbo-C wants the cast here since getenv must be declared as far * */
  clear_status ();
  set_cursor (0, YMAX);
  flush ();
  system (cmd);
  sleep (1);
  RDwin ();
}

/*
 * Pipe buffer
 */
void
CMD ()
{
  int fd;
  char cmd [maxLINE_LEN];	/* Buffer for command */
  char command [maxLINE_LEN];	/* Buffer for full command */

  if (restricted == True)
	{restrictederr (); return;}

  if ((fd = yankfile (READ, False)) == ERRORS) {
	error ("Buffer is empty", NIL_PTR);
	return;
  }
  (void) close (fd);
  if (get_string ("Command with buffer as input:", cmd, True, "") != FINE) {
	return;
  }
  build_string (command, "%s < %s", cmd, yank_file);
  clear_status ();
  set_cursor (0, YMAX);
#ifdef unix
  clear_window_title ();
#endif
  raw_mode (OFF);

  system (command);

  sleep (1);
  raw_mode (ON);
  RDwin ();
}

/*
 * Called if an operation is not implemented on this system
 */
void
notavailable ()
{
  error ("Command not available", NIL_PTR);
}

/*
 * Suspend editor after writing back the file.
 */
void
SUSP ()
{
  if (restricted == True)
	{restrictederr (); return;}

  if (cansuspendmyself == True) {
	if (hop_flag == 0 && modified == True) {
		if (write_text (False) == ERRORS) {
			return;
		}
	}
	set_cursor (0, YMAX);
#ifdef unix
	clear_window_title ();
#endif
	raw_mode (OFF);

	suspendmyself ();
	raw_mode (ON);
	clear_status ();
	RDwin ();
  }
  else notavailable ();
}

/*
 * Call an interactive shell.
 */
void
SH ()
{

#ifdef unix
#define SHimplemented
  register int w;
  int pid;
  int status;
  int waiterr;

  if (restricted == True)
	{restrictederr (); return;}

#ifdef FORK
  switch (pid = fork ()) {
#else
  switch (pid = vfork ()) {
#endif
	case -1:			/* Error */
		error ("Cannot fork: ", serror ());
		return;
	case 0:				/* This is the child */
		set_cursor (0, YMAX);
		putchar ('\n');
		clear_window_title ();
		raw_mode (OFF);

		if (rpipe == True) {	/* Fix stdin */
			if (close (STD_IN) < 0) {
				_exit (126);
			}
			if (open ("/dev/tty", O_RDONLY, 0) < 0) {
				_exit (126);
			}
		}
		execl (getenv ("SHELL"), getenv ("SHELL"), 0);
		_exit (127);	/* Exit with 127 */
		/* NOTREACHED */
	default:			/* This is the parent */
		do {
			w = wait (& status);
		} while (w != -1 && w != pid);
		waiterr = geterrno ();
  }

  raw_mode (ON);
  RDwin ();

  if (w == -1) {
	error ("Wait error: ", serrorof (waiterr));
	if (((status >> 8) == 127) || ((status >> 8) == 126)) sleep (2);
  }
  if ((status >> 8) == 127)		/* Child died with 127 */
	error (getenv ("SHELL"), ": error invoking ${SHELL} (not found / not enough memory ?)");
  else if ((status >> 8) == 126)
	error ("Cannot open /dev/tty as fd #0", NIL_PTR);
#endif /* unix */


#ifdef msdos
#define SHimplemented
  char old_dir [maxLINE_LEN];	/* Buffer to hold dir. name */

  if (restricted == True)
	{restrictederr (); return;}

  (void) getcwd (old_dir, maxLINE_LEN);

  set_cursor (0, YMAX);
  raw_mode (OFF);

  system (getenv ("COMSPEC"));

  raw_mode (ON);
  clear_status ();
  RDwin ();

  if (chdir (old_dir) == 0) {
	if (old_dir [0] != '\0' && old_dir [1] == ':')
		setdisk (((int) old_dir [0] & (int) '\137') - (int) 'A');
	RD ();	/* disk error dialog may be on screen after chdir */
  } else {
	overwriteOK = False;	/* Same file base name ... */
	writable = True;
/*	if (viewmode == False)	*/
	    set_modified ();	/* would mean different file now */
	RD ();	/* disk error dialog may be on screen */
	error ("Could not reset previous work dir: ", serror ());
  }
#endif /* msdos */


#ifdef vms
#define SHimplemented
  if (restricted == True)
	{restrictederr (); return;}

/* Who can tell me why this hangs the process after return from the CLI ?
  set_cursor (0, YMAX);
  raw_mode (OFF);

  system ("SPAWN");

  raw_mode (ON);
  clear_status ();
  RDwin ();
*/
  notavailable ();
#endif


#ifndef SHimplemented
  notavailable ();
#endif
}


/*======================================================================*\
|*			Main						*|
\*======================================================================*/

char * minedopt;

void
WordStar_keys ()
{
  int i;

  for (i = 0; i < 32; i ++) key_map [i] = ws_key_map [i];
  control_prefix = '\020';
}

void
config_markers ()
{
  char * Mark;

  Mark = (char *) getenv ("MINEDSHIFT");
/* Turbo-C wants the cast here since getenv must be declared as far * */
  if (Mark != NIL_PTR) {
	SHIFT_MARK = Mark [0];
	if (Mark [0] != '\0') SHIFT_BEG = Mark [1];
  }

  Mark = (char *) getenv ("MINEDTAB");
  if (Mark != NIL_PTR) {
/*	TABchar = (Mark [0] == '\0' ? TABdefault : Mark [0]);	*/
	if (Mark [0] == '\0') TABchar = TABdefault;
	else {
		TABchar = Mark [0];
		TABchar0 = Mark [1];
		if (TABchar0 != '\0') TABchar2 = Mark [2];
		if (TABchar2 != '\0') {
			TABcharmid = Mark [3];
			TABchar0 = '\0';
		}
	}
  } else
	TABchar = TABdefault;
  UTF_TAB = (char *) getenv ("MINEDUTFTAB");
  if (UTF_TAB != NIL_PTR) {
	UTF_TAB0 = UTF_TAB;
	if (* UTF_TAB0 != '\0') advance_utf8 (& UTF_TAB0);
	UTF_TAB2 = UTF_TAB0;
	if (* UTF_TAB2 != '\0') advance_utf8 (& UTF_TAB2);
	UTF_TABmid = UTF_TAB2;
	if (* UTF_TABmid != '\0') {
		advance_utf8 (& UTF_TABmid);
		if (* UTF_TABmid != '\0') UTF_TAB0 = "";
	}
  }

  Mark = (char *) getenv ("MINEDRET");
  if (Mark != NIL_PTR) {
	RET_MARK = Mark [0];
	if (RET_MARK != '\0') RET_BLANK = Mark [1];
	if (RET_BLANK != '\0') RET_BLANK2 = Mark [2];
  } else
	RET_MARK = RETdefault;
  Mark = (char *) getenv ("MINEDDOSRET");
  if (Mark != NIL_PTR) {
	DOSRET_MARK = Mark [0];
  } else
	DOSRET_MARK = DOSRETdefault;
  Mark = (char *) getenv ("MINEDPARA");
  if (Mark != NIL_PTR) {
	PARA_MARK = Mark [0];
  } else
	PARA_MARK = PARAdefault;
  UTF_RET = (char *) getenv ("MINEDUTFRET");
  if (UTF_RET != NIL_PTR) {
	UTF_RETblank = UTF_RET;
	if (* UTF_RETblank != '\0') advance_utf8 (& UTF_RETblank);
	UTF_RETblank2 = UTF_RETblank;
	if (* UTF_RETblank2 != '\0') advance_utf8 (& UTF_RETblank2);
  }
  UTF_DOSRET = (char *) getenv ("MINEDUTFDOSRET");
  UTF_PARA = (char *) getenv ("MINEDUTFPARA");
}

FLAG
eval_option ()
{
  switch (* minedopt) {
	case '+': plus_opt = True; break;
	case 'M': MENU = 1 - MENU; break;
	case 'v': viewonly = True; break;
	case 'm': if (plus_opt == True)
			multiexit = False;
		  else
			multiexit = True;
		  break;
	case 'P': proportional = True; break;
	case 'p': paradisp = True; break;
	case 'r': if (RET_opt == 'r')
			RET_opt = ' ';
		  else
			RET_opt = 'r';
		  break;
	case 'R': if (plus_opt == True)
			RET_opt = 'M';
		  else
			RET_opt = 'R';
		  break;
	case 'G': if (plus_opt == True)
			enforce_block_graphics_menu = True;
		  else
			if (enable_block_graphics == True)
				enable_block_graphics = False;
			else
				enable_block_graphics = True;
		  break;
	case '-': restricted = True; break;
	case 'o': if (disp_scrollbar == False) {
			disp_scrollbar = True;
			scrollbar_width = 1;
		  } else {
			disp_scrollbar = False;
			scrollbar_width = 0;
		  }
		  break;
	case 'O': update_scrollbar_lazy = True;
		  break;
	case 'c': if (plus_opt == False) {
			combining_mode = False;
			combining_screen = False;
		  } else if (utf8_screen == True) {
			combining_mode = True;
			combining_screen = True;
		  }
		  break;
	case 'C': auto_UTF_detect = False;
		  utf8_text = False;
		  utf8_lineends = False;
		  Chinese = True;
		  /* make sure lineend marker will always fit: */
		  scrollbar_width = 1;
		  break;
	case 'U': if (utf8_input == True) {
			utf8_screen = False;
			utf8_input = False;
		  } else {
			utf8_screen = True;
			utf8_input = True;
		  }
		  break;
	case 'u': auto_UTF_detect = False;
		  if (plus_opt == True)
			utf8_text = False;
		  else {
			if (utf8_text == True) {
				/* double 'u' option */
				utf8_lineends = True;
			} else
				utf8_text = True;
		  }
		  break;
	case 'b': if (bidi == True)
			bidi = False;
		  else
			bidi = True;
		  break;
	case 'w': if (wordnonblank == True)
			wordnonblank = False;
		  else
			wordnonblank = True;
		  break;
	case 'a': append_flag = True; break;
	case 'j': JUSlevel = 1; break;
	case 'J': JUSlevel = 2; break;
	case 'B': /*
		  key_map ['\010'] = DPC;
		  key_map ['\177'] = DCC;
		  */
		  key_map ['\010'] = MLF;
		  key_map ['\177'] = DPC;
		  break;
	case 'T': tab_left = False; break;
	case 'D': if (transdeutsch == True) auto_transdeutsch = True;
		  transdeutsch = True;
		  break;
	case 'W': WordStar_keys (); break;
	case 's': page_stay = True; break;
	case 'S': page_scroll = True; break;
	case 'x': xprot = exeprot; break;
	case 'X': no_window_title = True; break;
	case 't': minedopt ++;
		  if (* minedopt == '\0') {minedopt --; TABchar = TABdefault;}
		  else TABchar = * minedopt;
		  break;
/* deprecated
	case 'L': minedopt ++;
		  if (* minedopt == '\0') minedopt --;
		  else linesplitmarker = * minedopt;
		  break;
*/
	case 'd': minedopt ++;
		  if (* minedopt == '-') display_delay = -1;
		  else if (* minedopt >= '0' && * minedopt <= '9')
			display_delay = (int) * minedopt - (int) '0';
		  else minedopt --;
		  break;
	case '/': minedopt ++;
		  inisearch = minedopt;
		  break;
	case '4': tabsize = 4;
		  break;
	case 'k': mined_keypad = False;
		  break;
	default:  fprintf (stderr, "Unknown option %c\n", * minedopt);
		  return False;
  }
  return True;
}

int
main (argc, argv)
  int argc;
  char * argv [];
{
  register int index;	/* index in key table */
  int initlinenum;
  int initlini = 0;
  LINE * initline;
  FLAG goon;
  FLAG cursor_somewhere = False;

/* fprot = umask (0); */

/* determine character encoding modes */
  if (getenv ("utf8_screen")) utf8_screen = True;
  if (getenv ("utf8_input")) utf8_input = True;
  if (getenv ("utf8_term")) {
	utf8_screen = True;
	utf8_input = True;
  }
  if (is_locale_utf8 ()) {
	utf8_screen = True;
	utf8_input = True;
  }
  if (! getenv ("utf8_no_combining_screen"))
	if (utf8_screen == True) {
		combining_screen = True;
		combining_mode = True;
	}

  if (getenv ("MINEDCHIN")) {
	auto_UTF_detect = False;
	utf8_text = False;
	utf8_lineends = False;
	Chinese = True;
	/* make sure lineend marker will always fit: */
	scrollbar_width = 1;
  }

/* determine various modes and display options from environment settings */
  if (getenv ("NoCtrlSQ") || getenv ("NoControlSQ")) {
	/* ^S and ^Q may come arbitrarily from terminal, so don't use them */
	controlQS = True;
	key_map ['\021'] = I;
	key_map ['\023'] = I;
  }
  if (getenv ("MINEDPROP")) proportional = True;
  if (getenv ("MINEDTURKISH")) Turkish = True;

  if (getenv ("MINEDMAC")) RET_opt = 'R';
  if (getenv ("MINEDWS")) WordStar_keys ();
  transout = (char *) getenv ("MINEDOUT");
  if (transout != NIL_PTR && * transout != '\0') {
	translate_output = True;
	translen = strlen (transout);
  }
#ifdef unix
  if (getenv ("MINEDMODF")) mined_modf = getenv ("MINEDMODF");
#endif
#ifdef pc_winlowdelay
  if (getenv ("windir")) display_delay = 0;
#endif

  dimansi = (char *) getenv ("MINEDDIM");
  if (dimansi == NIL_PTR) {
#ifdef pc
	dimansi = "1;31";
#else
	dimansi = "31";
#endif
  }

  uniansi = (char *) getenv ("MINEDUNI");
  if (uniansi == NIL_PTR) uniansi = "46";
  else if ((character) * uniansi > '9') {
	UNI_MARK = * uniansi;
	do {
		uniansi ++;
	} while (* uniansi == ' ');
  }
  unimarkansi = (char *) getenv ("MINEDUNIMARK");
  if (unimarkansi == NIL_PTR) unimarkansi = "36";
  combiningansi = (char *) getenv ("MINEDCOMBINING");
  if (combiningansi == NIL_PTR) combiningansi = "46";	/* 45? */

  ctrlansi = (char *) getenv ("MINEDCTRL");
  if (ctrlansi == NIL_PTR) ctrlansi = "";
  menuansi = (char *) getenv ("MINEDMENU");
  if (menuansi == NIL_PTR) menuansi = "";
  HTMLansi = (char *) getenv ("MINEDHTML");
  if (HTMLansi == NIL_PTR) HTMLansi = "34";
  diagansi = (char *) getenv ("MINEDDIAG");
  if (diagansi == NIL_PTR) diagansi = "";
  scrollfgansi = (char *) getenv ("MINEDSCROLLFG");
  if (scrollfgansi == NIL_PTR) scrollfgansi = "44";
  scrollbgansi = (char *) getenv ("MINEDSCROLLBG");
  if (scrollbgansi == NIL_PTR) scrollbgansi = "40;104;48;5;45";

  if ((minedopt = (char *) getenv ("MINED")) != NIL_PTR) {
     while (* minedopt != '\0') {
	(void) eval_option ();
	minedopt ++;
     }
  }

/* evaluate command line options (not file names) */
  fnami = 1;
  goon = True;
  do {
    if (fnami < argc) {
      if (* argv [fnami] == '+' && argv [fnami] [1] >= '0' && argv [fnami] [1] <= '9') {
	initlini = fnami;
	fnami += 1;
      }
      else if (* argv [fnami] == '-'
	      || * argv [fnami] == '+'
#ifdef msdos
	      || * argv [fnami] == '/'
#endif
	      ) {
	minedopt = argv [fnami];
	if (* minedopt == '+')
		plus_opt = True;
	else
		plus_opt = False;
	minedopt ++;
	while (goon == True && * minedopt != '\0') {
		goon = eval_option ();
		minedopt ++;
	}
	fnami += 1;
      } else {
	goon = False;
      }
    } else {
      goon = False;
    }
  } while (goon == True);

/* configure special markers unless Chinese mode is active */
  if (Chinese == False) {
	config_markers ();
  } else {
	SHIFT_MARK = ' ';
  }

/*
  get_term ();
*/

  fnami_min = fnami;
  fnami_max = argc - 1;
  fnami_cnt = argc - fnami_min;
  fnamv = argv;	/* Why did this produce a warning? C is such a stupid language! */
  if (! (fnami < argc)) {
     fnami = 0;
  }

  if (! isatty (STD_IN)) {	/* Reading from pipe */
	if (fnami != 0) {
		panic ("Cannot read both pipe and file", NIL_PTR);
	}
	rpipe = True;
	set_modified ();	/* Set modified flag not to loose buffer */
#ifdef msdos
	panic ("Cannot edit after input from pipe", "MSDOS C incompatibility");
#else
	if ((input_fd = open ("/dev/tty", O_RDONLY, 0)) < 0)
	   panic ("Cannot open /dev/tty for read", serror ());
#endif
  }
  if (! isatty (STD_OUT)) {
	wpipe = True;
	set_modified (); /* Set modified flag not to ignore buffer on exit */
    /*	if ((output_fd = open ("/dev/tty", O_WRONLY, 0)) < 0)
	   panic ("Cannot open /dev/tty for write", serror ()); */
  }

  get_term ();
  raw_mode (ON);	/* Set tty to appropriate mode */
  if (erase_char != '\0') {
	key_map [erase_char] = DPC;
  }
  clearscreen ();

/* select map of escape sequences for terminal function keys */
#ifndef msdos
  set_fkeymap (getenv ("MINEDTERM"));
  setKEYMAP (getenv ("MINEDKEYMAP"));
#endif

/*
 * Generate names of paste files and of panic-file
 */
#ifdef unix
  temp_dir = getenv ("MINEDTMP");
  if (temp_dir == NIL_PTR || temp_dir [0] == '\0' || access (temp_dir, R_OK | X_OK) < 0) {
	temp_dir = getenv ("TMPDIR");
  }
  if (temp_dir == NIL_PTR || temp_dir [0] == '\0' || access (temp_dir, R_OK | X_OK) < 0) {
	temp_dir = getenv ("TMP");
  }
  if (temp_dir == NIL_PTR || temp_dir [0] == '\0' || access (temp_dir, R_OK | X_OK) < 0) {
	temp_dir = getenv ("TEMP");
  }
  if (temp_dir == NIL_PTR || temp_dir [0] == '\0' || access (temp_dir, R_OK | X_OK) < 0) {
	temp_dir = "/usr/tmp";
  }
  if (access (temp_dir, R_OK | X_OK) < 0) {
	temp_dir = "/tmp";
  }
  if (getenv ("MINEDUSER")) {
	build_string (yankie_file, "%s/minedbuf.%s", temp_dir, getenv ("MINEDUSER"));
	build_string (panic_file, "%s/minedpanic.%s.%d", temp_dir, getenv ("MINEDUSER"), getpid ());
  } else if (getenv ("USER")) {
	build_string (yankie_file, "%s/minedbuf.%s", temp_dir, getenv ("USER"));
	build_string (panic_file, "%s/minedpanic.%s.%d", temp_dir, getenv ("USER"), getpid ());
  } else {
	build_string (yankie_file, "%s/minedbuf.%d", temp_dir, geteuid ());
	build_string (panic_file, "%s/minedpanic.%d.%d", temp_dir, geteuid (), getpid ());
  }
#endif
#ifdef vms
  if (getenv ("SYS$MINEDTMP")) {
	temp_dir = "SYS$MINEDTMP";
  } else if (getenv ("SYS$SCRATCH")) {
	temp_dir = "SYS$SCRATCH";
  } else {
  	temp_dir = "SYS$LOGIN";
  }
  if (getenv ("MINEDUSER")) {
	build_string (yankie_file, "%s:$MINEDBUF$%s", temp_dir, getenv ("MINEDUSER"));
	build_string (panic_file, "%s:$MINEDPANIC$%s.%d", temp_dir, getenv ("MINEDUSER"), getpid ());
  } else if (getenv ("USER")) {
	build_string (yankie_file, "%s:$MINEDBUF$%s", temp_dir, getenv ("USER"));
	build_string (panic_file, "%s:$MINEDPANIC$%s.%d", temp_dir, getenv ("USER"), getpid ());
  } else {
	build_string (yankie_file, "%s:$MINEDBUF$%d", temp_dir, geteuid ());
	build_string (panic_file, "%s:$MINEDPANIC$%d.%d", temp_dir, geteuid (), getpid ());
  }
#endif
#ifdef msdos
  temp_dir = (char *) getenv ("MINEDTMP");
  if (temp_dir == NIL_PTR || temp_dir [0] == '\0') temp_dir = (char *) getenv ("TEMP");
  if (temp_dir == NIL_PTR || temp_dir [0] == '\0') temp_dir = (char *) getenv ("TMP");
  if (temp_dir == NIL_PTR || temp_dir [0] == '\0') temp_dir = "\\";
  build_string (yankie_file, "%s\\minedbuf", temp_dir);
  build_string (panic_file, "%s\\mined-pa.nic", temp_dir);
#endif

  header = tail = alloc_header (); /* Make header of list */
  if (header == NIL_LINE) panic ("Cannot allocate memory", NIL_PTR);
  header->text = NIL_PTR;
  header->next = tail->prev = header;
  header->syntax_marker = syntax_none;

/* Load the file (if any) */
  if (fnami == 0) {
	load_file (NIL_PTR, False);
  } else {
	/* This should be applied to all file names, or better, not at all:
	if (length_of (argv [fnami]) > maxLINE_LEN) {
		argv [fnami] [maxLINE_LEN] = '\0';
	}
	*/
	load_wild_file (argv [fnami], False);
  }
  loading = True;	/* keep loading flag True until entering main loop */

  if (initlini != 0) {
     make_number (& initlinenum, argv [initlini] + 1);
     if (initlinenum > 0) {
	if (initlinenum <= 0 || (initline = proceed (header->next, initlinenum - 1)) == tail) {
	   error ("Illegal line number: ", num_out ((long) initlinenum));
	} else {
	   move_to (x, find_y_w_o_RD (initline));
	   fstatus ("Read", -1L, -1L);
	}
     }
  }

  if (inisearch != NIL_PTR) {
	search_for (inisearch, FORWARD);
	fstatus ("Read", -1L, -1L);
  }

  if (wpipe == True) {
	file_name [0] = '\0'; /* don't let user believe he's editing a file */
	fstatus ("Editing for standard output", -1L, -1L);
  }
  RD ();
  flush ();
  catch_signals (catch_interrupt);
  loading = False;

/* Main loop of the editor */
  for (;;) {
	if (fstat_always == True && stat_visible == False) {
		FSTAT ();
	} else if (always_disp_code == True && stat_visible == False) {
		display_the_code ();
	}

	if (MENU && top_line_scrolled == True) {
		displaymenuline ();
		flags_changed = False;
		cursor_somewhere = True;
	}
	if (MENU && flags_changed == True) {
		displayflags ();
		cursor_somewhere = True;
	}
	if (scrollbar == True) {
		if (display_scrollbar (update_scrollbar_lazy))
			cursor_somewhere = True;
	}
	if (cursor_somewhere == True) {
		set_cursor_xy ();
	}
	flush ();	/* Flush output (if any) */

	index = readchar ();

	if (stat_visible == True) {
		clear_status ();
	}
	if (quit == False) {	/* Call the function for the typed key */
		(* key_map [index]) (index);
		if (hop_flag > 0) {
			hop_flag --;
			flags_changed = True;
		}
	}
	if (quit == True) {
		if (hop_flag > 0) {
			flags_changed = True;
		}
		CANCEL ();
		quit = False;
	}
  }
  /* NOTREACHED */
}


/*======================================================================*\
|*				End					*|
\*======================================================================*/
