/*======================================================================*\
|*		Editor mined						*|
|*		part 2							*|
\*======================================================================*/

#include "mined.h"


/*======================================================================*\
|*			Definitions specific for mined2.c		*|
\*======================================================================*/

#ifdef unix
#define linkyank
#endif

#ifdef __pcgcc__
#undef linkyank
#endif
#ifdef __CYGWIN__
#define linkyank
#endif

/*
 * viewonlyerr () outputs an error message with a beep
 */
void
viewonlyerr ()
{
  ring_bell ();
  error ("View only mode", NIL_PTR);
}

/*
 * restrictederr () outputs an error message with a beep
 */
void
restrictederr ()
{
  ring_bell ();
  error ("Restricted mode", NIL_PTR);
}

FLAG yank_status = NOT_VALID;	/* Status of yank_file */
FLAG html_status = NOT_VALID;	/* Status of html_file */
FLAG redraw_pending = False;	/* was a redraw suppressed in find_y ? */


/*======================================================================*\
|*			Forward declarations				*|
\*======================================================================*/

void file_insert ();
void yank_text ();
int reverse_scroll ();
int forward_scroll ();
int legal ();
FLAG checkmark ();


/*======================================================================*\
|*			Move commands					*|
\*======================================================================*/

/*
 * Move one line up.
 */
void
MUP ()
{
  if (hop_flag > 0) HIGH ();
  else if (y == 0) {		/* Top line of screen. Scroll one line */
	if (reverse_scroll (True) != ERRORS) {
		move_y (y);
	}
  }
  else			/* Move to previous line */
	move_y (y - 1);
}

/*
 * Move one line down.
 */
void
MDN ()
{
  if (hop_flag > 0) LOW ();
  else if (y == last_y) {	/* Last line of screen. Scroll one line */
	if (bot_line->next == tail && bot_line->text [0] != '\n') {
		return;
	}
	else {
		(void) forward_scroll (True);
		move_y (y);
	}
  }
  else			/* Move to next line */
	move_y (y + 1);
}

/*
 * Move left one position.
 */
void
MLF ()
{
  if ((keyshift & ctrlshift_mask) == ctrlshift_mask) {
	keyshift = '0';
	BLINE ();
	return;
  }
  if (keyshift & ctrl_mask) {
	keyshift = '0';
	ctrl_MLF ();
	return;
  }
  if (keyshift & shift_mask) {
	keyshift = '0';
	MPW ();
	return;
  }

  if (hop_flag > 0) BLINE ();
  else if (x == 0 && get_shift (cur_line->shift_count) == 0) {/* Begin of line */
	if (cur_line->prev != header) {
		MUP ();					/* Move one line up */
		move_to (LINE_END, y);
	}
  }
  else
	move_to (x - 1, y);
}

/*
 * Move right one position.
 */
void
MRT ()
{
  if ((keyshift & ctrlshift_mask) == ctrlshift_mask) {
	keyshift = '0';
	ELINE ();
	return;
  }
  if (keyshift & ctrl_mask) {
	keyshift = '0';
	ctrl_MRT ();
	return;
  }
  if (keyshift & shift_mask) {
	keyshift = '0';
	MNW ();
	return;
  }

  if (hop_flag > 0) ELINE ();
  else if (* cur_text == '\n') {
	if (cur_line->next != tail) {		/* Last char of file */
		MDN ();				/* Move one line down */
		move_to (LINE_START, y);
	}
  }
  else
	move_to (x + 1, y);
}

/*
 * Move left one Unicode character (may enter into combined character).
 */
void
ctrl_MLF ()
{
  char * curpoi;

  if (hop_flag > 0) BLINE ();
  else if (cur_text == cur_line->text) {/* Begin of line */
	if (cur_line->prev != header) {
		MUP ();					/* Move one line up */
		move_to (LINE_END, y);
	}
  } else {
	curpoi = cur_text;
	precede_char (& curpoi, cur_line->text);
	move_address (curpoi, y);
  }
}

/*
 * Move right one Unicode character (may enter into combined character).
 */
void
ctrl_MRT ()
{
  char * curpoi;

  if (hop_flag > 0) ELINE ();
  else if (* cur_text == '\n') {
	if (cur_line->next != tail) {		/* Last char of file */
		MDN ();				/* Move one line down */
		move_to (LINE_START, y);
	}
  } else {
	curpoi = cur_text;
	advance_char (& curpoi);
	move_address (curpoi, y);
  }
}

/*
 * Move to top of screen
 */
void
HIGH ()
{
  move_y (0);
}

/*
 * Move to bottom of screen
 */
void
LOW ()
{
  move_y (last_y);
}

/*
 * Move to begin of line.
 */
void
BLINE ()
{
  move_to (LINE_START, y);
}

/*
 * Move to end of line.
 */
void
ELINE ()
{
  move_to (LINE_END, y);
}

/*
 * GOTO () prompts for a linenumber and moves to that line.
 */
void
goline (number)
  int number;
{
  LINE * line;
  if (number <= 0 || (line = proceed (header->next, number - 1)) == tail)
	error ("Illegal line number: ", num_out ((long) number));
  else	{
	  clear_status ();
	  move_y (find_y (line));
	}
}

void
goproz (number)
  int number;
{
  goline ((long) (total_lines - 1) * number / 100 + 1);
}

void
GOTO ()
{
  character c;
  char end;
  int number;

  if (MENU) {
	hop_flag = 1;
	displayflags ();
	set_cursor_xy ();
	flush ();
	hop_flag = 0;
  }
  if (! char_ready_within (500)) status_msg ("HOP: type command (fortified) or number...");
  if (quit == True) return;
  c = readchar ();
  if (quit == True) return;
  if ('0' <= c && c <= '9') {
	if (lines_per_page > 0)
		end = get_number ("...number, then go line / %(of lines) / p(age / m(ark / g(o marker / f(ile #", c, & number);
	else	end = get_number ("...number, then go line / %(of lines) / m(ark / g(o marker / f(ile #", c, & number);
	if (end == '%')
		goproz (number);
	else if (end == 'm' || end == 'M' || end == ',')
		MARKn (number);
	else if (end == '\'' || end == '.' || end == 'g' || end == 'G')
		GOMAn (number);
	else if (end == 'f' || end == 'F' || end == '#')
		edit_nth_file (number);
	else if (lines_per_page > 0 && (end == 'p' || end == 'P') && number > 0)
		goline (number * lines_per_page - lines_per_page + 1);
	else if (end != ERRORS)
		goline (number);
	return;
  }
  else {
	clear_status ();
	hop_flag = 1;
	(* key_map [c]) (c);
	return;
  }
}

/*
 * Scroll forward one page or to eof, whatever comes first. (Bot_line becomes
 * top_line of display.) Try to leave the cursor on the same line. If this is
 * not possible, leave cursor on the line halfway the page.
 */
void
PD ()
{
  register int i;
  int new_y;

  if (keyshift & ctrl_mask) {
	keyshift = '0';
	SD ();
	return;
  }

  if (hop_flag > 0) {
	hop_flag = 0;
	EFILE ();
	return;
  }

  for (i = 0; i < SCREENMAX; i ++)
	if (forward_scroll (page_scroll) == ERRORS)
		break;			/* EOF reached */

  if (y - i < 0)			/* Line no longer on screen */
	new_y = (page_stay == True) ? 0 : SCREENMAX >> 1;
  else	new_y = y - i;

  if (page_scroll == False) display (0, top_line, last_y, new_y);

  move_y (new_y);
}

/*
 * Scroll backwards one page or to top of file, whatever comes first.
 * (Top_line becomes bot_line of display).
 * The very bottom line (YMAX) is always blank.
 * Try to leave the cursor on the same line.
 * If this is not possible, leave cursor on the line halfway the page.
 */
void
PU ()
{
  register int i;
  int new_y;

  if (keyshift & ctrl_mask) {
	keyshift = '0';
	SU ();
	return;
  }

  if (hop_flag > 0) {
	hop_flag = 0;
	BFILE ();
	return;
  }

  for (i = 0; i < SCREENMAX; i ++) {
	if (reverse_scroll (page_scroll) == ERRORS)
		/* should also flag reverse_scroll that clearing of 
		   bottom line is not desired */
		break;			/* Top of file reached */
  }

  if (y + i > SCREENMAX)		/* line no longer on screen */
	new_y = (page_stay == True) ? last_y : SCREENMAX >> 1;
  else
	new_y = y + i;

  if (can_scroll_reverse == True && page_scroll == True) {
	set_cursor (0, YMAX);	/* Erase very bottom line */
	clear_lastline ();
  }
  else display (0, top_line, last_y, new_y);

  move_y (new_y);
}

/*
 * Go to top of file, scrolling if possible, else redrawing screen.
 */
void
BFILE ()
{
  if (proceed (top_line, - SCREENMAX) == header) {
	PU ();			/* It fits. Let PU do it */
  } else {
	reset (header->next, 0);/* Reset top_line, etc. */
	RD_y (0);		/* Display full page */
  }
  move_to (LINE_START, 0);
}

/*
 * Go to last position of text, scrolling if possible, else redrawing screen
 */
void
EFILE ()
{
  if (proceed (bot_line, SCREENMAX) == tail) {
	PD ();			/* It fits. Let PD do it */
  } else {
	reset (proceed (tail->prev, - SCREENMAX), SCREENMAX);
	RD_y (last_y);		/* Display full page */
  }
  move_to (LINE_END, last_y);
}

/*
 * Scroll one line up. Leave the cursor on the same line (if possible).
 */
void
SU ()
{
  register int i;

  if (hop_flag > 0) {
	hop_flag = 0;
	for (i = 0; i < (SCREENMAX >> 1); i ++) SU ();
	return;
  }

  if (reverse_scroll (True) != ERRORS) {	/* else we are at top of file */
	move_y ((y == SCREENMAX) ? SCREENMAX : y + 1);
  }
}

/*
 * Scroll one line down. Leave the cursor on the same line (if possible).
 */
void
SD ()
{
  register int i;

  if (hop_flag > 0) {
	hop_flag = 0;
	for (i = 0; i < (SCREENMAX >> 1); i ++) SD ();
	return;
  }

  if (forward_scroll (True) != ERRORS)
	move_y ((y == 0) ? 0 : y - 1);
}

/*
 * Perform a forward scroll. It returns ERRORS if we're at the last line of
 * the file.
 */
int
forward_scroll (update)
  FLAG update;
{
  if (bot_line->next == tail)		/* Last line of file. No dice */
	return ERRORS;
  top_line = top_line->next;
  bot_line = bot_line->next;
  cur_line = cur_line->next;
  line_number ++;

/* Perform the scroll on screen */
  if (update == True) {
	clean_menus ();
	scroll_forward ();
	set_cursor (0, SCREENMAX);
	line_print (bot_line);
  }

  return FINE;
}

/*
 * Perform a backwards scroll. It returns ERRORS if we're at the first line 
 * of the file. It updates the display completely if update is True. 
 * Otherwise it leaves that to the caller (page up function).
 */
int
reverse_scroll (update)
  FLAG update;
{
  if (top_line->prev == header)
	return ERRORS;		/* Top of file. Can't scroll */

  if (last_y != SCREENMAX)	/* Reset last_y if necessary */
	last_y ++;
  else
	bot_line = bot_line->prev;	/* Else adjust bot_line */
  top_line = top_line->prev;
  cur_line = cur_line->prev;
  line_number --;

/* Perform the scroll on screen */
  if (update == True) {
    if (can_scroll_reverse == True) {
	clean_menus ();
	set_cursor (0, 0);
	scroll_reverse ();
	set_cursor (0, YMAX);	/* Erase very bottom line */
	clear_lastline ();
	set_cursor (0, 0);
	line_print (top_line);
    }
    else display (0, top_line, last_y, y);
  }

  return FINE;
}

/*----------------------*
 *	Word moves	*
 *----------------------*/

/*
 * A word was previously defined as a number of non-blank characters
 * separated by tabs, spaces or linefeeds.
 * By consulting idfchar (), sequences of real letters only or digits
 * or underlines are recognized as words.
 */
extern int idfchar ();

/*
 * BSEN () and ESEN () look for the beginning or end of the current sentence.
 */
void
BSEN ()
{
  search_for ("[;.]", REVERSE);
}

void
ESEN ()
{
  search_for ("[;.]", FORWARD);
}

void
search_tag (poi)
  char * poi;
{
  char pat [maxLINE_LEN];
  FLAG direction = FORWARD;
  char * patpoi = & pat [3];

  strcpy (pat, "</*");
  poi ++;
  if (* poi == '/') {
	direction = REVERSE;
	poi ++;
  }
  while (! white_space (* poi) && * poi != '\n' && * poi != '>' && * poi != '\0') {
	* patpoi = * poi;
	patpoi ++;
	poi ++;
  }
  * patpoi = '\0';

  search_corresponding (pat, direction, '/');
}

/*
 * SCORR () looks for a corresponding bracket ([<->] etc.)
 */
void
SCORR ()
{
  char * poi;

  switch (* cur_text) {
    case '(':	search_corresponding ("[()]", FORWARD, ')'); return;
    case ')':	search_corresponding ("[()]", REVERSE, '('); return;
    case '[':	search_corresponding ("[\\[\\]]", FORWARD, ']'); return;
    case ']':	search_corresponding ("[\\[\\]]", REVERSE, '['); return;
    case '{':	search_corresponding ("[{}]", FORWARD, '}'); return;
    case '}':	search_corresponding ("[{}]", REVERSE, '{'); return;
    case '':	search_corresponding ("[]", FORWARD, ''); return;
    case '':	search_corresponding ("[]", REVERSE, ''); return;
    default:	if (dim_HTML == True) {
			poi = cur_text;
			if (* poi == '>') {
				MLF ();
			}
			while (poi != cur_line->text && * poi != '<') {
				precede_char (& poi, cur_line->text);
			}
			if (* poi == '<') {
				search_tag (poi);
			} else {
				error ("No bracket or tag to match", NIL_PTR); return;
			}
		} else {
			if (* cur_text == '<') {
				search_corresponding ("[<>]", FORWARD, '>'); return;
			} else if (* cur_text == '>') {
				search_corresponding ("[<>]", REVERSE, '<');
			} else {
				error ("No bracket to match", NIL_PTR); return;
			}
		}
  }
}

int
get_idf (idf_buf, text, start)
  char * idf_buf;
  char * text;
  char * start;
{
  char * idf_buf_poi = idf_buf;
  char * idf_poi;

  if (! idfchar (* text)) {
	error ("No identifier", NIL_PTR);
	return ERRORS;
  } else {
	idf_poi = text;
	while (idfchar (* idf_poi) && idf_poi != start) {
		idf_poi --;
	}
	if (! idfchar (* idf_poi)) idf_poi ++;
	while (idfchar (* idf_poi)) {
		* idf_buf_poi ++ = * idf_poi ++;
	}
	* idf_buf_poi = '\0';
	return FINE;
  }
}

/*
 * SIDF () searches for the identifier at the current position
 */
void
SIDF (method)
  FLAG method;
{
  char idf_buf [MAX_CHARS];	/* identifier to search for */

  int ret = get_idf (idf_buf, cur_text, cur_line->text);
  if (ret == ERRORS)
	return;

  search_expr (idf_buf, method);
}

/*
 * MPW () moves to the start of the previous word. A word is defined as a
 * number of non-blank characters separated by tabs spaces or linefeeds.
 */
void
move_previous_word (remove)
  FLAG remove;
{
  register char * begin_line;
  char * textp;
  char start_char = * cur_text;
  char * start_pos = cur_text;
  FLAG idfsearch;

  if (remove == DELETE && viewonly == True)
	{viewonlyerr (); return;}

/* First check if we're at the beginning of line. */
  if (cur_text == cur_line->text) {
	if (cur_line->prev == header)
		return;
	start_char = '\0';
  }

  MLF ();

  begin_line = cur_line->text;
  textp = cur_text;

/* Check if we're in the middle of a word. */
  if (! alpha (* textp) || ! alpha (start_char)) {
	while (textp != begin_line 
		&& (white_space (* textp) || * textp == '\n'))
	{
		precede_char (& textp, begin_line);
	}
  }

/* Now we're at the end of previous word. Skip non-blanks until a blank comes */
  if (wordnonblank == True) {
	while (textp != begin_line && alpha (* textp)) {
		precede_char (& textp, begin_line);
	}
  } else {
	if (idfchar (* textp)) {
		idfsearch = True;
		while (textp != begin_line && idfchar (* textp)) {
			precede_char (& textp, begin_line);
		}
	}
	else {
		idfsearch = False;
		while (textp != begin_line 
			&& alpha (* textp) && ! idfchar (* textp))
		{
			precede_char (& textp, begin_line);
		}
	}
  }

/* Go to the next char if we're not at the beginning of the line */
/* At the beginning of the line, check whether to stay or to go to the word */
  if (textp != begin_line && * textp != '\n')
	advance_char (& textp);
  else if (textp == begin_line && * textp != '\n' &&
	   ((wordnonblank == True) ? * textp == ' '
		:   ((idfsearch == True) ? ! idfchar (* textp)
			: (! alpha (* textp) || idfchar (* textp))))) {
	advance_char (& textp);
	if (white_space (* textp) || textp == start_pos)
		/* no word there or not moved, so go back */
		precede_char (& textp, begin_line);
  }

/* Find the x-coordinate of this address, and move to it */
  move_address (textp, y);
  if (remove == DELETE)
	(void) delete_text (cur_line, textp, cur_line, start_pos);
}

void
MPW ()
{
  if (hop_flag > 0) BSEN ();
  else move_previous_word (NO_DELETE);
}

/*
 * MNW () moves to the start of the next word. A word is defined as a number of
 * non-blank characters separated by tabs spaces or linefeeds. Always keep in
 * mind that the pointer shouldn't pass the '\n'.
 */
void
move_next_word (remove)
  FLAG remove;
{
  char * textp = cur_text;

  if (remove == DELETE && viewonly == True)
	{viewonlyerr (); return;}

/* Move to the end of the current word. */
  if (wordnonblank == True) {
	if (* textp != '\n')
		advance_char (& textp);
	while (alpha (* textp)) {
		advance_char (& textp);
	}
  } else {
	if (idfchar (* textp))
		while (* textp != '\n' && idfchar (* textp)) {
			advance_char (& textp);
		}
	else
		while (alpha (* textp) && ! idfchar (* textp)) {
			advance_char (& textp);
		}
  }

/* Skip all white spaces */
  while (* textp != '\n' && white_space (* textp)) {
	textp ++;
  }
/* If we're deleting, delete the text in between */
  if (remove == DELETE) {
	(void) delete_text (cur_line, cur_text, cur_line, textp);
	return;
  }

/* If we're at end of line, move to the beginning of (first word on) the next line */
  if (* textp == '\n' && cur_line->next != tail) {
	MDN ();
	move_to (LINE_START, y);
	textp = cur_text;
/*
	while (* textp != '\n' && white_space (* textp)) {
		textp ++;
	}
*/
  }
  move_address (textp, y);
}

void
MNW ()
{
  if (hop_flag > 0) ESEN ();
  else move_next_word (NO_DELETE);
}

/*
 * find_y () checks if the matched line is on the current page. If it is, it
 * returns the new y coordinate, else it displays the correct page with the
 * matched line in the middle (unless redrawflag is set to False) 
 * and returns the new y value.
 */
int
find_y_RD (match_line, redrawflag)
  LINE * match_line;
  FLAG redrawflag;
{
  register LINE * line;
  register int count = 0;

/* Check if match_line is on the same page as currently displayed. */
  for (line = top_line; line != match_line && line != bot_line->next;
						      line = line->next)
	count ++;
  if (line != bot_line->next)
	return count;

/* Display new page, with match_line in center. */
  if ((line = proceed (match_line, - (SCREENMAX >> 1))) == header) {
  /* Can't display in the middle. Make first line of file top_line */
	count = 0;
	for (line = header->next; line != match_line; line = line->next)
		count ++;
	line = header->next;
  }
  else	/* New page is displayed. Set cursor to middle of page */
	count = SCREENMAX >> 1;

/* Reset pointers and redraw the screen */
  reset (line, 0);
  if (redrawflag == True) {
	RD_y (count);
	redraw_pending = False;
  }
  else	redraw_pending = True;

  return count;
}

int
find_y (match_line)
  LINE * match_line;
{
  return find_y_RD (match_line, True);
}

int
find_y_w_o_RD (match_line)
  LINE * match_line;
{
  return find_y_RD (match_line, False);
}


/*======================================================================*\
|*			Modify commands					*|
\*======================================================================*/

/*
 * DCC deletes the character under the cursor. If this character is a '\n' the
 * current line is joined with the next one.
 * If this character is the only character of the line, the current line will
 * be deleted.
 * DCC0 deletes without justification.
 */
void
delete_char (with_combinings)
  FLAG with_combinings;
{
  char * after_char;
  unsigned long unichar;
  int follow;

  if (* cur_text == '\n') {
	if (cur_line->next == tail) {
		if (cur_line->return_type != lineend_NONE) {
			set_modified ();
			cur_line->return_type = lineend_NONE;
			set_cursor_xy ();
			put_line (cur_line, x, True, False);
			status_msg ("Trailing line-end deleted");
		}
	} else {
		(void) delete_text (cur_line, cur_text, cur_line->next, cur_line->next->text);
	}
  } else {
	after_char = cur_text;

	if (utf8_text == True && combining_mode == True) {
		utf8_info (after_char, & follow, & unichar);
		advance_char (& after_char);
		if (with_combinings == True && ! iscombining (unichar)) {
		/* delete combining accents together with base character */
			utf8_info (after_char, & follow, & unichar);
			while (iscombining (unichar)) {
				advance_char (& after_char);
				utf8_info (after_char, & follow, & unichar);
			}
		}
	} else {
		advance_char (& after_char);
	}

	(void) delete_text (cur_line, cur_text, cur_line, after_char);
  }
}
void
DCC ()
{
  delete_char (True);
}
void
DCC0 ()
{
  DCC ();
}

/*
   DPC0 deletes the character on the left side of the cursor.
   If the cursor is at the beginning of the line, the last character 
   if the previous line is deleted. With hop flag, delete left part of 
   line from current point.
 */
void
DPC0 ()
{
  char * delete_pos;

  if (x == 0 && cur_line->prev == header)
	return;			/* Top of file */

  if (viewonly == True)
	{viewonlyerr (); return;}

  if (hop_flag > 0) {
	hop_flag = 0;
	if (cur_text != cur_line->text) {
	  delete_pos = cur_text;
	  BLINE ();
	  (void) delete_text (cur_line, cur_line->text, cur_line, delete_pos);
	}
  }
  else {
	MLF ();			/* Move one left */
	DCC ();			/* Delete character under cursor */
  }
}

/*
   DPC normally deletes the character left just as DPC0 does.
   However, unless the hop flag is set, it first checks if there is 
   anything but white space on the current line left of the current position.
   If there is only white space, it tries to perform a "backtab" function, 
   reverting the indentation to the previous amount above in the text 
   (unless if it's in the line immediately above the current line).
 */
void
DPC ()
{
  char * cp;
  int column;
  int previous_col;
  LINE * lp;

  if (hop_flag > 0) {
	DPC0 ();
  } else {
	cp = cur_line->text;
	column = 0;
	while (* cp != '\0' && cp != cur_text && white_space (* cp)) {
		advance_char_col (& cp, & column);
	}
	if (cp == cur_text) {
		/* only white space left of current position */
		previous_col = column;
		lp = cur_line->prev;
		while (previous_col >= column && lp != header) {
			/* count white space on line lp */
			cp = lp->text;
			previous_col = 0;
			while (* cp != '\0' && previous_col < column && white_space (* cp)) {
				advance_char_col (& cp, & previous_col);
			}
			if (* cp == '\n' || * cp == '\0') {
				/* don't count space lines */
				previous_col = column;
			}

			lp = lp->prev;
		}

		/* if less indented previous line was found, 
		   and this was not on the line immediately 
		   preceeding the current line,
		   perform the back TAB function */
		if (previous_col < column && cur_line->prev != lp->next) {
			while (column > previous_col) {
				DPC0 ();
				column = 0;
				cp = cur_line->text;
				while (* cp != '\0' && cp != cur_text) {
					advance_char_col (& cp, & column);
				}
			}
			while (column < previous_col) {
				S (' ');
				column ++;
			}
		} else {
			DPC0 ();
		}
	} else {
		DPC0 ();
	}
  }
}

/*
 * DLINE delete the whole current line.
 */
void
DLINE ()
{
  if (viewonly == True)
	{viewonlyerr (); return;}

  if (hop_flag > 0) {
    hop_flag = 0;
    if (* cur_text != '\n')
	(void) delete_text (cur_line, cur_text, cur_line, cur_text + length_of (cur_text) - 1);
  }
  else {
/*
    if (cur_line->next == tail)
	(void) delete_text (cur_line, cur_line->text, cur_line, cur_text + length_of (cur_text) - 1);
    else
	(void) delete_text (cur_line, cur_line->text, cur_line->next, cur_line->next->text);
    BLINE ();
*/
    BLINE ();
    if (* cur_text != '\n')
	DLN ();
    DCC ();
  }
}

/*
 * DLN deletes all characters until the end of the line. If the current
 * character is a '\n', then delete that char.
 */
void
DLN ()
{
  if (hop_flag > 0) {
	hop_flag = 0;
	DLINE ();
  }
  else if (* cur_text == '\n')
	DCC ();
  else	(void) delete_text (cur_line, cur_text, cur_line, cur_text + length_of (cur_text) - 1);
}

/*
 * DNW () deletes the next word (as defined in MNW ())
 */
void
DNW ()
{
  if (* cur_text == '\n')
	DCC ();
  else
	move_next_word (DELETE);
}

/*
 * DPW () deletes the previous word (as defined in MPW ())
 */
void
DPW ()
{
  if (cur_text == cur_line->text)
	DPC0 ();
  else
	move_previous_word (DELETE);
}

void
enterNUL ()
{
  if (viewonly == True)
	{viewonlyerr (); return;}

  S ('\n');				/* Insert a new line */
  MUP ();				/* Move one line up */
  move_to (LINE_END, y);		/* Move to end of this line */
  cur_line->return_type = lineend_NUL;
  put_line (cur_line, x, True, False);
  MRT ();				/* move behind inserted NUL */
}


/*
 * Functions to insert character at current location.
 * S0 inserts without justification.
 */

static unsigned long previous_unichar = 0;

/*
 * enter a character or a byte of a multi-byte character 
 * (with intermediate storage)
 */
static
void
S1 (newchar, JUSlvl, utf8_transform)
  register character newchar;
  int JUSlvl;
  FLAG utf8_transform;
{
  static character buffer [7];
  static character * utfpoi = buffer;
  static int utfcount = 1;
  static character firstbyte;
  int width = 1;
  static unsigned long unichar;
  int newx;

  if (newchar == '\0') {
	enterNUL ();
	return;
  }

  if (utf8_text == True) {
	if (utf8_transform == True && last_char_latin1 == False) {
	/* UTF-8 input for UTF-8 text */
		if (newchar < 0x80) {
			unichar = newchar;
			* utfpoi = newchar;
			utfpoi ++;
			* utfpoi = '\0';
			utfpoi = buffer;
		} else if ((newchar & 0xC0) == 0x80) {
		/* UTF-8 sequence byte */
			unichar = (unichar << 6) | (newchar & 0x3F);
			* utfpoi = newchar;
			utfpoi ++;
			utfcount --;
			if (utfcount == 0) {
				* utfpoi = '\0';
				utfpoi = buffer;
				width = uniwidth (unichar);
			}
			else
				return;
		} else { /* first UTF-8 byte */
			utfpoi = buffer;
			* utfpoi = newchar;

			if ((newchar & 0xE0) == 0xC0) {
				utfcount = 2;
				unichar = newchar & 0x1F;
			} else if ((newchar & 0xF0) == 0xE0) {
				utfcount = 3;
				unichar = newchar & 0x0F;
			} else if ((newchar & 0xF8) == 0xF0) {
				utfcount = 4;
				unichar = newchar & 0x07;
			} else if ((newchar & 0xFC) == 0xF8) {
				utfcount = 5;
				unichar = newchar & 0x03;
			} else if ((newchar & 0xFE) == 0xFC) {
				utfcount = 6;
				unichar = newchar & 0x01;
			} else /* ignore illegal UTF-8 code */
				return;

			utfpoi ++;
			utfcount --;
			return;
		}
	} else {
	/* 8-bit input for UTF-8 text */
		unichar = newchar;
		if (newchar < 0x80) {
			buffer [0] = newchar;
			buffer [1] = '\0';
		} else {
			buffer [0] = (newchar >> 6) | 0xC0;
			buffer [1] = (newchar & 0x3F) | 0x80;
			buffer [2] = '\0';
		}
	}
  } else if (Chinese == True) {
	/* 8/16-bit (Chinese) input for Chinese text */
	if (firstbyte != '\0') {
		buffer [0] = firstbyte;
		buffer [1] = newchar;
		width = 2;
	} else if (multichar (newchar)) {
		firstbyte = newchar;
		return;
	} else {
		buffer [0] = newchar;
		buffer [1] = '\0';
		width = 1;
	}
	firstbyte = '\0';
  } else if (utf8_transform == True && last_char_latin1 == False) {
	/* UTF-8 input for 8-bit text */
	buffer [1] = '\0';
	if (newchar < 0x80) {
		buffer [0] = newchar;
	} else if ((newchar & 0xC0) == 0x80) {
	/* UTF-8 sequence byte */
		unichar = (unichar << 6) | (newchar & 0x3F);
		utfcount --;
		if (utfcount == 0) {
			if ((unichar & 0xFF) == unichar)
				buffer [0] = unichar & 0xFF;
			else
				buffer [0] = '';
		}
		else {
			return;
		}
	} else { /* first UTF-8 byte */
		if ((newchar & 0xE0) == 0xC0) {
			utfcount = 2;
			unichar = newchar & 0x1F;
		} else if ((newchar & 0xF0) == 0xE0) {
			utfcount = 3;
			unichar = newchar & 0x0F;
		} else if ((newchar & 0xF8) == 0xF0) {
			utfcount = 4;
			unichar = newchar & 0x07;
		} else if ((newchar & 0xFC) == 0xF8) {
			utfcount = 5;
			unichar = newchar & 0x03;
		} else if ((newchar & 0xFE) == 0xFC) {
			utfcount = 6;
			unichar = newchar & 0x01;
		} else { /* ignore illegal UTF-8 code */
			return;
		}
		utfcount --;
		return;
	}
  } else {
	/* 8-bit input for 8-bit text */
	buffer [0] = newchar;
	buffer [1] = '\0';
  }

/* right-to-left support */
  if (bidi == True && utf8_text == True && is_right_to_left (previous_unichar)) {
	if (newchar == '\n') {
		ELINE ();
	} else if (iscombining (unichar) && * cur_text != '\n') {
		MRT ();
	}
  }


/* Insert the new character */
  if (insert (cur_line, cur_text, buffer) == ERRORS) {
	return;
  }

/* Fix screen */
  if (newchar == '\n') {
	set_cursor (0, y);
	if (y == SCREENMAX) {		/* Can't use display () */
		line_print (cur_line);
		(void) forward_scroll (True);
		move_to (0, y);
	}
	else {
		reset (top_line, y);	/* Reset pointers */
		if (can_add_line == True) {
			clean_menus ();
			add_line (y + 1);
			clear_status ();
			display (y, cur_line, 1, y + 1);
		}
		else	display (y, cur_line, last_y - y, y + 1);
		move_to (0, y + 1);
	}
  }
  else if (x + width == XBREAK) { /* If line must be shifted, just call move_to */
	move_to (x + width, y);
  }
  else {			/* else display rest of line */
	if (width == 0) {
		/* combining char added to left char */
		newx = x + width;
		move_to (x - 1, y);
		put_line (cur_line, x, False, False);
		move_to (newx, y);
	} else {
		put_line (cur_line, x, False, False);
		move_to (x + width, y);
	}
  }

/* right-to-left support */
  if (bidi == True && utf8_text == True) {
	if (unichar == ' ' || unichar == '\t') {
		if (is_right_to_left (previous_unichar)) {
			move_to (x - 1, y);
		}
	} else if (unichar != '\n') {
		if (iscombining (unichar)) {
			if (is_right_to_left (previous_unichar)) {
				move_to (x - 1, y);
			}
		} else {
			if (is_right_to_left (unichar)) {
				move_to (x - 1, y);
			}
			previous_unichar = unichar;
		}
	}
  }

  if (JUSlvl > 0) {
	JUSandreturn ();
  }
}

/*
 * Insert newline with auto indentation.
 */
void
SNLindent ()
{
  char * coptext;

  S ('\n');
  coptext = cur_line->prev->text;

  while (* coptext == ' ' || * coptext == '\t') {
	S (* coptext);
	coptext ++;
  }
}

/*
 * Insert new line at current location.
 */
void
SNL ()
{
  if (autoindent == False 
      || last_delta_readchar < 10 || average_delta_readchar < 10) {
	S ('\n');
  } else {
	SNLindent ();
  }
}

void
Spair (l, r)
  character l;
  character r;
{
  S1 (l, JUSlevel, utf8_input);
  SNLindent ();
  S1 (r, JUSlevel, utf8_input);

  MUP ();
  ELINE ();
}

void
S (newchar)
  register character newchar;
{
  if (hop_flag > 0) {
	hop_flag = 0;
	flags_changed = True;
	if (newchar == '(') {
		Spair ('(', ')');
		return;
	}
	else if (newchar == '[') {
		Spair ('[', ']');
		return;
	}
	else if (newchar == '{') {
		Spair ('{', '}');
		return;
	}
	else if (newchar == '<') {
		Spair ('<', '>');
		return;
	}
	else if (newchar == '/') {
		if (* cur_text != '\n') {
			SNLindent ();
			MUP ();
		}
		S1 ('/', JUSlevel, utf8_input);
		S1 ('*', JUSlevel, utf8_input);
		S1 ('*', JUSlevel, utf8_input);
		SNLindent ();
		S1 (' ', JUSlevel, utf8_input);
		SNLindent ();
		S1 ('*', JUSlevel, utf8_input);
		S1 ('/', JUSlevel, utf8_input);
		MUP ();
		S1 (' ', JUSlevel, utf8_input);
		S1 (' ', JUSlevel, utf8_input);
		return;
	}
  }

  S1 (newchar, JUSlevel, utf8_input);
}

void
Suni1 (newchar)
  register character newchar;
{
  S1 (newchar, JUSlevel, False);
}

void
Sutf8 (newchar)
  unsigned long newchar;
{
  if (newchar < 0x80) {
	S1 ((character) newchar, JUSlevel, True);
  } else if (newchar < 0x800) {
	S1 ((character) (0xC0 | (newchar >> 6)), JUSlevel, True);
	S1 ((character) (0x80 | (newchar & 0x3F)), JUSlevel, True);
  } else if (newchar < 0x10000) {
	S1 ((character) (0xE0 | (newchar >> 12)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 6) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | (newchar & 0x3F)), JUSlevel, True);
  } else if (newchar < 0x200000) {
	S1 ((character) (0xF0 | (newchar >> 18)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 12) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 6) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | (newchar & 0x3F)), JUSlevel, True);
  } else if (newchar < 0x4000000) {
	S1 ((character) (0xF8 | (newchar >> 24)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 18) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 12) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 6) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | (newchar & 0x3F)), JUSlevel, True);
  } else if (newchar < 0x80000000) {
	S1 ((character) (0xFC | (newchar >> 30)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 24) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 18) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 12) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | ((newchar >> 6) & 0x3F)), JUSlevel, True);
	S1 ((character) (0x80 | (newchar & 0x3F)), JUSlevel, True);
  } else {
	error ("Invalid Unicode value", NIL_PTR);
  }
}

void
S0 (newchar)
  register character newchar;
{
  S1 (newchar, 0, utf8_input);
}


#define arrlen(arr)	(sizeof (arr) / sizeof (* arr))

/* Keep this consistent with mousemen.c: itemtype Quotemenu [] ! */
quoteset quotes [] = {
/*
	“ 201C; LEFT DOUBLE QUOTATION MARK
		DOUBLE TURNED COMMA QUOTATION MARK
	” 201D; RIGHT DOUBLE QUOTATION MARK
		DOUBLE COMMA QUOTATION MARK
	„ 201E; DOUBLE LOW-9 QUOTATION MARK
		LOW DOUBLE COMMA QUOTATION MARK
	‟ 201F; DOUBLE HIGH-REVERSED-9 QUOTATION MARK
		DOUBLE REVERSED COMMA QUOTATION MARK
	‘ 2018; LEFT SINGLE QUOTATION MARK
		SINGLE TURNED COMMA QUOTATION MARK
	’ 2019; RIGHT SINGLE QUOTATION MARK
		SINGLE COMMA QUOTATION MARK
	‚ 201A; SINGLE LOW-9 QUOTATION MARK
		LOW SINGLE COMMA QUOTATION MARK
	‛ 201B; SINGLE HIGH-REVERSED-9 QUOTATION MARK
		SINGLE REVERSED COMMA QUOTATION MARK
	« 00AB; LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
		LEFT POINTING GUILLEMET
	» 00BB; RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
		RIGHT POINTING GUILLEMET
	‹ 2039; SINGLE LEFT-POINTING ANGLE QUOTATION MARK
		LEFT POINTING SINGLE GUILLEMET
	› 203A; SINGLE RIGHT-POINTING ANGLE QUOTATION MARK
		RIGHT POINTING SINGLE GUILLEMET
*/
	{0x0022, 0x0022, 0x0027, 0x0027, "\"", "\""},	/* ASCII ""'' */
	{0x201C, 0x201D, 0x2018, 0x2019, "“", "”"},	/* English “”‘’ */
	{0x201E, 0x201C, 0x201A, 0x2018, "„", "“"},	/* German „“‚‘ */
	{0x201D, 0x201D, 0x2019, 0x2019, "”", "”"},	/* ””’’ */
	{0x00AB, 0x00BB, 0x2039, 0x203A, "«", "»"},	/* French «»‹› */
	{0x00BB, 0x00AB, 0x203A, 0x2039, "»", "«"},	/* »«›‹ */
	{0x201E, 0x201D, 0x201A, 0x2019, "„", "”"},	/* „”‚’ */
	{0x00BB, 0x00BB, 0x203A, 0x203A, "»", "»"},	/* »»›› */
};
int quote_type = 0;

void
quote_type_up ()
{
  if (quote_type < arrlen (quotes) - 1) {
	quote_type ++;
  } else {
	quote_type = 0;
  }
}
void
quote_type_down ()
{
  if (quote_type > 0) {
	quote_type --;
  } else {
	quote_type = arrlen (quotes) - 1;
  }
}
void
set_quote_type (qt)
{
  if (qt >= 0 && qt < arrlen (quotes)) {
	quote_type = qt;
  }
}

void
Squote (doublequote)
  FLAG doublequote;
{
  unsigned long prevchar;

  if (utf8_text == True && smart_quotes == True) {
	prevchar = precedingchar (cur_text, cur_line->text);
	if (prevchar == (unsigned char) '('
	 || prevchar == (unsigned char) '['
	 || prevchar == (unsigned char) '{'
	 || prevchar == (unsigned char) ' '
	 || prevchar == (unsigned char) '	'
	 || prevchar == (unsigned char) '\n'
	 || prevchar == quotes [quote_type].leftdouble
	 || prevchar == quotes [quote_type].leftsingle
	) {
		/* insert left quote */
		if (doublequote == True) {
			Sutf8 (quotes [quote_type].leftdouble);
		} else {
			Sutf8 (quotes [quote_type].leftsingle);
		}
	} else {
		/* insert right quote */
		if (doublequote == True) {
			Sutf8 (quotes [quote_type].rightdouble);
		} else {
			Sutf8 (quotes [quote_type].rightsingle);
		}
	}
  } else {
	if (doublequote == True) {
		S ('"');
	} else {
		S ('\'');
	}
  }
}
void
Sdoublequote ()
{
  Squote (True);
}
void
Ssinglequote ()
{
  Squote (False);
}

/*
 * Replace current character with its hex/octal/decimal representation.
 */
character
hexdig (c)
  character c;
{
  if (c < 10) {
	return c + '0';
  } else {
	return c - 10 + 'A';
  }
}

void
insertcode (c, radix)
  character c;
  int radix;
{
  int radix2;

  if (radix == 8) {
	S (hexdig ((c >> 6) & 007));
	S (hexdig ((c >> 3) & 007));
	S (hexdig ((c) & 007));
  } else if (radix == 16) {
	S (hexdig ((c >> 4) & 017));
	S (hexdig ((c) & 017));
  } else {	/* assume radix = 10 or, at least, three digits suffice */
	radix2 = radix * radix;
	S (hexdig (c / radix2));
	S (hexdig ((c % radix2) / radix));
	S (hexdig (c % radix));
  }
}
void
changetocode (radix, unicode)
  int radix;
  FLAG unicode;
{
  character c = * cur_text;
  int utfcount = 1;
  unsigned long unichar;
#ifdef endlessloop
  int follow;
#endif
  char buffer [7];
  char * textpoi;
  char * utfpoi;

  if (c == '\n') {
#ifdef msdos
	insertcode ('\r', radix);
#endif
	insertcode ('\n', radix);
  } else if (utf8_text == True) {
	if ((c & 0x80) == 0x00) {
		utfcount = 1;
		unichar = c;
	} else if ((c & 0xE0) == 0xC0) {
		utfcount = 2;
		unichar = c & 0x1F;
	} else if ((c & 0xF0) == 0xE0) {
		utfcount = 3;
		unichar = c & 0x0F;
	} else if ((c & 0xF8) == 0xF0) {
		utfcount = 4;
		unichar = c & 0x07;
	} else if ((c & 0xFC) == 0xF8) {
		utfcount = 5;
		unichar = c & 0x03;
	} else if ((c & 0xFE) == 0xFC) {
		utfcount = 6;
		unichar = c & 0x01;
	} else /* illegal UTF-8 code */ {
	/*	DCC ();	*/
		insertcode (c, radix);
		error ("Invalid UTF-8 sequence", NIL_PTR);
		return;
	}
	utfcount --;
	utfpoi = buffer;
	* utfpoi ++ = c;
	textpoi = cur_text;
	textpoi ++;
	while (utfcount > 0 && (* textpoi & 0xC0) == 0x80) {
		c = * textpoi ++;
		* utfpoi ++ = c;
		unichar = (unichar << 6) | (c & 0x3F);
		utfcount --;
	}
	* utfpoi = '\0';
/*	delete_char (False);	*/
	if (unicode == True) {
		if (unichar > 0xFFFF) {
			insertcode (unichar >> 24, radix);
			insertcode (unichar >> 16, radix);
		}
		insertcode (unichar >> 8, radix);
		insertcode (unichar, radix);
	} else {
		utfpoi = buffer;
		while (* utfpoi != '\0') {
			insertcode (* utfpoi ++, radix);
		}
	}

#ifdef endlessloop
	utf8_info (cur_text, & follow, & unichar);
	if (iscombining (unichar)) {
		changetocode (radix, unicode);
	}
#endif

	if (utfcount > 0) {
		error ("Invalid UTF-8 sequence", NIL_PTR);
	}
  } else if (Chinese == True && multichar (c)) {
	buffer [0] = c;
	textpoi = cur_text;
	textpoi ++;
	c = * textpoi;
/*	DCC ();	*/
	insertcode (buffer [0], radix);
	insertcode (c, radix);
  } else {
/*	DCC ();	*/
	insertcode (c, radix);
  }
}
void
changefromcode (format)
  char * format;
{
  unsigned long code;

  sscanf (cur_text, format, & code);
  if (utf8_text == True) {
	Sutf8 (code);
  } else if (Chinese == True && code > 0xFF) {
	S ((code >> 8) & 0xFF);
	S (code & 0xFF);
  } else {
	S (code & 0xFF);
  }
}
void
changeuni ()
{
  if (hop_flag > 0) {
	changefromcode ("%lx");
  } else {
	changetocode (16, True);
  }
}
void
changehex ()
{
  if (hop_flag > 0) {
	changefromcode ("%2lx");
  } else {
	changetocode (16, False);
  }
}
void
changeoct ()
{
  if (hop_flag > 0) {
	changefromcode ("%lo");
  } else {
	changetocode (8, False);
  }
}
void
changedec ()
{
  if (hop_flag > 0) {
	changefromcode ("%lu");
  } else {
	changetocode (10, False);
  }
}

/*
 * display_code shows UTF-8 code sequence and Unicode value on the status line
 */
char hexbuf [20];
char * hexbufpoi = hexbuf;

static
struct scriptentry {
	unsigned long first, last;
	char * script;
} scripttable [] =
{
#include "scriptbl.h"
};

/*
   Determine script of Unicode character according to script range table.
 */
char *
script (ucs)
  unsigned long ucs;
{
  int min = 0;
  int max = sizeof (scripttable) / sizeof (struct scriptentry) - 1;
  int mid;

  /* binary search in table */
  while (max >= min) {
    mid = (min + max) / 2;
    if (scripttable [mid].last < ucs) {
      min = mid + 1;
    } else if (scripttable [mid].first > ucs) {
      max = mid - 1;
    } else if (scripttable [mid].first <= ucs && scripttable [mid].last >= ucs) {
      return scripttable [mid].script;
    }
  }

  return "";
}

void
appendcode (c)
  character c;
{
  * hexbufpoi ++ = hexdig ((c >> 4) & 0x0F);
  * hexbufpoi ++ = hexdig ((c) & 0x0F);
}

void
display_the_code ()
{
  character c = * cur_text;
  int utfcount = 1;
  unsigned long unichar;
  int utfcount2;
  unsigned long unichar2;
  char * textpoi;
  FLAG invalid = False;
  char * scriptmsg;
  char * scriptsep;
  char * widemsg;
  char * combinedmsg;

  hexbufpoi = hexbuf;

  if (c == '\n') {
	switch (cur_line->return_type) {
	    case lineend_NONE:	textpoi = "No lineend (split line)";
				break;
	    case lineend_NUL:	appendcode ('\0');
				textpoi = "Nul character";
				break;
	    case lineend_LF:	appendcode ('\n');
				textpoi = "Line end";
				break;
	    case lineend_CRLF:	appendcode ('\r');
				appendcode ('\n');
				textpoi = "DOS line end";
				break;
	    case lineend_CR:	appendcode ('\r');
				textpoi = "Mac line end";
				break;
	    case lineend_LS:	/* Unicode line separator 2028:   */
				appendcode ('\342');
				appendcode ('\200');
				appendcode ('\250');
				textpoi = "Unicode line separator, code 2028";
				break;
	    case lineend_PS:	/* Unicode paragraph separator 2029:   */
				appendcode ('\342');
				appendcode ('\200');
				appendcode ('\251');
				textpoi = "Unicode paragraph separator, code 2029";
				break;
	    default:		appendcode ('\n');
				textpoi = "Unknown line end";
				break;
	}
	* hexbufpoi = '\0';
	build_string (text_buffer, "%s: %s", textpoi, hexbuf);
	status_msg (text_buffer);
	return;
  } else if (utf8_text == True) {
	if ((c & 0x80) == 0x00) {
		utfcount = 1;
		unichar = c;
	} else if ((c & 0xE0) == 0xC0) {
		utfcount = 2;
		unichar = c & 0x1F;
	} else if ((c & 0xF0) == 0xE0) {
		utfcount = 3;
		unichar = c & 0x0F;
	} else if ((c & 0xF8) == 0xF0) {
		utfcount = 4;
		unichar = c & 0x07;
	} else if ((c & 0xFC) == 0xF8) {
		utfcount = 5;
		unichar = c & 0x03;
	} else if ((c & 0xFE) == 0xFC) {
		utfcount = 6;
		unichar = c & 0x01;
	} else /* illegal UTF-8 code */ {
		appendcode (c);
		invalid = True;
	}
	textpoi = cur_text;
	if (invalid == False) {
		utfcount --;
		appendcode (c);
		textpoi ++;
		while (utfcount > 0 && (* textpoi & 0xC0) == 0x80) {
			c = * textpoi ++;
			appendcode (c);
			unichar = (unichar << 6) | (c & 0x3F);
			utfcount --;
		}
		if (utfcount > 0) {
			invalid = True;
		}
	} else {
		textpoi ++;
	}

	utf8_info (textpoi, & utfcount2, & unichar2);
	if (iscombining (unichar2)) {
		combinedmsg = " - combined ...";
	} else {
		combinedmsg = "";
	}

	* hexbufpoi = '\0';
	if (invalid == False) {
	    scriptmsg = script (unichar);
	    if (* scriptmsg == '\0') {
		scriptsep = scriptmsg;
	    } else {
		scriptsep = " ";
	    }

	    if (iswide (unichar)) {
		if (iscombining (unichar)) {
			widemsg = "wide combining ";
		} else {
			widemsg = "wide ";
		}
	    } else if (iscombining (unichar)) {
		widemsg = "combining ";
	    } else if ((unichar & 0xFC00) == 0xD800) {
		widemsg = "single high surrogate ";
	    } else if ((unichar & 0xFC00) == 0xDC00) {
		widemsg = "single low surrogate ";
	    } else if ((unichar & 0xFFFE) == 0xFFFE) {
		widemsg = "reserved ";
	    } else {
		widemsg = "";
	    }

	    if (unichar > 0xFFFF) {
		build_string (text_buffer, 
			"UTF-8 sequence: %s, %s%s%sUCS-4: %08lX%s", 
			hexbuf, scriptmsg, scriptsep, widemsg, unichar, combinedmsg);
	    } else {
		build_string (text_buffer, 
			"UTF-8 sequence: %s, %s%s%sUCS-2 (Unicode): %04lX%s", 
			hexbuf, scriptmsg, scriptsep, widemsg, unichar, combinedmsg);
	    }
	} else {
	    build_string (text_buffer, "Invalid UTF-8 sequence: %s%s", hexbuf, combinedmsg);
	}
	status_msg (text_buffer);
	return;
  } else if (Chinese == True && multichar (c)) {
	appendcode (c);
	textpoi = cur_text;
	textpoi ++;
	c = * textpoi;
	appendcode (c);
  } else {
	appendcode (c);
  }
  * hexbufpoi = '\0';
  build_string (text_buffer, "Character code: %s", hexbuf);
  status_msg (text_buffer);
}

void
display_code ()
{
  if (hop_flag > 0) {
	hop_flag = 0;
	if (always_disp_code == True) {
		always_disp_code = False;
	} else {
		always_disp_code = True;
	}
  } else {
	display_the_code ();
  }
}

/*
 * Replace current diacritic transcription with real character
 */
void
UML ()
{
  character c1 = * cur_text;
  char * c2poi = cur_text;
  character c2;
  unsigned long uc;

  if (Chinese == True) return;

  c2poi ++;
  c2 = * c2poi;

  if (c1 == 'a' && c2 == 'e') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'o' && c2 == 'e') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'u' && c2 == 'e') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'A' && c2 == 'e') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'O' && c2 == 'e') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'U' && c2 == 'e') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'A' && c2 == 'E') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'O' && c2 == 'E') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 'U' && c2 == 'E') {
	DCC (); DCC (); Suni1 ('');
  } else if (c1 == 's' && c2 == 's') {
	DCC (); DCC (); Suni1 ('');
  } else {
	uc = compose (c1, c2);
	if (uc != '') {
		DCC (); DCC (); Sutf8 (uc);
	}
  }
}

void
delete_basechar ()
{
  char * after_char = cur_text;
  int text_offset = cur_text - cur_line->text;

  advance_char (& after_char);
  (void) delete_text (cur_line, cur_text, cur_line, after_char);
  /* fieser Trick: */
  move_address (cur_line->text + text_offset, y);
}

struct {
	unsigned long base;
	short toupper, tolower;
} caseconv_table [] =
{
#include "casetabl.h"
};

#define caseconv_table_size	(sizeof (caseconv_table) / sizeof (* caseconv_table))

typedef struct {unsigned short u1, u2, u3;} uniseq;
#define U_cond_Final_Sigma		1
#define U_cond_After_Soft_Dotted	2
#define U_cond_More_Above		4
#define U_cond_Not_Before_Dot		8
#define U_cond_tr			16
#define U_cond_lt			32
#define U_cond_az			64
struct {
	unsigned long base;
	uniseq lower, title, upper;
	short condition;
} caseconv_special [] =
{
#include "casespec.h"
};

#define caseconv_special_size	(sizeof (caseconv_special) / sizeof (* caseconv_special))

int
lookup_caseconv (basechar)
  unsigned long basechar;
{
	int low = 0;
	int high = caseconv_table_size - 1;
	int i;

	while (low <= high) {
		i = (low + high) / 2;
		if (caseconv_table [i].base == basechar) {
			return i;
		} else if (caseconv_table [i].base >= basechar) {
			high = i - 1;
		} else {
			low = i + 1;
		}
	}
	/* notify "not found" */
	return -1;
}

int
lookup_caseconv_special (basechar)
  unsigned long basechar;
{
	int low = 0;
	int high = caseconv_special_size - 1;
	int i;

	while (low <= high) {
		i = (low + high) / 2;
		if (caseconv_special [i].base == basechar) {
			return i;
		} else if (caseconv_special [i].base >= basechar) {
			high = i - 1;
		} else {
			low = i + 1;
		}
	}
	/* notify "not found" */
	return -1;
}

/*
 * Flip lower and upper case letters
 */
void
LOWCAP ()
{
  /* for UTF-8 mode */
  int utfcount;
  unsigned long unichar;
  /* for single-byte mode */
  character c1;
  int prev_x;
  int tabix;
  unsigned long convchar;
  short condition;
  char * after_char;
  unsigned long unichar2;

  if (* cur_text == '\n') {
	MRT ();
  } else if (utf8_text == True) {
    do {
	tabix = 0;
	utf8_info (cur_text, & utfcount, & unichar);
	if (unichar == (character) '')
	{
		prev_x = x;
		delete_basechar ();
		Sutf8 ('S'); Sutf8 ('S');
		if (x <= prev_x + 1)	/* may occur with combining chars */
			move_to (prev_x + 2, y);
	}
#define dont_optimise_Latin1
#ifdef optimise_Latin1
	else if ((unichar >= 'a' && unichar <= 'z' && unichar != 'i')
	 || (unichar >= (character) '' && unichar <= (character) '' && unichar != 247)
	   )
	{
		prev_x = x;
		delete_basechar ();
		Sutf8 (unichar - 0x20);
		if (x == prev_x)	/* may occur with combining chars */
			move_to (prev_x + 1, y);
	}
	else if ((unichar >= 'A' && unichar <= 'Z' && unichar != 'I')
	      || (unichar >= (character) '' && unichar <= (character) '' && unichar != 215)
		)
	{
		prev_x = x;
		delete_basechar ();
		Sutf8 (unichar + 0x20);
		if (x == prev_x)	/* may occur with combining chars */
			move_to (prev_x + 1, y);
	}
#endif
	else if ((tabix = lookup_caseconv_special (unichar)) >= 0)
	{
	    condition = caseconv_special [tabix].condition;
	    if (condition == U_cond_Final_Sigma) {
		after_char = cur_text;
		advance_char (& after_char);
		utf8_info (after_char, & utfcount, & unichar2);
		while (iscombining (unichar2)) {
			advance_char (& after_char);
			utf8_info (after_char, & utfcount, & unichar2);
		}
		if (unichar2 < (unsigned long) 'A'
		|| (unichar2 > (unsigned long) 'Z' && unichar2 < (unsigned long) 'a')
		|| (unichar2 > (unsigned long) 'z' && unichar2 < (unsigned long) (character) '')
		) {	/* final position detected */
			condition = 0;
		}
	    } else if (condition == U_cond_tr && Turkish == True) {
		condition = 0;
	    }

	    if (condition == 0) {
		prev_x = x;
		delete_basechar ();
		if (caseconv_special [tabix].base == caseconv_special [tabix].lower.u1)
		{
			/* convert to upper (or title ...) */
			Sutf8 (caseconv_special [tabix].upper.u1);
			if (caseconv_special [tabix].upper.u2 != 0) {
				Sutf8 (caseconv_special [tabix].upper.u2);
				if (caseconv_special [tabix].upper.u3 != 0) {
					Sutf8 (caseconv_special [tabix].upper.u3);
				}
			}
		} else {
			/* convert to lower */
			Sutf8 (caseconv_special [tabix].lower.u1);
			if (caseconv_special [tabix].lower.u2 != 0) {
				Sutf8 (caseconv_special [tabix].lower.u2);
				if (caseconv_special [tabix].lower.u3 != 0) {
					Sutf8 (caseconv_special [tabix].lower.u3);
				}
			}
		}
		if (x == prev_x)	/* may occur with combining chars */
			move_to (prev_x + 1, y);
	    } else {
		/* notifiy to try further */
		tabix = -1;
	    }
	}
	if (tabix == -1 && (tabix = lookup_caseconv (unichar)) >= 0)
	{
		if (caseconv_table [tabix].toupper != 0) {
			convchar = unichar + caseconv_table [tabix].toupper;
		} else {
			convchar = unichar + caseconv_table [tabix].tolower;
		}
		prev_x = x;
		delete_basechar ();
		Sutf8 (convchar);
		if (x == prev_x)	/* may occur with combining chars */
			move_to (prev_x + 1, y);
	}
	else if (tabix == -1)
	{
		move_to (x + 1, y);
	}
    } while (hop_flag > 0 && idfchar (* cur_text));
  } else {
    do {
	c1 = * cur_text;
	if ((c1 >= 'a' && c1 <= 'z')
	 || (Chinese == False && c1 >= (character) '' && c1 <= (character) '' && c1 != (character) 247)
	   ) {
		DCC ();
		Suni1 ((character) ((unsigned int) c1 - 0x20));
	}
	else if ((c1 >= 'A' && c1 <= 'Z')
	      || (Chinese == False && c1 >= (character) '' && c1 <= (character) '' && c1 != (character) 215)
		) {
		DCC ();
		Suni1 ((character) ((unsigned int) c1 + 0x20));
	}
	else if (Chinese == False && c1 == (character) '') {
		DCC ();
		Suni1 ('S'); Suni1 ('S');
	}
	else move_to (x + 1, y);
    } while (hop_flag > 0 && idfchar (* cur_text));
  }
}

/*
 * insert_unichar inserts unicode character
 */
static
void
insert_unichar (composed)
  unsigned long composed;
{
  if (utf8_text == True) {
	Sutf8 (composed);
  } else if ((composed & 0xFF) == composed) {
	Suni1 (composed);
  } else {
	Suni1 ('');
  }
}

/*
   insert next char with defined accent prefix (entered by function key)
 */
void
insert_diaeresis ()
{
  insert_unichar (diaeresis (readchar ()));
}
/*
   insert next char with defined accent prefix (entered by function key)
 */
void
insert_angstrom ()
{
  insert_unichar (angstrom (readchar ()));
}
/*
   insert next char with defined accent prefix (entered by function key)
 */
void
insert_acute ()
{
  insert_unichar (acute (readchar ()));
}
/*
   insert next char with defined accent prefix (entered by function key)
 */
void
insert_grave ()
{
  insert_unichar (grave (readchar ()));
}
/*
   insert next char with defined accent prefix (entered by function key)
 */
void
insert_circumflex ()
{
  insert_unichar (circumflex (readchar ()));
}
/*
   insert next char with defined accent prefix (entered by function key)
 */
void
insert_tilde ()
{
  insert_unichar (tilde (readchar ()));
}

/*
 * insert_accent combines and inserts accented character
 */
static
void
insert_accent (name, routine)
  char * name;
  charfunc routine;
{
  unsigned long composed;

  build_string (text_buffer, "Enter character to compose with %s...", name);
  status_msg (text_buffer);
  composed = (* routine) (readchar ());
  clear_status ();
  insert_unichar (composed);
}

/*
 * CTRl inserts a control-char at the current location. A message that this
 * function is called is displayed at the status line.
 */
void
CTRl ()
{
  register character ctrl;
  unsigned long unichar;
  char mnemonic [maxLINE_LEN];

  status_msg ("Enter control char / # hex / ## octal / compose char / <blank> mnemonic ...");
  ctrl = readcharacter ();

  if (ctrl == ring /*|| ctrl == ','*/) {
	insert_accent ("angstrom/cedilla", angstrom); return;
  } else switch (ctrl) {
	/*case ':':*/
	case '"':	{insert_accent ("diaeresis", diaeresis); return;}
	case (character) '':	{insert_accent ("acute (d'aigu)", acute); return;}
	case '\'':	{insert_accent ("acute (d'aigu)", acute); return;}
	case '`':	{insert_accent ("grave", grave); return;}
	case '^':	{insert_accent ("circumflex", circumflex); return;}
	case '~':	{insert_accent ("tilde", tilde); return;}
  }

  if (ctrl == FUNcmd && keyproc != I) {
	if (keyproc == MLF) {
		ctrl_MLF ();
		return;
	} else if (keyproc == MRT) {
		ctrl_MRT ();
		return;
	} else if (keyproc == DCC) {
		delete_char (False);
		return;
	}
	ring_bell ();
	clear_status ();
	return;
  }

  if (utf8_text == True) {
    if (ctrl == '#') {
	if (get_hex_number ("Enter hex digits for Unicode value to compose character...", & unichar) 
		!= ERRORS) {
		Sutf8 (unichar);
	}
	return;
    } else if (ctrl == ' ') {
	if (get_string ("Enter character mnemonic: ", mnemonic, False, " ") 
		== ERRORS) {
		return;
	}
	unichar = compose_mnemo (mnemonic);
	clear_status ();
	Sutf8 (unichar);
	return;
    } else if (
	((ctrl > ' ') && (ctrl <= 'Z')) || (ctrl > '^')
	) {
	build_string (text_buffer, "Enter character to compose with %c...", ctrl);
	status_msg (text_buffer);
	unichar = compose (ctrl, readchar ());
	clear_status ();
	Sutf8 (unichar);
	return;
    }
  } else if (ctrl == '#') {
	if (get_hex_number ("Enter hex digits for character...", & unichar) 
		!= ERRORS) {
		Suni1 (unichar);
	}
	return;
  } else if (
	((ctrl > ' ') && (ctrl <= 'Z')) || (ctrl > '^')
	) {
	build_string (text_buffer, "Enter character to compose with %c...", ctrl);
	status_msg (text_buffer);
	unichar = compose (ctrl, readchar ());
	clear_status ();
	Suni1 (unichar);
	return;
  }

  clear_status ();
  if (ctrl == '\177') {S ('\177'); return;}
  ctrl = ctrl & '\237';
  S (ctrl);
}

/*
 * LIB insert a line at the current position and moves back to the end of
 * the previous line. It keeps the line end type of the current line.
 */
void
LIB ()
{
  lineend_type return_type;

  if (viewonly == True)
	{viewonlyerr (); return;}

  if (hop_flag > 0) {
	return_type = lineend_NONE;
	hop_flag = 0;
  } else {
	return_type = cur_line->return_type;
  }

  S ('\n');				/* Insert the line */
  MUP ();				/* Move one line up */
  move_to (LINE_END, y);		/* Move to end of this line */
  if (cur_line->return_type != return_type) {
	cur_line->return_type = return_type;
	put_line (cur_line, x, True, False);
  }
}


/*======================================================================*\
|*			Yank commands					*|
\*======================================================================*/

LINE * mark_line = NIL_LINE;		/* For marking position. */
char * mark_text = NIL_PTR;
#define maxmarkers 10
LINE * mark_n_line [maxmarkers] = {
	NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE, 
	NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE, NIL_LINE};
char * mark_n_text [maxmarkers] = {
	NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR, 
	NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR, NIL_PTR};

/*
 * PT () inserts the buffer at the current location.
 */
void
PT ()
{
  register int fd;		/* File descriptor for buffer */

  if (viewonly == True)
	{viewonlyerr (); return;}

  if (hop_flag > 0) {
	if ((fd = open (yankie_file, O_RDONLY | O_BINARY, 0)) < 0) {
		error ("No inter window buffer present", NIL_PTR);
		return;
	}
  }
  else if ((fd = yankfile (READ, False)) == ERRORS) {
	error ("Buffer is empty", NIL_PTR);
	return;
  }
  /* Insert the buffer */
/*	file_insert (fd, False); => positioning error when TAB in last line */
/*		also: UTF-8 positioning correct? (char_count vs. length_of) */
	file_insert (fd, True);
}

/*
 * paste_HTML () inserts the HTML embedding buffer at the current location.
 */
void
paste_HTML ()
{
  int fd;		/* File descriptor for buffer */

  if (viewonly == True)
	{viewonlyerr (); return;}

  if ((fd = open (html_file, O_RDONLY | O_BINARY, 0)) < 0) {
	error ("HTML paste buffer vanished", NIL_PTR);
	return;
  }
  file_insert (fd, True);
}

/*
 * INSFILE () prompts for a filename and inserts the file at the current
 * location in the file.
 */
void
INSFILE ()
{
  register int fd;		/* File descriptor of file */
  char name [maxLINE_LEN];	/* Buffer for file name */

  if (restricted == True)
	{restrictederr (); return;}

  if (viewonly == True)
	{viewonlyerr (); return;}

/* Get the file name */
  if (get_file ("Get and insert file:", name) != FINE)
	return;
  clear_status ();

  status_line ("Inserting ", name);
  if ((fd = open (name, O_RDONLY | O_BINARY, 0)) < 0)
	error ("Cannot open file: " /*, name */, serror ());
  else {	/* Insert the file */
	file_insert (fd, True);	/* leave cursor at begin of insertion */
  }
}

/*
 * File_insert () inserts the contents of an opened file (as given by
 * filedescriptor fd) at the current location.
 * After the insertion, if old_pos is True, the cursor remains at the
 * start of the inserted text, if old_pos is False, it is placed to
 * its end. If old_pos is False, this works erroneously if the last line
 * inserted contains a TAB character!!
 */
void
file_insert (fd, old_pos)
  int fd;
  FLAG old_pos;
{
  char line_buffer [MAX_CHARS];		/* Buffer for next line */
  register LINE * line = cur_line;
  register int line_count = total_lines;	/* Nr of lines inserted */
  LINE * page = cur_line;
  int ret;
  int len;
  lineend_type return_type;

  reset_get_line ();

/* Get the first piece of text (might be ended with a '\n') from fd */
  if ((ret = get_line (fd, line_buffer, & len)) == ERRORS)
	return;				/* Empty file */

/* Adjust line end type if present */
  if (ret == SPLIT_LINE)
	return_type = lineend_NONE;
  else if (ret == NUL_LINE)
	return_type = lineend_NUL;
  else if (ret != NO_LINE)
	return_type = extract_lineend_type (line_buffer);
  else
	return_type = cur_line->return_type;

/* Insert this text at the current location */
  if (insert (line, cur_text, line_buffer) == ERRORS)
	return;
  cur_line->return_type = return_type;


/* Repeat getting lines (and inserting lines) until EOF is reached */
  while (line != NIL_LINE
	 && (ret = get_line (fd, line_buffer, & len)) != ERRORS && ret != NO_LINE)
  {
	if (ret == SPLIT_LINE)
		return_type = lineend_NONE;
	else if (ret == NUL_LINE)
		return_type = lineend_NUL;
	else
		return_type = extract_lineend_type (line_buffer);
	line = line_insert (line, line_buffer, len, return_type);
  }

  if (line == NIL_LINE) sleep (2) /* show memory allocation error msg */;
  else if (ret == NO_LINE) {	/* Last line read not ended by a '\n' */
	line = line->next;
	if (insert (line, line->text, line_buffer) == ERRORS)
		sleep (2) /* give time to read error msg */;
  }

  (void) close (fd);

/* If illegal lines were input, report */
  show_get_l_errors ();

/* Calculate nr of lines added */
  line_count = total_lines - line_count;

/* Fix the screen */
  if (line_count == 0) {		/* Only one line changed */
	set_cursor (0, y);
	line_print (line);
	move_to ((old_pos == True) ? x : x + length_of (line_buffer), y);
  }
  else {				/* Several lines changed */
	reset (top_line, y);	/* Reset pointers */
	while (page != line && page != bot_line->next)
		page = page->next;
	if (page != bot_line->next || old_pos == True)
		display (y, cur_line, SCREENMAX - y, y);
		/* screen display style parameter (last) may be inaccurate */
	if (old_pos == True)
		move_to (x, y);
	else {
		if (ret == NO_LINE)
			move_to (col_count (line_buffer), find_y (line));
		else
			move_to (0, find_y (line->next));
	}
  }

/* If nr of added line >= REPORT, print the count */
  if (line_count >= REPORT)
	status_line (num_out ((long) line_count), " lines added");
}

/*
 * WB () writes the buffer (yank_file) into another file, which
 * is prompted for.
 */
void
WB ()
{
  register int new_fd;		/* Filedescriptor to copy file */
  int yank_fd;			/* Filedescriptor to buffer */
  register int cnt;		/* Count check for read/write */
  int ret = FINE;		/* Error check for write */
  char wfile_name [maxLINE_LEN];	/* Output file name */
  char * msg_doing; char * msg_done;

  if (restricted == True)
	{restrictederr (); return;}

/* Checkout the buffer */
  if ((yank_fd = yankfile (READ, False)) == ERRORS) {
	error ("Buffer is empty", NIL_PTR);
	return;
  }

/* Get file name */
  if (get_file ((hop_flag > 0) ? "Append buffer to file:"
			       : "Write buffer to file:", wfile_name) != FINE)
	return;

/* Create the new file or open previous file for appending */
  if (hop_flag > 0) {
    status_line ("Opening ", wfile_name);
    if ((new_fd = open (wfile_name, O_WRONLY | O_CREAT | O_APPEND | O_BINARY, fprot)) < 0) {
	error ("Cannot append to file: ", serror ());
	return;
    }
    msg_doing = "Appending "; msg_done = "Appended";
  }
  else {
    if (checkoverwrite (wfile_name) != True)
	return;
    else {
	status_line ("Opening ", wfile_name);
	if ((new_fd = open (wfile_name, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, fprot)) < 0) {
		error ("Cannot create file: ", serror ());
		return;
	}
    }
    msg_doing = "Writing "; msg_done = "Wrote";
  }

  status_line (msg_doing, wfile_name);

/* Copy buffer into file */
  while ((cnt = read (yank_fd, text_buffer, sizeof (text_buffer))) > 0) {
	if (write (new_fd, text_buffer, (unsigned int) cnt) != cnt) {
		bad_write (new_fd);
		ret = ERRORS;
		break;
	}
  }

/* Clean up open files and status_line */
  (void) close (yank_fd);
  if (close (new_fd) < 0) {
	error ("Write buffer to file failed: ", serror ());
	ret = ERRORS;
  }

  if (ret != ERRORS)			/* Bad write */
	file_status (msg_done, bytes_saved, wfile_name, lines_saved, 
			False, True, False, False, chars_saved);
}

/*
 * MARK sets mark_line / mark_text to the current line / current text pointer.
 */
void
MARK ()
{
  if (hop_flag > 0) GOMA ();
  else {
	Pushmark ();
	mark_line = cur_line;
	mark_text = cur_text;
	status_msg ("Mark set");
  }
}

/*
 * GOMA moves to the marked position
 */
void
GOMA ()
{
  if (checkmark (mark_line, mark_text) == NOT_VALID)
	error ("Mark not set", NIL_PTR);
  else
	move_address (mark_text, find_y (mark_line));
}

/*
 * MARKn sets mark n to the current line / current text pointer.
 * Markn sets it silently.
 * Pushmark saves the currrent marker to marker 0.
 */
void
MARKn (n)
  int n;
{
  if (hop_flag > 0) GOMAn (n);
  else {
	if (n < 0 || n >= maxmarkers) {
		error ("Marker # out of range", NIL_PTR);
		return;
	}
	mark_n_line [n] = cur_line;
	mark_n_text [n] = cur_text;
	status_msg ("Mark set");
  }
}

void
Markn (n)
  int n;
{
	if (n == -1) {	/* initial mark */
		mark_line = cur_line;
		mark_text = cur_text;
	} else if (n < 0 || n >= maxmarkers) {
		error ("Marker # out of range", NIL_PTR);
		return;
	} else {
		mark_n_line [n] = cur_line;
		mark_n_text [n] = cur_text;
	}
}

void
Pushmark ()
{
	mark_n_line [0] = mark_line;
	mark_n_text [0] = mark_text;
}

/*
 * GOMAn moves to the marked position n
 */
void
GOMAn (n)
  int n;
{
  if (n < 0 || n >= maxmarkers) {
	error ("Marker # out of range", NIL_PTR);
	return;
  }

  if (checkmark (mark_n_line [n], mark_n_text [n]) == NOT_VALID)
	error ("Mark not set", NIL_PTR);
  else
	move_address (mark_n_text [n], find_y (mark_n_line [n]));
}

#ifndef linkyank
/*
 * copy one file to the other
 */
void
copyfile (yank_file, yankie_file)
  char * yank_file;
  char * yankie_file;
{
  int yank_fd;		/* Filedescriptor to yank buffer */
  int new_fd;		/* Filedescriptor to yankie file */
  register int cnt;	/* Count check for read/write */

/* Checkout the buffer */
  if ((yank_fd = open (yank_file, O_RDONLY | O_BINARY, 0)) < 0) {
	return;
  }

/* Create yankie file */
  if ((new_fd = open (yankie_file, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, fprot)) < 0) {
	return;
  }

/* Copy buffer into file */
  while ((cnt = read (yank_fd, text_buffer, sizeof (text_buffer))) > 0) {
	if (write (new_fd, text_buffer, (unsigned int) cnt) != cnt) {
		break;
	}
  }

/* Clean up open files and status_line */
  (void) close (yank_fd);
  (void) close (new_fd);
}
#endif

/*
 * Yankie () provides a reference to the last saved buffer to be read
 * by other mined invocations.
 */
void
yankie ()
{
#ifdef linkyank
  delete_file (yankie_file);
  (void) link (yank_file, yankie_file);
#else
  (void) copyfile (yank_file, yankie_file);
#endif
}

/*
 * yank_block is an interface to the actual yank.
 * It calls checkmark () to check if the marked position is still valid.
 * If it is, yank_text is called.
 */
void
yank_block (remove, append)
  FLAG remove;	/* == DELETE if text should be deleted */
  FLAG append;	/* == True if text should only be appended to yank buffer */
{
  switch (checkmark (mark_line, mark_text)) {
	case NOT_VALID :
		error ("Mark not set", NIL_PTR);
		return;
	case SMALLER :
		yank_text (yankfile (WRITE, append), & yank_status, 
				mark_line, mark_text, cur_line, cur_text, 
				remove, append);
		yankie ();
		break;
	case BIGGER :
		yank_text (yankfile (WRITE, append), & yank_status, 
				cur_line, cur_text, mark_line, mark_text, 
				remove, append);
		yankie ();
		break;
	case SAME :		/* Ignore stupid behaviour */
		status_msg ("Nothing to save");
		break;
	default :
		error ("Internal mark error", NIL_PTR);
		return;
  }
}

void
yank_HTML (remove)
  FLAG remove;	/* == DELETE if text should be deleted */
{
  switch (checkmark (mark_line, mark_text)) {
	case NOT_VALID :
		error ("Mark not set", NIL_PTR);
		return;
	case SMALLER :
		yank_text (htmlfile (WRITE, False), & html_status, 
				mark_line, mark_text, cur_line, cur_text, 
				remove, False);
		break;
	case BIGGER :
		yank_text (htmlfile (WRITE, False), & html_status, 
				cur_line, cur_text, mark_line, mark_text, 
				remove, False);
		break;
	case SAME :		/* Ignore stupid behaviour */
		status_msg ("Nothing to save");
		break;
	default :
		error ("Internal mark error", NIL_PTR);
		return;
  }
}

/*
 * YA () puts the text between the marked position and the current
 * in the buffer.
 */
void
YA ()
{
  if (append_flag == True)
	yank_block (NO_DELETE, True);
  else if (hop_flag > 0) {
	if (MENU) {
		append_flag = True;
		displayflags ();
		set_cursor_xy ();
		flush ();
	}
	yank_block (NO_DELETE, True);
	if (MENU) {
		append_flag = False;
		displayflags ();
		set_cursor_xy ();
		flush ();
	}
  } else
	yank_block (NO_DELETE, False);
}

/*
 * DT () is essentially the same as YA (), but in DT () the text is deleted.
 */
void
DT ()
{
  if (viewonly == True)
	{viewonlyerr (); return;}

  if (append_flag == True)
	yank_block (DELETE, True);
  else if (hop_flag > 0)
	yank_block (DELETE, True);
  else
	yank_block (DELETE, False);
}

/*
 * Check_mark () checks if mark_line and mark_text are still valid pointers.
 * If they are it returns
 * SMALLER if the marked position is before the current,
 * BIGGER if it isn't or SAME if somebody didn't get the point.
 * NOT_VALID is returned when mark_line and/or mark_text are no longer valid.
 * Legal () checks if mark_text is valid on the mark_line.
 */
FLAG
checkmark (mark_line, mark_text)
  register LINE * mark_line;
  register char * mark_text;
{
  register LINE * line;
  FLAG cur_seen = False;

/* Special case: check is mark_line and cur_line are the same. */
  if (mark_line == cur_line) {
	if (mark_text == cur_text)	/* Even same place */
		return SAME;
	if (legal (mark_line, mark_text) == ERRORS) /* mark_text out of range */
		return NOT_VALID;
	if (mark_text < cur_text)
		return SMALLER;
	else	return BIGGER;
  }

/* Start looking for mark_line in the line structure */
  for (line = header->next; line != tail; line = line->next) {
	if (line == cur_line)
		cur_seen = True;
	else if (line == mark_line)
		break;
  }

/* If we found mark_line (line != tail) check for legality of mark_text */
  if (line == tail || legal (mark_line, mark_text) == ERRORS)
	return NOT_VALID;

/* cur_seen is True if cur_line is before mark_line */
  if (cur_seen == True)
	return BIGGER;
  else	return SMALLER;
}

/*
 * Legal () checks if mark_text is still a valid pointer.
 */
int
legal (mark_line, mark_text)
  register LINE * mark_line;
  register char * mark_text;
{
  register char * textp = mark_line->text;

/* Locate mark_text on mark_line */
  while (textp != mark_text && * textp != '\0') {
	textp ++;
  }
  return (* textp == '\0') ? ERRORS : FINE;
}

/*
 * Yank puts all the text between start_position and end_position into
 * the buffer.
 * The caller must check that the arguments to yank_text () are valid (e.g. in
 * the right order).
 */
void
yank_text (fd, buf_status, 
	start_line, start_textp, end_line, end_textp, 
	remove, append)
  int fd;
  FLAG * buf_status;
  LINE * start_line;
  LINE * end_line;
  char * start_textp;
  char * end_textp;
  FLAG remove;	/* == DELETE if text should be deleted */
  FLAG append;	/* == True if text should only be appended to yank buffer */
{
  register LINE * line = start_line;
  register char * textp = start_textp;
  long chars_written = 0L;	/* chars written to buffer this time */
  long bytes_written = 0L;	/* bytes written to buffer this time */
  int lines_written = 0;	/* lines written to buffer this time */
  FLAG in_Chinese = False;
  int return_len;

/* Create file to hold buffer */
  if (fd == ERRORS) {
	return;
  }

  if (append == True) {
	status_msg ("Appending text ...");
  } else {
	status_msg ("Saving text ...");
	chars_saved = 0L;
	bytes_saved = 0L;
	lines_saved = 0;
  }

/* Keep writing chars until the end_location is reached. */
  while (textp != end_textp) {
	if (* textp == '\n') {
		/* handle different line ends */
		return_len = write_lineend (fd, line->return_type);
		if (return_len == ERRORS) {
			(void) close (fd);
			return;
		}

		lines_written ++;
		if (line->return_type != lineend_NONE) {
			chars_written ++;
		}
		bytes_written += return_len;

		/* move to the next line */
		line = line->next;
		textp = line->text;
	} else {
		if (writechar (fd, * textp) == ERRORS) {
			(void) close (fd);
			return;
		}

		if (utf8_text == True) {
			chars_written -= UTFseq_len (textp) - 1;
		} else if (Chinese == True) {
			if (in_Chinese == True) {
				chars_written --;
				in_Chinese = False;
			} else if (multichar (* textp)) {
				in_Chinese = True;
			}
		}

		chars_written ++;
		bytes_written ++;

		/* move to the next byte */
		textp ++;
	}
  }

/* Flush the I/O buffer and close file */
  if (flush_buffer (fd) == ERRORS) {
	(void) close (fd);
	return;
  }
  if (close (fd) < 0) {
	error ("Write to buffer failed: ", serror ());
	return;
  }
  * buf_status = VALID;

  /*
   * Check if the text should be deleted as well. In case it should,
   * the following hack is used to save a lot of code.
   * First move back to the start_position (this might be the current
   * location) and then delete the text.
   * This might look a bit confusing to the user the first time.
   * Delete () will fix the screen.
   */
  if (remove == DELETE) {
	move_to (find_x (start_line, start_textp), find_y (start_line));
	if (delete_text (start_line, start_textp, end_line, end_textp)
		== ERRORS) {
		sleep (2) /* give time to read allocation error msg */;
	}
	mark_line = cur_line;
	mark_text = cur_text;
  }

  bytes_saved += bytes_written;
  chars_saved += chars_written;
  lines_saved += lines_written;

  build_string (text_buffer, "%s %d lines to buffer (chars/bytes: %ld/%ld)", 
	(append == True) ? "Appended" : "Written", lines_written, 
	chars_written, bytes_written);
  status_msg (text_buffer);
}

/*
 * scratchfile/yankfile () tries to create a unique file in a temporary directory.
 * It tries several different filenames until one can be created
 * or MAXTRIALS attempts have been made.
 * After MAXTRIALS times, an error message is given and ERRORS is returned.
 */

#define MAXTRIALS 99

int
scratchfile (mode, append, buf_name, which, buf_status)
  FLAG mode;	/* Can be READ or WRITE permission */
  FLAG append;	/* == True if text should only be appended to yank buffer */
  char * buf_name;
  char * which;
  FLAG * buf_status;
{
  static int trials = 0;	/* Keep track of trials */
  int fd = 0;			/* Filedescriptor to buffer */

/* If * buf_status == NOT_VALID, scratchfile is called for the first time */
  if (* buf_status == NOT_VALID && mode == WRITE) { /* Create new file */
	/* Generate file name. */
#ifdef msdos
	build_string (buf_name, "%s.%s%d", yankie_file, which, trials);
#else
	build_string (buf_name, "%s.%s%d-%d", yankie_file, which, getpid (), trials);
#endif
	/* Check file existence */
	if (access (buf_name, 0 /* F_OK */) == 0
	    || (
/*		fd = creat (buf_name, bufprot)		*/
		fd = open (buf_name, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY, bufprot)
		) < 0)
	{
		if (++ trials >= MAXTRIALS) {
		    build_string (text_buffer, "Unable to create scratchfile %s: ", buf_name);
		    if (fd == 0) {
			error (text_buffer, "File exists");
		    } else {
			error (text_buffer, serror ());
		    }
		    return ERRORS;
		} else {	/* try again */
		    return scratchfile (mode, append, buf_name, which, buf_status);
		}
	}
  }
  else if (* buf_status == NOT_VALID && mode == READ) {
	return ERRORS;
  }
  else /* * buf_status == VALID */
	if (  (mode == READ && (fd = open (buf_name, O_RDONLY | O_BINARY, 0)) < 0)
	   || (mode == WRITE &&
		(fd = open (buf_name, O_WRONLY | O_CREAT
				| ((append == True) ? O_APPEND : O_TRUNC)
				| O_BINARY
				, bufprot)) < 0))
  {
	* buf_status = NOT_VALID;
	return ERRORS;
  }

  clear_buffer ();
  return fd;
}

int
yankfile (mode, append)
  FLAG mode;	/* Can be READ or WRITE permission */
  FLAG append;	/* == True if text should only be appended to yank buffer */
{
  return scratchfile (mode, append, yank_file, "", & yank_status);
}

int
htmlfile (mode, append)
  FLAG mode;	/* Can be READ or WRITE permission */
  FLAG append;	/* == True if text should only be appended to yank buffer */
{
  return scratchfile (mode, append, html_file, "h", & html_status);
}


/*======================================================================*\
|*			Configurable functions				*|
\*======================================================================*/

void
HOMEkey ()
{
  if (mined_keypad)
	MARK ();
  else	BLINE ();
}

void
ENDkey ()
{
  if (mined_keypad)
	YA ();
  else	ELINE ();
}

void
shiftHOMEkey ()
{
  if (mined_keypad)
	BLINE ();
  else	MARK ();
}

void
shiftENDkey ()
{
  if (mined_keypad)
	ELINE ();
  else	YA ();
}


/*======================================================================*\
|*				End					*|
\*======================================================================*/
