#############################################################################
# mined text editor make actions (make include file for make targets)

# Mined modules
OBJ	= $(SCREENOBJ) mined1.o minedaux.o mined2.o output.o keymap.o mousemen.o search.o iswide.o iscombin.o isrtol.o locale.o

# Default make target:
all:	mined

# Unicode data tables:

SpecialCasing.txt:
	wget http://unicode.org/Public/UNIDATA/SpecialCasing.txt

UnicodeData.txt:
	wget http://unicode.org/Public/UNIDATA/UnicodeData.txt

Scripts.txt:
	wget http://unicode.org/Public/UNIDATA/Scripts.txt

Blocks.txt:
	wget http://unicode.org/Public/UNIDATA/Blocks.txt

casespec.h:	SpecialCasing.txt mkcasespec
	./mkcasespec

casetabl.h:	UnicodeData.txt mkcasetabl
	./mkcasetabl

# Unicode character information tables:
scriptbl.h:	Scripts.txt mkscript
	./mkscript

colours.h:	colours.cfg mkinclud.mak
	sed -e 's/[ 	"]*\([A-Z][A-Z]*\)[ 	,"]*\([0-9][0-9]*\).*/	{"\1", \2},/' -e t -e d colours.cfg | sort > colours.h

# Character input support tables:
mnemos.h:	mnemos.1345 mnemos.html mnemos.more mnemos.tex
	cat mnemos.??* > mnemos.h

# Documentation page for mined additional mnemos
mnemodoc:	../doc/mnemodoc.html
../doc/mnemodoc.html:	mnemos.more mkmnemodoc
	./mkmnemodoc

# Source compilation:
mined1.o:	mined1.c mined.h
	$(CC) $(CFLAGS) -c mined1.c
mined2.o:	mined2.c mined.h casespec.h casetabl.h scriptbl.h
	$(CC) $(CFLAGS) -c mined2.c
minedaux.o:	minedaux.c mined.h
	$(CC) $(CFLAGS) $(DECLFLAGS) -c minedaux.c
search.o:	search.c mined.h
	$(CC) $(CFLAGS) $(DECLFLAGS) -c search.c
mousemen.o:	mousemen.c mined.h
	$(CC) $(CFLAGS) $(DECLFLAGS) -c mousemen.c
output.o:	output.c colours.h mined.h
	$(CC) $(CFLAGS) $(DECLFLAGS) -c output.c
io.o:	io.c io.h mined.h keyboard.c keytrans.h
	$(CC) $(CFLAGS) $(PROTOFLAGS) -c io.c
iocurses.o:	io.c io.h keyboard.c keytrans.h mined.h
	$(CC) $(CFLAGS) $(PROTOFLAGS) -c io.c $(ICURSES) -o iocurses.o
keymap.o:	keymap.c mined.h mnemos.h
	$(CC) $(CFLAGS) $(DECLFLAGS) -c keymap.c
iswide.o:	iswide.c
	$(CC) $(CFLAGS) -c iswide.c
iscombin.o:	iscombin.c
	$(CC) $(CFLAGS) -c iscombin.c
isrtol.o:	isrtol.c
	$(CC) $(CFLAGS) -c isrtol.c
locale.o:	locale.c
	$(CC) $(CFLAGS) -c locale.c

# Mined generation:
mined:	$(OBJ)
	$(CC) $(OBJ) $(SLIB) $(LINKOPTS) -o mined

# Online help:
help:	../mined.hlp

../mined.hlp:	../VERSION help/* mkhelp
	./mkhelp > ../mined.hlp

# Manual page:
man:	../man/mined.1

../man/mined.1:	../doc/mined.html
	cd ../man; make

# Installation:
install:	mined ../mined.hlp ../man/mined.1
	mkdir -p $(bindir)
	install mined $(bindir)
	mkdir -p $(helpdir)
	install ../mined.hlp $(helpdir)
	mkdir -p $(mandir)/man1
	install ../man/mined.1 $(mandir)/man1

# Cleanup after compilation:
clean:
	rm -f $(OBJ) core # mined


#############################################################################
# end
