/***********************************************************************\
		Editor mined
		menu handling part
\***********************************************************************/

#include "syntax.h"

#include "mined.h"
#define gotoxy(x, y)	move_cursor (x, y - 1)

#define arrlen(arr)	(sizeof (arr) / sizeof (* arr))

#include "io.h"


/***********************************************************************\
	Menu types
\***********************************************************************/

typedef struct {void (* toggle) ();
		char * (* dispflag) ();
		} flagitemtype;

typedef struct {char * itemname;
		voidfunc itemfu;
		char * hopitemname;
		} itemtype;

typedef struct {char * menuname; int menulen;
		itemtype * menuitems;
		} menutype;


/***********************************************************************\
	Global variables
\***********************************************************************/

/* mode tuning */
FLAG use_block_graphics = True;
local int menuborder = 1;
local int popupmenuborder = 1;

local int flags_pos = 62;
local int flags_displayed = 9;

/* layout tuning */
local int menu_width = 12;
local int popupmenu_width = 16;

/* state */
local int first_dirty_line = 0;
local int last_dirty_line = 0;
local itemtype * last_menu;
local int last_menulen;
local int last_menuwidth;
local char * last_menutitle;

local FLAG pulldownmenu_displayed = False;
local FLAG popupmenu_displayed = False;

local FLAG menuline_dirty = False;

#ifdef full_Keymap_menu
local char keymap_menuname [22];
#endif


/***********************************************************************\
	Flag operations
\***********************************************************************/

local void
	togglenothing ()
Begin
End

local char *
	dispnothing ()
Begin
	return " ";
End


local void
	toggleHOP ()
Begin
	If hop_flag > 0
	Then	hop_flag = 0;
	Else	hop_flag = 1;
	Fi
	displayflags ();
End

local char *
	dispHOP ()
Begin
	return hop_flag > 0 ? "#H" : "h";
End


local void
	toggleVIEW ()
Begin
	If viewonly == True
	Then	EDITmode ();
	Else	VIEWmode ();
	Fi
End

local char *
	dispVIEW ()
Begin
	return viewonly == True ? "#V" : "E";
End


local void
	toggleJUSmode ()
Begin
	JUSmode = 1 - JUSmode;
End

local char *
	dispJUSmode ()
Begin
	return JUSmode == 1 ? "#«" : "# ";
End


local void
	toggleJUSlevel ()
Begin
	JUSlevel ++;
	If JUSlevel > 2
	Then	JUSlevel = 0;
	Fi
End

local char *
	dispJUSlevel ()
Begin
	If JUSlevel == 0
	Then	return "j";
	Elsif JUSlevel == 1
	Then	return "#j";
	Else	return "#J";
	Fi
End


global void
	toggleappend ()
Begin
	If append_flag == True
	Then	append_flag = False;
	Else	append_flag = True;
	Fi
End

local char *
	dispappend ()
Begin
	If append_flag == True
	Then	return "#+";
	Else	return "=";
	Fi
End


local void
	toggleUTF ()
Begin
	char * curpos = cur_text;
	If utf8_text == True
	Then	utf8_text = False;
	Else	utf8_text = True;
	Fi
	RD ();
	menuline_dirty = True;
	/* move cursor to or behind actual previous character position */
	move_address (curpos, y);
	/* adjust not to stay amidst a character */
	move_to (x, y);
End

local char *
	dispUTF ()
Begin
	If utf8_text == True
	Then	return "#U";
	Else	return "L";
	Fi
End


local void
	toggle_combining ()
Begin
	char * curpos = cur_text;
	If combining_mode == True
	Then	combining_mode = False;
	Else	combining_mode = True;
	Fi
	RD ();
	menuline_dirty = True;
	/* move cursor to actual previous character position */
	move_address (curpos, y);
	/* adjust not to stay amidst a combined character */
/*	move_to (x, y);	*/
End

local char *
	disp_combining ()
Begin
	If utf8_text == False || combining_screen == False
	Then	return " ";
	Elsif combining_mode == True
	Then	return "C";
	Else	return "#c";
	Fi
End


local char *
	disp_leftquote ()
Begin
	If utf8_text == True
	Then	return quotes [quote_type].leftquotation;
	Else	return " ";
	Fi
End

local char *
	disp_rightquote ()
Begin
	If utf8_text == True
	Then	return quotes [quote_type].rightquotation;
	Else	return " ";
	Fi
End


local void
	click_leftquote (col) int col;
Begin
	If mouse_button == rightbutton
	Then	handleQuotemenu (col - 2);
	Else	quote_type_down ();
	Fi
End

local void
	click_rightquote (col) int col;
Begin
	If mouse_button == rightbutton
	Then	handleQuotemenu (col - 2);
	Else	quote_type_up ();
	Fi
End


local void
	toggle_HTML ()
Begin
	If dim_HTML == True
	Then	dim_HTML = False;
	Else	dim_HTML = True;
	Fi
	RD ();
End


local void
	toggle_autoindent ()
Begin
	If autoindent == True
	Then	autoindent = False;
	Else	autoindent = True;
	Fi
End

local char *
	disp_autoindent ()
Begin
	If autoindent == True
	Then	return "#»";
	Else	return "#¦";
	Fi
End


local void
	select_keymap_entry (menu, i) itemtype * menu; int i;
Begin
	setKEYMAP (menu [i].hopitemname);
End


local void
	select_quote_type (menu, i) itemtype * menu; int i;
Begin
	set_quote_type (i);
End


/***********************************************************************\
	Flag table
\***********************************************************************/

#define keymap_in_flags
local void clickKEYMAP ();
local char * dispKEYMAP0 ();
local char * dispKEYMAP1 ();


local flagitemtype Flagmenu [] =
{
#ifdef keymap_in_flags
	{clickKEYMAP, dispKEYMAP0},
	{clickKEYMAP, dispKEYMAP1},
	{togglenothing, dispnothing},
#endif
	{click_leftquote, disp_leftquote},
	{click_rightquote, disp_rightquote},
	{togglenothing, dispnothing},
	{toggleUTF, dispUTF},
	{toggle_combining, disp_combining},
	{togglenothing, dispnothing},
	{toggleHOP, dispHOP},
	{togglenothing, dispnothing},
	{toggleVIEW, dispVIEW},
	{toggleappend, dispappend},
	{togglenothing, dispnothing},
	{toggle_autoindent, disp_autoindent},
	{toggleJUSlevel, dispJUSlevel},
	{toggleJUSmode, dispJUSmode},
};


/***********************************************************************\
	Menu tables
\***********************************************************************/

local void
	separator ()
Begin
End


local itemtype Filemenu [] =
{
	{"Open", EDIT, ""},
	{"View", VIEW, ""},
	{"Save", WT, ""},
	{"New Name", NN, ""},
	{"Save & Exit", EXED, ""},
	{"", separator, ""},
	{"paste file", INSFILE, ""},
	{"copy to file", WB, "append to file"},
	{"print buffer", PBUF, ""},
	{"", separator, ""},
	{"check out", checkout, ""},
	{"check in", checkin, ""},
	{"", separator, ""},
	{"next file", NXTFILE, "first file"},
	{"prev file", PRVFILE, "last file"},
	{"n-th file", NTHFILE, ""},
	{"save position", SAVPOS, ""},
	{"Discard & Quit", QUED, ""},
};

local itemtype Editmenu [] =
{
	{"mark", MARK, "Go to Mark"},
	{"cut", DT, "cut & append"},
	{"copy", YA, "append"},
	{"paste", PT, "paste external buf"},
	{"", separator, ""},
	{"Go to Mark", GOMA, ""},
	{"set marker N", MARKER, ""},
	{"go marker N", GOMARKER, ""},
};

local itemtype Searchmenu [] =
{
	{"find", SFW, "find Identf."},
	{"find backw.", SRV, "find Id. backw."},
	{"find again", RS, "previous search"},
	{"", separator, ""},
	{"find Identf.", SIDFW, ""},
	{"find Id. backw.", SIDRV, ""},
	{"find matching ()", SCORR, ""},
	{"go to Idf. def.", Stag, "go to def. ..."},
	{"", separator, ""},
	{"substitute", GR, ""},
	{"replace (?)", REPL, ""},
	{"replace on line", LR, ""},
	{"", separator, ""},
	{"Go to Mark", GOMA, ""},
	{"Go to ...", GOTO, ""},
};

local itemtype Paragraphmenu [] =
{
	{"justify clever", JUSclever, "clever justify other"},
	{"justify simple", JUS, "justify other mode"},
	{"", separator, ""},
	{"set left margin", ADJLM, ""},
	{"set first line left m.", ADJFLM, ""},
	{"set other lines l. m.", ADJNLM, ""},
	{"set right margin", ADJRM, ""},
	{"toggle auto-justif.", toggleJUSlevel, ""},
	{"toggle line-end mode", toggleJUSmode, ""},
};

local itemtype Specialeditmenu [] =
{
	{"insert control char", CTRl, ""},
	{"insert HTML command", HTML, ""},
	{"case switch", LOWCAP, "case switch word"},
	{"", separator, ""},
	{"UTF-8 text", toggleUTF, ""},
	{"combined display", toggle_combining, ""},
	{"file info", FS, "toggle file info"},
	{"char info", display_code, "toggle char info"},
	{"toggle auto indent", toggle_autoindent, ""},
	{"toggle HTML disp", toggle_HTML, ""},
	{"toggle TAB width", toggletab, ""},
	{"", separator, ""},
	{"2 chars -> umlaut", UML, ""},
	{"unicode (char)", changeuni, "char (unicode)"},
	{"hex bytes (char)", changehex, "char (hex)"},
	{"octal bytes (char)", changeoct, "char (octal)"},
	{"dec bytes (char)", changedec, "char (dec)"},
};

#ifdef msdos
local itemtype Screensizemenu [] =
{
#ifdef msdos
	{"bigger", screenbigger, "video mode"},
	{"smaller", screensmaller, "graphic mode"},
	{"more lines", screenmorelines, "font bank"},
	{"less lines", screenlesslines, "char height"},
#else
	{"bigger", screenbigger, ""},
	{"smaller", screensmaller, ""},
	{"more lines", screenmorelines, ""},
	{"less lines", screenlesslines, ""},
#endif
	{"switch line #", LNSW, ""},
	{"circul line #", LNCI, ""},
};
#endif

local itemtype Keymapmenu [] =
{
	{"none", select_keymap_entry, ""},
	{"Greek", select_keymap_entry, "gr"},
	{"Cyrillic", select_keymap_entry, "cy"},
	{"Arabic", select_keymap_entry, "ar"},
	{"Hebrew", select_keymap_entry, "he"},
};

/* Keep this consistent with mined2.c: quoteset quotes [] ! */
local itemtype Quotemenu [] =
{
	{"\"plain\"", select_quote_type, "\"\" ''"},
	{"“English”", select_quote_type, "“” ‘’"},
	{"„German“", select_quote_type, "„“ ‚‘"},
	{"”Swedish”", select_quote_type, "”” ’’"},
	{"«French»", select_quote_type, "«» ‹›"},
	{"»?«", select_quote_type, "»« ›‹"},
	{"„?”", select_quote_type, "„” ‚’"},
	{"»?»", select_quote_type, "»» ››"},
};

local menutype menus [] =
{
	{"File", arrlen (Filemenu), Filemenu},
	{"Edit", arrlen (Editmenu), Editmenu},
	{"Search", arrlen (Searchmenu), Searchmenu},
	{"eXtra", arrlen (Specialeditmenu), Specialeditmenu},
	{"Paragraph", arrlen (Paragraphmenu), Paragraphmenu},
#ifdef msdos
	{"Screen size", arrlen (Screensizemenu), Screensizemenu},
#endif
/* this is replaced by the flags keymap menu: */
#ifdef full_Keymap_menu
	{keymap_menuname, arrlen (Keymapmenu), Keymapmenu},
#endif
};

local itemtype Popupmenu [] =
{
	{"Mark", MARK, "Go to Mark"},
	{"Cut", DT, "Cut-Append"},
	{"Copy", YA, "Append"},
	{"Paste", PT, "Paste external"},
	{"HTML command", HTML, ""},
	{"", separator, ""},
	{"Find", SFW, "FindIdf"},
	{"FindReverse", SRV, "FindIdfRev"},
	{"FindNext", RS, "Previous Find"},
	{"FindIdf", SIDFW, ""},
	{"FindIdfRev", SIDRV, ""},
	{"go to Idf. def.", Stag, "go to def. ..."},
	{"Go to Mark", GOMA, ""},
	{"matching paren", SCORR, ""},
};


/***********************************************************************\
	Flag-triggered menus
\***********************************************************************/

global void
	handleQuotemenu (col) int col;
Begin
	If col == 0
	Then	col = flags_pos + 3;
	Fi
	popup_menu (Quotemenu, arrlen (Quotemenu), col, 0, "Quotes");
End


#ifdef keymap_in_flags

global void
	handleKeymapmenu (col) int col;
Begin
	If col == 0
	Then	col = flags_pos;
	Fi
	popup_menu (Keymapmenu, arrlen (Keymapmenu), col, 0, "Keyboard Map");
End

local void
	clickKEYMAP (col) int col;
Begin
	If mouse_button == rightbutton
	Then	handleKeymapmenu (col - 2);
	Else	cycleKEYMAP ();
	Fi
End

local char km [3];

local char *
	dispKEYMAP0 ()
Begin
	If allow_keymap
	Then	km [0] = '#';
		km [1] = keyboard_mapping [0];
		km [2] = '\0';
		return km;
	Else	return " ";
	Fi
End

local char *
	dispKEYMAP1 ()
Begin
	If allow_keymap
	Then	km [0] = '#';
		km [1] = keyboard_mapping [1];
		km [2] = '\0';
		return km;
	Else	return " ";
	Fi
End

#endif


/***********************************************************************\
	putnstr () writes a string to screen at adjusted length
\***********************************************************************/
local void
	putnstr (s, l) char * s; int l;
Begin
	int i;
/*
	char buf [99];

	strcpy (buf, "");
	strncat (buf, s, (unsigned int) l);
	For i = strlen (s) While i < l Step i ++
	Do
		strcat (buf, " ");
	Done
	putstring (buf);
*/

	For i = 0 While i < l Step i ++
	Do	If * s == '\0'
		Then	put_char (' ');
		Elsif (* s & 0x80) != 0
		Then	put_utfchar (s);
			advance_char (& s);
		Else	put_char (* s ++);
		Fi
	Done
End




/***********************************************************************\
	putblock () writes a block graphics character
	menu*_on/off () switch menu border display mode
	putborder_top/middle/bottom () writes menu borders
\***********************************************************************/
FLAG in_menu_border = False;

local void
	putblock (c) char c;
Begin
	If in_menu_border == True
	Then	putblockchar (c);
	Else	menuborder_on ();
		putblockchar (c);
		menuborder_off ();
	Fi
End

local void
	putborder_top (x, y, width, title) int x, y, width; char * title;
Begin
	int i;

	gotoxy (x, y);
	If title != NIL_PTR
	Then	If use_block_graphics == True
		Then	menuborder_on ();
			putblock ('l');
			menuheader_on ();
			putnstr ("", menuborder);
		Else	menudisp_on ();
			putnstr ("", 1 + menuborder);
		Fi
		putnstr (title, width - 2 - 2 * menuborder);
		If use_block_graphics == True
		Then	putnstr ("", menuborder);
			menuheader_off ();
			menuborder_on ();
			putblock ('k');
		Else	putnstr ("", 1 + menuborder);
		Fi
	Else
		If use_block_graphics == True
		Then	menuborder_on ();
			putblock ('l');
			For i = 2 While i < width Step i ++
			Do	putblock ('q');
			Done
			putblock ('k');
		Else	menudisp_on ();
			putnstr ("", width);
		Fi
	Fi
End

local void
	putborder_bottom (x, y, width) int x, y, width;
Begin
	int i;

	gotoxy (x, y);
	If use_block_graphics == True
	Then
		putblock ('m');
		For i = 2 While i < width Step i ++
		Do	putblock ('q');
		Done
		putblock ('j');
		menuborder_off ();
	Else	putnstr ("", width);
		menudisp_off ();
	Fi
End

local void
	putborder_middle (x, y, width) int x, y, width;
Begin
	int i;

	gotoxy (x, y);
	If use_block_graphics == True
	Then
		putblock ('t');
		For i = 2 While i < width Step i ++
		Do	putblock ('q');
		Done
		putblock ('u');
		menuborder_off ();
	Else	putnstr ("", width);
	Fi
End


/***********************************************************************\
	displaymenuheader () displays a menu header
\***********************************************************************/
local void
	displaymenuheader (meni) int meni;
Begin
	gotoxy (meni * menu_width, 0);
	putchar (' ');
	menudisp_on ();
	putnstr ("", menuborder);
	putnstr (menus [meni].menuname, menu_width - 2 - 2 * menuborder);
	putnstr ("", menuborder);
	menudisp_off ();
	putchar (' ');
End


/***********************************************************************\
	number_menus () determines the number of pull-down menus
	set_scripttag () sets script tag into name of Keymap menu
\***********************************************************************/
local int
	number_menus ()
Begin
#ifdef full_Keymap_menu
	If allow_keymap
	Then	return arrlen (menus);
	Else	return arrlen (menus) - 1;
	Fi
#else
	return arrlen (menus);
#endif
End

#ifdef full_Keymap_menu
global void
	set_scripttag ()
Begin
	strcpy (keymap_menuname, keyboard_mapping);
	strcat (keymap_menuname, " Keymap");
End
#endif


/***********************************************************************\
	calcmenuvalues () determines menu screen positions
	width_of_menu () determines the display width of the menu
\***********************************************************************/
local int
	width_of_menu (menu, menulen) itemtype * menu; int menulen;
Begin
	int i;
	itemtype * item;
	int itemlen;
	int width = 0;
	For i = 0 While i < menulen Step i ++
	Do
		item = & (menu [i]);
		If hop_flag > 0 && item->hopitemname != NIL_PTR && item->hopitemname [0] != '\0'
		Then	itemlen = col_count (item->hopitemname);
		Else	itemlen = col_count (item->itemname);
		Fi
		itemlen += 2 + 2 * menuborder;
		If itemlen > width
		Then	width = itemlen;
		Fi
	Done
	If width > XMAX
	Then	width = XMAX;
	Fi
	return width;
End


local void
	calcmenuvalues ()
Begin
	int flags_gap;

#ifdef pc_dont_use_block_graphics
	If use_curses == False
	Then	use_block_graphics = False;
	Fi
#endif
	If Chinese == True
	Then	use_block_graphics = False;
	Fi

	/* determine positions for menus and flags display */
	menu_width = (XMAX - arrlen (Flagmenu) - 4) / number_menus ();

	If menu_width < 10 || use_block_graphics
	Then	menuborder = 0;
	Else	menuborder = 1;
	Fi

	/* determine popup menu width: */
	popupmenu_width = width_of_menu (Popupmenu, arrlen (Popupmenu));

	If popupmenu_width < 15 || use_block_graphics
	Then	popupmenuborder = 0;
	Else	popupmenuborder = 1;
	Fi

	/* determine position and length of displayed flags */
	flags_pos = XMAX - 2 - arrlen (Flagmenu);
	flags_gap = flags_pos - number_menus () * menu_width;
	flags_displayed = arrlen (Flagmenu);
	If flags_gap < 2
	Then	flags_pos += 2 - flags_gap;
		flags_displayed -= 2 - flags_gap;
	Fi
End


/***********************************************************************\
	displaymenuheaders () displays the menu headers
\***********************************************************************/
local void
	displaymenuheaders_from (firsti) int firsti;
Begin
	int i;
	int starti;

	If menuline_dirty == True
	Then	starti = 0;
		menuline_dirty = False;
	Else	starti = firsti;
	Fi

	calcmenuvalues ();

	If menu_width < 3
	Then	return;
	Fi

#ifdef full_Keymap_menu
	If number_menus () == arrlen (menus)
	Then	set_scripttag ();
	Fi
#endif
	For i = starti While i < number_menus () Step i ++
	Do
		displaymenuheader (i);
	Done
End

local void
	displaymenuheaders ()
Begin
	int i;

	calcmenuvalues ();

	If menu_width < 3
	Then	return;
	Fi

#ifdef full_Keymap_menu
	If number_menus () == arrlen (menus)
	Then	set_scripttag ();
	Fi
#endif
	For i = 0 While i < number_menus () Step i ++
	Do
		displaymenuheader (i);
	Done
End


/***********************************************************************\
	cleargap () clears the gap between menu headers and flags
\***********************************************************************/
local void
	cleargap ()
Begin
	int i;
	int gappos = number_menus () * menu_width;

	gotoxy (gappos, 0);
	For i = gappos While i < flags_pos Step i ++
	Do
		putchar (' ');
	Done
End


/***********************************************************************\
	displayflags () displays the menu flags
\***********************************************************************/
global void
	displayflags ()
Begin
	int i;
	char * flags;
	FLAG is_reverse;

	calcmenuvalues ();

	If menu_width < 3
	Then	return;
	Fi

#ifdef full_Keymap_menu
	If number_menus () == arrlen (menus)
	Then	set_scripttag ();
		displaymenuheader (number_menus () - 1);
	Fi
#endif

	gotoxy (flags_pos, 0);
	For i = 0 While i < flags_displayed Step i ++
	Do
		flags = (* Flagmenu [i].dispflag) ();

		If * flags == '#'
		Then	is_reverse = True;
			flags ++;
			menudisp_on ();
		Else	is_reverse = False;
		Fi
		If (* flags & 0x80) != 0
		Then	put_utfchar (flags);
		Else	put_char (* flags);
		Fi
		If	is_reverse == True
		Then	menudisp_off ();
		Fi
	Done
End


/***********************************************************************\
	displaymenuline () displays the menu line
\***********************************************************************/
local void
	displaymenuline_from (firsti) int firsti;
Begin
	displaymenuheaders_from (firsti);
	cleargap ();
	displayflags ();
	clear_eol ();
	top_line_scrolled = False;
End


global void
	displaymenuline ()
Begin
	displaymenuheaders ();
	cleargap ();
	displayflags ();
	clear_eol ();
	top_line_scrolled = False;
End


/***********************************************************************\
	prepare_menuline prepares menu line to prevent 
	right-to-left menu obstruction
\***********************************************************************/
local void
	prepare_menuline (line, menu_line_poi)
		int line; LINE * * menu_line_poi;
Begin
	char * cp;
	int utfcount;
	unsigned long unichar;

	If * menu_line_poi == (LINE *) 0
	Then	return;
	Fi

	If bidi == False && disp_scrollbar == False && utf8_text == True && utf8_screen == True
	Then	For cp = (* menu_line_poi)->text
		While * cp != '\0'
		Step advance_utf8 (& cp)
		Do	utf8_info (cp, & utfcount, & unichar);
			If is_right_to_left (unichar)
			Then	break;
			Fi
		Done
		If * cp != '\0'
		Then	gotoxy (0, line);
			clear_eol ();
		Fi
	Fi
	If * menu_line_poi == bot_line
	Then	* menu_line_poi = (LINE *) 0;
	Else	* menu_line_poi = (* menu_line_poi)->next;
	Fi
End




/***********************************************************************\
	menu state and saved menu status for redrawmenu ()
\***********************************************************************/

local FLAG pulldownmenu_active = False;
local FLAG popupmenu_active = False;

local int last_pulldown_menu;
local int last_pulldown_item;

local int last_popup_column;
local int last_popup_line;
local int last_popup_item;


/***********************************************************************\
	pulldown_menu () displays a pulldown menu
\***********************************************************************/
local int
	pulldown_menu (meni) int meni;
Begin
	int i;
	int menulen = menus [meni].menulen;
	int meniwidth = menu_width;
	int menuxpos = meni * menu_width;
	itemtype * item;
	int itemlen;
	LINE * menu_line;

	menu_mouse_mode (True);

	/* save redraw values: */
	last_pulldown_menu = meni;
	last_pulldown_item = -1;

	If menulen > YMAX - 1
	Then menulen = YMAX - 1;
	Fi

	/* determine menu width: */
	For i = 0 While i < menulen Step i ++
	Do
		item = & (menus [meni].menuitems [i]);
		If hop_flag > 0 && item->hopitemname != NIL_PTR && item->hopitemname [0] != '\0'
		Then	itemlen = col_count (item->hopitemname);
		Else	itemlen = col_count (item->itemname);
		Fi
		itemlen += 2 + 2 * menuborder;
		If itemlen > meniwidth
		Then	meniwidth = itemlen;
		Fi
	Done
	If menuxpos + meniwidth > XMAX - 1
	Then	meniwidth = XMAX + 1 - menuxpos;
	Fi

	/* top menu border, with menu name: */
	putborder_top (menuxpos, 0, meniwidth, menus [meni].menuname);

	/* text lines may have to be cleared: */
	menu_line = top_line;

	/* menu items: */
	For i = 0 While i < menulen Step i ++
	Do
	    prepare_menuline (1 + i, & menu_line);
	    item = & (menus [meni].menuitems [i]);
	    If item->itemfu == separator
	    Then
		putborder_middle (menuxpos, 1 + i, meniwidth);
	    Else
		gotoxy (menuxpos, 1 + i);
		If use_block_graphics == True
		Then	putblock ('x');
			menuborder_off ();
		Else	putchar (' ');
			menudisp_off ();
		Fi
		putnstr ("", menuborder);

		If hop_flag > 0 && item->hopitemname != NIL_PTR && item->hopitemname [0] != '\0'
		Then	putnstr (item->hopitemname, meniwidth - 2 - 2 * menuborder);
		Else	putnstr (item->itemname, meniwidth - 2 - 2 * menuborder);
		Fi

		putnstr ("", menuborder);
		If use_block_graphics == True
		Then	menuborder_on ();
			putblock ('x');
		Else	menudisp_on ();
			putchar (' ');
		Fi
	    Fi
	Done

	/* bottom menu border: */
	prepare_menuline (1 + menulen, & menu_line);
	putborder_bottom (menuxpos, 1 + menulen, meniwidth);

	gotoxy (menuxpos, 0);
	If menulen > last_dirty_line
	Then	last_dirty_line = menulen;
	Fi
	pulldownmenu_displayed = True;
	pulldownmenu_active = True;

	return meniwidth;
End


/***********************************************************************\
	display_menu () displays the given menu, returns menu width
\***********************************************************************/
local int
	display_menu (menu, menulen, column, line, title)
		itemtype * menu; int menulen;
		int column; int line;
		char * title;
Begin
	int i;
	itemtype * item;
	int menu_width = width_of_menu (menu, menulen);
	LINE * menu_line;

	menu_mouse_mode (True);

	/* save redraw values: */
	last_popup_column = column;
	last_popup_line = line;
	last_popup_item = -1;

	/* text lines may have to be cleared: */
	menu_line = proceed (top_line, line - 1);

	/* top menu border: */
	prepare_menuline (line, & menu_line);
	putborder_top (column, line, menu_width, title);

	/* menu items: */
	For i = 0 While i < menulen Step i ++
	Do
	    prepare_menuline (line + 1 + i, & menu_line);
	    item = & (menu [i]);
	    If item->itemfu == separator
	    Then
		putborder_middle (column, line + 1 + i, menu_width);
	    Else
		gotoxy (column, line + 1 + i);
		If use_block_graphics == True
		Then	putblock ('x');
			menuborder_off ();
		Else	putchar (' ');
			menudisp_off ();
		Fi
		putnstr ("", popupmenuborder);

		If hop_flag > 0 && item->hopitemname != NIL_PTR && item->hopitemname [0] != '\0'
		Then	putnstr (item->hopitemname, menu_width - 2 - 2 * popupmenuborder);
		Else	putnstr (item->itemname, menu_width - 2 - 2 * popupmenuborder);
		Fi

		putnstr ("", popupmenuborder);
		If use_block_graphics == True
		Then	menuborder_on ();
			putblock ('x');
		Else	menudisp_on ();
			putchar (' ');
		Fi
	    Fi
	Done

	/* bottom menu border: */
	prepare_menuline (line + 1 + menulen, & menu_line);
	putborder_bottom (column, line + 1 + menulen, menu_width);

	gotoxy (column, line);
	first_dirty_line = line - 1;
	last_dirty_line = first_dirty_line + menulen + 2;
	last_menu = menu;
	last_menulen = menulen;
	last_menuwidth = menu_width;
	last_menutitle = title;
	popupmenu_displayed = True;
	popupmenu_active = True;

	return menu_width;
End


/***********************************************************************\
	menselected () highlights a selected pulldown menu item
\***********************************************************************/
local void
	menselected (selected, meni, itemno)
		FLAG selected; int meni; int itemno;
Begin
	itemtype * item;
	int meniwidth = menu_width;
	int menulen = menus [meni].menulen;
	int itemlen;
	int menuxpos = meni * menu_width;
	int i;

	/* save redraw values: */
	If selected == True
	Then	last_pulldown_item = itemno;
	Else	last_pulldown_item = -1;
	Fi

	/* determine menu width: */
	For i = 0 While i < menulen Step i ++
	Do
		item = & (menus [meni].menuitems [i]);
		If hop_flag > 0 && item->hopitemname != NIL_PTR && item->hopitemname [0] != '\0'
		Then	itemlen = col_count (item->hopitemname);
		Else	itemlen = col_count (item->itemname);
		Fi
		itemlen += 2 + 2 * menuborder;
		If itemlen > meniwidth
		Then	meniwidth = itemlen;
		Fi
	Done
	If menuxpos + meniwidth > XMAX - 1
	Then	meniwidth = XMAX + 1 - menuxpos;
	Fi

	gotoxy (menuxpos, 1 + itemno);
	If use_block_graphics == True
	Then	putblock ('x');
	Else	menudisp_on ();
		putchar (' ');
		menudisp_off ();
	Fi
	If selected == True
	Then	menuheader_on ();
	Fi
	putnstr ("", menuborder);

	item = & (menus [meni].menuitems [itemno]);
	If hop_flag > 0 && item->hopitemname != NIL_PTR && item->hopitemname [0] != '\0'
	Then	putnstr (item->hopitemname, meniwidth - 2 - 2 * menuborder);
	Else	putnstr (item->itemname, meniwidth - 2 - 2 * menuborder);
	Fi

	putnstr ("", menuborder);
	If selected == True
	Then	menuheader_off ();
	Fi
	If use_block_graphics == True
	Then	putblock ('x');
	Else	menudisp_on ();
		putchar (' ');
		menudisp_off ();
	Fi
End


/***********************************************************************\
	popupmenselected () highlights a selected popup menu item
\***********************************************************************/
local void
	popupmenselected (menu, menu_width, selected, column, line, i)
		itemtype * menu; FLAG selected; int column; int line; int i;
Begin
	itemtype * item;

	/* save redraw values: */
	If selected == True
	Then	last_popup_item = i;
	Else	last_popup_item = -1;
	Fi

	gotoxy (column, line + 1 + i);
	If use_block_graphics == True
	Then	putblock ('x');
	Else	menudisp_on ();
		putchar (' ');
		menudisp_off ();
	Fi
	If selected == True
	Then	menuheader_on ();
	Fi
	putnstr ("", popupmenuborder);

	item = & (menu [i]);
	If hop_flag > 0 && item->hopitemname != NIL_PTR && item->hopitemname [0] != '\0'
	Then	putnstr (item->hopitemname, menu_width - 2 - 2 * popupmenuborder);
	Else	putnstr (item->itemname, menu_width - 2 - 2 * popupmenuborder);
	Fi

	putnstr ("", popupmenuborder);
	If selected == True
	Then	menuheader_off ();
	Fi
	If use_block_graphics == True
	Then	putblock ('x');
	Else	menudisp_on ();
		putchar (' ');
		menudisp_off ();
	Fi
End


/***********************************************************************\
	clean_menus () wipes menus which are still displayed
	clear_menus () clears menus which are still displayed
\***********************************************************************/
global void
	clean_menus ()
Begin
	LINE * start_line;

	menu_mouse_mode (False);
	If pulldownmenu_displayed == True
	Then
		pulldownmenu_displayed = False;
		display (0, top_line, last_dirty_line, last_dirty_line);
		last_dirty_line = 0;
		set_cursor_xy ();
	Fi
	If popupmenu_displayed == True
	Then
		start_line = proceed (top_line, first_dirty_line);
		popupmenu_displayed = False;
		display (first_dirty_line, start_line, 1 + last_menulen, first_dirty_line + 1 + last_menulen);
		last_dirty_line = 0;
		set_cursor_xy ();
	Fi
End

local void
	clear_menus ()
Begin
	clean_menus ();
	pulldownmenu_active = False;
	popupmenu_active = False;
End


/***********************************************************************\
	redrawmenu () redraws an open menu after screen refresh
\***********************************************************************/
global void
	redrawmenu ()
Begin
	If pulldownmenu_active == True
	Then
		int item = last_pulldown_item;
		pulldown_menu (last_pulldown_menu);
		If item >= 0
		Then	menselected (True, last_pulldown_menu, item);
		Fi
	Fi
	If popupmenu_active == True
	Then
		int item = last_popup_item;
		display_menu (last_menu, last_menulen, last_popup_column, last_popup_line, last_menutitle);
		If item >= 0
		Then	popupmenselected (last_menu, last_menuwidth, True, 
					last_popup_column, last_popup_line, 
					item);
		Fi
	Fi
End


/***********************************************************************\
	is_menu_open () checks if a menu is open
\***********************************************************************/
global int
	is_menu_open ()
{
	return pulldownmenu_active == True || popupmenu_active == True;
}


/***********************************************************************\
	MINMEN (screen_column, itemno) handles menu dialog or flag toggling
		if itemno > 0, the item is pre-selected
	openmenu (i) opens the i-th menu
	openmenuat (c) opens the menu at screen column c
\***********************************************************************/
global void
	openmenu (i) int i;
Begin
	MINMEN (i * menu_width, 0);
End

global void
	openmenuat (c) int c;
Begin
	If mouse_button == middlebutton && hop_flag == 0
	Then	toggleHOP ();
	Fi

	MINMEN (c, 0);
End

global void
	MINMEN (column, itemno) int column; int itemno;
Begin
	int meni;
	character c;
	int popmenwidth;	/* width of the menu, determined dynamically */

	calcmenuvalues ();

	If menu_width < 3
	Then	return;
	Fi
	meni = column / menu_width;
	If meni < 0 || meni >= number_menus ()
	Then	/* check if it's a flag toggle position */
		meni = column - flags_pos;
		If meni >= 0 && meni < flags_displayed
		Then	(* Flagmenu [meni].toggle) (column);
			displayflags ();
			set_cursor_xy ();
		Fi
		return;
	Fi

	popmenwidth = pulldown_menu (meni);
	If itemno > 0
	Then	menselected (True, meni, itemno - 1);
	Fi

	flush ();
	c = readchar_allbuttons ();
	Dowhile command (c) == MDN || command (c) == MUP
		|| (command (c) == MOUSEescape
			&& mouse_button != leftbutton
			&& (mouse_button != releasebutton ||
			    mouse_ypos < 0 ||
			    ! (mouse_prevbutton == leftbutton 
			       || mouse_prevbutton == rightbutton
			      )
			   )
			&& mouse_button != middlebutton
			&& (mouse_button != rightbutton ||
			    mouse_ypos >= 0 || mouse_xpos / menu_width == meni)
			&& (mouse_button != movebutton ||
			    mouse_ypos >= 0 || mouse_xpos / menu_width == meni)
		   )
	Do
	    If command (c) == MOUSEescape
	    Then
		If itemno > 0
		Then	menselected (False, meni, itemno - 1);
		Fi
		itemno = mouse_ypos + 1;
		If itemno <= 0 || itemno > menus [meni].menulen
		|| menus [meni].menuitems [itemno - 1].itemfu == separator
		Then	itemno = 0;
		Else
			menselected (True, meni, itemno - 1);
		Fi
	    Else
		If itemno > 0
		Then	menselected (False, meni, itemno - 1);
		Fi
		If command (c) == MDN
		Then	itemno ++;
			If menus [meni].menuitems [itemno - 1].itemfu == separator
			Then	itemno ++;
			Fi
			If itemno > menus [meni].menulen
			Then	itemno = 1;
			Fi
		Else	itemno --;
			If menus [meni].menuitems [itemno - 1].itemfu == separator
			Then	itemno --;
			Fi
			If itemno <= 0
			Then	itemno = menus [meni].menulen;
			Fi
		Fi
		menselected (True, meni, itemno - 1);
	    Fi
	    flush ();
	    c = readchar_allbuttons ();
	Done
	If c == ' ' || command (c) == HOP || command (c) == GOTO
		|| (command (c) == MOUSEescape && mouse_button == middlebutton)
	Then	toggleHOP ();
		clear_menus ();
		displaymenuline_from (meni);
		MINMEN (column, itemno);
	Elsif command (c) == MOUSEescape
	Then	/* mouse click */
		If mouse_xpos >= menu_width * meni
		&& mouse_xpos < menu_width * meni + popmenwidth
		&& mouse_ypos >= 0 && mouse_ypos < menus [meni].menulen
		&& (mouse_button == leftbutton || mouse_button == releasebutton)
		Then
			menselected (True, meni, mouse_ypos);
			(* menus [meni].menuitems [mouse_ypos].itemfu)
				(menus [meni].menuitems [mouse_ypos].hopitemname);
		Elsif mouse_ypos == -1 && mouse_xpos / menu_width != meni
		Then	clear_menus ();
		/*	displaymenuheaders ();	*/
			displaymenuline_from (meni);
			MINMEN (mouse_xpos, 0);
		Fi
	Elsif (c == '\n' || command (c) == SNL) && itemno > 0
	Then	(* menus [meni].menuitems [itemno - 1].itemfu)
			(menus [meni].menuitems [itemno - 1].hopitemname);
	Elsif command (c) == MLF || command (c) == MRT
	Then	clear_menus ();
		displaymenuline_from (meni);
		If command (c) == MLF
		Then	If meni > 0
			Then	meni --;
			Else	meni = number_menus () - 1;
			Fi
		Else	If meni < number_menus () - 1
			Then	meni ++;
			Else	meni = 0;
			Fi
		Fi
		openmenu (meni);
	Fi
	clear_menus ();
/*	displaymenuheaders ();	*/
/*	displaymenuline_from (meni);	*/
	displaymenuline ();
End


/***********************************************************************\
	QUICKMEN () handles right mouse button menu
	popup_menu () handles given menu
\***********************************************************************/
global void
	popup_menu (menu, menulen, column, line, title)
		itemtype * menu; int menulen;
		int column; int line;
		char * title;
Begin
	character c;
	int startline;
	int startcolumn;
	int itemno;

	int menu_width = width_of_menu (menu, menulen);

	startline = line + 1;
	If startline > YMAX - menulen - 1
	Then	startline = YMAX - menulen - 1;
	Fi
	startcolumn = column;
	If startcolumn > XMAX - menu_width
	Then	startcolumn = XMAX - menu_width;
	Fi
	If startline < 1 || startcolumn < 0 || menu_width < 3
	Then	return;
	Fi

	menu_width = display_menu (menu, menulen, startcolumn, startline, title);
	itemno = 0;
	flush ();

	c = readchar_allbuttons ();
	Dowhile command (c) == MDN || command (c) == MUP
		|| c == ' ' || command (c) == HOP || command (c) == GOTO
		|| (command (c) == MOUSEescape
			&& mouse_button != leftbutton
			&& (mouse_button != releasebutton
			    || ! (mouse_prevbutton == leftbutton 
			          || mouse_prevbutton == rightbutton
			         )
			    || ! (
				  mouse_xpos > startcolumn
				   && mouse_xpos < startcolumn + menu_width
				  && mouse_ypos >= startline
				   && mouse_ypos < startline + menulen
			         )
			   )
		   )
	Do
	    If c == ' ' || command (c) == HOP || command (c) == GOTO
		  || (command (c) == MOUSEescape
			&& mouse_button == middlebutton
		     )
	    Then
		toggleHOP ();
		clear_menus ();
		menu_width = display_menu (menu, menulen, startcolumn, startline, title);
		If itemno > 0
		Then	popupmenselected (menu, menu_width, True, startcolumn, startline, itemno - 1);
		Fi
	    Elsif command (c) == MOUSEescape
	    Then
		If itemno > 0
		Then	popupmenselected (menu, menu_width, False, startcolumn, startline, itemno - 1);
		Fi
		itemno = mouse_ypos - startline + 1;
		If itemno <= 0 || itemno > menulen
		|| menu [itemno - 1].itemfu == separator
		Then	itemno = 0;
		Else
			popupmenselected (menu, menu_width, True, startcolumn, startline, itemno - 1);
		Fi
	    Else
		If itemno > 0
		Then	popupmenselected (menu, menu_width, False, startcolumn, startline, itemno - 1);
		Fi
		If command (c) == MDN
		Then	itemno ++;
			If menu [itemno - 1].itemfu == separator
			Then	itemno ++;
			Fi
			If itemno > menulen
			Then	itemno = 1;
			Fi
		Else	itemno --;
			If menu [itemno - 1].itemfu == separator
			Then	itemno --;
			Fi
			If itemno <= 0
			Then	itemno = menulen;
			Fi
		Fi
		popupmenselected (menu, menu_width, True, startcolumn, startline, itemno - 1);
	    Fi
	    flush ();
	    c = readchar_allbuttons ();
	Done
	If command (c) == MOUSEescape
	Then	/* mouse click */
		mouse_xpos -= startcolumn;
		mouse_ypos -= startline;
		If mouse_xpos > 0 && mouse_xpos < menu_width
		&& mouse_ypos >= 0 && mouse_ypos < menulen
		Then
			popupmenselected (menu, menu_width, True, startcolumn, startline, mouse_ypos);
			(* menu [mouse_ypos].itemfu)
				(menu, mouse_ypos);
		Elsif mouse_ypos + startline == -1
		Then	clear_menus ();
			displaymenuheaders ();
			MINMEN (mouse_xpos + startcolumn, 0);
		Fi
	Elsif (c == '\n' || command (c) == SNL) && itemno > 0
	Then
		(* menu [itemno - 1].itemfu)
			(menu, itemno - 1);
	Fi
	clear_menus ();
End


global void
	QUICKMEN (column, line) int column; int line;
Begin
	popup_menu (Popupmenu, arrlen (Popupmenu), column, line, NIL_PTR);
End


/***********************************************************************\
	End
\***********************************************************************/
