/*======================================================================*\
|*		Editor mined						*|
|*		auxiliary functions					*|
\*======================================================================*/

#include "mined.h"

#undef XBREAK_bottom
#define XBREAK_bottom (XMAX - 1)

/*#define disp_combining_at_basechar_width*/


FLAG pagewrapped = False;	/* Did output on the bottom line wrap and scroll? */
FLAG char_on_status_line = False; /* is output active on status line ? */


/*======================================================================*\
|*			UTF-8 and special indication handling		*|
\*======================================================================*/

void put_unichar ();

#define illegal_UTF indicate_UTF
static
void
indicate_UTF (c)
  character c;
{
  if (char_on_status_line == True) {
	reverse_off ();
  }
  unidisp_on ();
  put_unichar ((unsigned long) c);	/* needed for Turbo-C */
  unidisp_off ();
  if (char_on_status_line == True) {
	reverse_on ();
  }
}

static
void
indicate_special (c)
  character c;
{
  if (char_on_status_line == True) {
	reverse_off ();
  }
  unimarkdisp_on ();
  put_unichar ((unsigned long) c);	/* needed for Turbo-C */
  unimarkdisp_off ();
  if (char_on_status_line == True) {
	reverse_on ();
  }
}

static int utfcount = 0;
static unsigned long unichar;

/*
 * put Unicode character to screen;
 * put as UTF-8 on UTF-8 terminal, or as Latin-1 otherwise
	7 bit	0xxxxxxx
	8..11	110xxxxx 10xxxxxx
	12..16	1110xxxx 10xxxxxx 10xxxxxx
 */
void
put_unichar (unichar)
  unsigned long unichar;
{
  if (utf8_screen) {
	if (unichar < 0x80) {
		__putchar (unichar);
	} else if (unichar < 0x800) {
		__putchar (0xC0 | (unichar >> 6));
		__putchar (0x80 | (unichar & 0x3F));
	} else if (unichar < 0x10000) {
		__putchar (0xE0 | (unichar >> 12));
		__putchar (0x80 | ((unichar >> 6) & 0x3F));
		__putchar (0x80 | (unichar & 0x3F));
	} else if (unichar < 0x200000) {
		__putchar (0xF0 | (unichar >> 18));
		__putchar (0x80 | ((unichar >> 12) & 0x3F));
		__putchar (0x80 | ((unichar >> 6) & 0x3F));
		__putchar (0x80 | (unichar & 0x3F));
	} else if (unichar < 0x4000000) {
		__putchar (0xF8 | (unichar >> 24));
		__putchar (0x80 | ((unichar >> 18) & 0x3F));
		__putchar (0x80 | ((unichar >> 12) & 0x3F));
		__putchar (0x80 | ((unichar >> 6) & 0x3F));
		__putchar (0x80 | (unichar & 0x3F));
	} else if (unichar < 0x80000000) {
		__putchar (0xFC | (unichar >> 30));
		__putchar (0x80 | ((unichar >> 24) & 0x3F));
		__putchar (0x80 | ((unichar >> 18) & 0x3F));
		__putchar (0x80 | ((unichar >> 12) & 0x3F));
		__putchar (0x80 | ((unichar >> 6) & 0x3F));
		__putchar (0x80 | (unichar & 0x3F));
	}
  } else {
	if (unichar < 0x100) {
		__putchar (unichar);
	} else {
		indicate_UTF (UNI_MARK);
	}
  }
}

static
void
put_uniend ()
{
  if (utf8_text && utfcount > 0) {
	illegal_UTF ('');
	utfcount = 0;
  }
}

/*
 * put_char () puts a character byte to the screen (via buffer)
 * If UTF-8 text is being output through put_char () but the screen 
 * is not in UTF-8 mode, put_char () transforms the byte sequences.
 *
 * putchar (c) is __putchar (c)
 * __putchar (c) does (void) writeout (c) (except with curses)
 */
void
put_char (c)
  character c;
{
  if (utf8_text) {
	if (c < 0x80) {
		if (utfcount > 0) {
			/* previous character not terminated properly */
			illegal_UTF ('');
			utfcount = 0;
		}
		__putchar (c);
	} else if ((c & 0xC0) == 0x80) {
		if (utfcount == 0) {
			illegal_UTF ('8');
			return;
		}
		unichar = (unichar << 6) | (c & 0x3F);
		utfcount --;

		if (utfcount == 0) {
			/* final UTF-8 byte */
			if (unichar < ' ') {
				indicate_UTF (unichar + '@');
			} else if (unichar == 0x7F) {
				indicate_UTF ('?');
			} else if (unichar >= 0x80 && unichar <= 0x9F) {
				indicate_UTF (unichar + '@');
			} else if (unichar == 0xA0) {
				indicate_special ('');
			} else if (utf8_screen) {
				put_unichar (unichar);
			} else if ((unichar & 0xFF) == unichar) {
				__putchar (unichar & 0xFF);
			} else if (iscombining (unichar)) {
				indicate_UTF ('\'');
				if (iswide (unichar)) {
					indicate_UTF (' ');
				}
			} else {
				indicate_UTF (UNI_MARK);
				if (iswide (unichar)) {
					indicate_UTF (' ');
				}
			}
		} else {
			/* character continues */
			return;
		}
	} else { /* first UTF-8 byte */
		if (utfcount > 0) {
			/* previous character not terminated properly */
			illegal_UTF ('');
			utfcount = 0;
		}
		if ((c & 0xE0) == 0xC0) {
			utfcount = 2;
			unichar = c & 0x1F;
		} else if ((c & 0xF0) == 0xE0) {
			utfcount = 3;
			unichar = c & 0x0F;
		} else if ((c & 0xF8) == 0xF0) {
			utfcount = 4;
			unichar = c & 0x07;
		} else if ((c & 0xFC) == 0xF8) {
			utfcount = 5;
			unichar = c & 0x03;
		} else if ((c & 0xFE) == 0xFC) {
			utfcount = 6;
			unichar = c & 0x01;
		} else {
			/* illegal UTF-8 code 254 (0xFE) or 255 (0xFF) */
			illegal_UTF ('4' + (c & 1));
			return;
		}
		utfcount --;
		return;
	}
  } else
	__putchar (c);
}

/*
 * putcharacter (c) puts one non-UTF character to the screen
 */
static
void
putcharacter (c)
  character c;
{
  if (utf8_screen && ! utf8_text) {
	put_unichar (c);
  } else {
	put_char (c);
  }
}

/*
 * put 1 UTF-8 character (from string)
 */
void
put_utfchar (s)
  char * s;
{
  unsigned long unichar;
  int follow;

  utf8_info (s, & follow, & unichar);
  put_unichar (unichar);
}

/*
 * put mark character, possibly switching alternate character set
 * putmarkmode and endmarkmode are only called by putmark and for 
 * display of TAB markers
 */
FLAG dim_mode = False;
FLAG mark_mode = False;
FLAG use_extracset = False;

static
void
putmarkmode (c, utfmark)
  char c;
  char * utfmark;
{
  if (dim_mode == False) {
	dimon ();
	dim_mode = True;
  }

  if (utf8_screen == True) {
	if (utf8_text == True && utfmark != NIL_PTR && * utfmark != '\0') {
		put_utfchar (utfmark);
	} else {
		put_unichar (c & 0xFF);
	}
  }
  else if ((c & '\200') != '\0') {
	if (mark_mode == True) {
		end_extracset ();
		mark_mode = False;
	}
	putchar (c);
  }
  else {
	if (mark_mode == False && use_extracset == True) {
		start_extracset ();
		mark_mode = True;
	}
	putchar (c);
  }
}

static
void
endmarkmode ()
{
  if (dim_mode == True) {
	dimoff ();
	dim_mode = False;
  }
  if (mark_mode == True) {
	end_extracset ();
	mark_mode = False;
  }
}

static
void
putmark (c, utfmark)
  char c;
  char * utfmark;
{
  putmarkmode (c, utfmark);
  endmarkmode ();
}

static
void
putChinmark (mark)
  char * mark;
{
	boldon ();
	__putchar (mark [0]);
	__putchar (mark [1]);
	boldoff ();
}

static
void
putUmark (c, utfmark)
  char c;
  char * utfmark;
{
  unimarkdisp_on ();
  if (utf8_screen == True) {
	if (utf8_text == True && utfmark != NIL_PTR && * utfmark != '\0') {
		put_utfchar (utfmark);
	} else {
		put_unichar (c & 0xFF);
	}
  } else {
	putchar (c);
  }
  unimarkdisp_off ();
}

static
void
putret (type)
  unsigned char type;
{
/*	Latin-1	DOS mapping table for some useful special characters
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
		
*/
  if (type == lineend_LF) {
	putmark (RET_MARK, UTF_RET);
  } else if (type == lineend_CRLF) {
	/* MSDOS line separator */
	putmark (DOSRET_MARK, UTF_DOSRET);
  /* other marks:  */
  } else if (type == lineend_CR) {
	/* Mac line separator */
#ifdef pc_charset
	putmark ('', "¦");
#else
/*	putmark ('', "¥");	*/
/*	putmark ('', "¦");	*/
	putmark ('@', "@");
#endif
  } else if (type == lineend_NONE) {
#ifdef pc_charset
	putmark ('', "¬");
#else
	putmark ('', "¬");
#endif
  } else if (type == lineend_NUL) {
#ifdef pc_charset
	putmark ('', "º");
#else
	putmark ('', "º");
#endif
  } else if (type == lineend_LS) {
	/* Unicode line separator 2028:   */
	putUmark (RET_MARK, UTF_RET);
  } else if (type == lineend_PS) {
	/* Unicode paragraph separator 2029:   */
	putUmark (PARA_MARK, UTF_PARA);
  } else {
#ifdef pc_charset
	putmark ('', "¤");
#else
	putmark ('', "¤");
#endif
  }
}

static
int
iscontrol (c)
  character c;
{
#ifdef pc_charset
	return ((c) == '\177') 
		|| ((character) c < (character) ' ');
#else
  if (utf8_text == True || Chinese == True) {
	return ((c) == '\177') 
		|| ((character) c < (character) ' ');
  } else {
	return ((c) == '\177') 
		|| ((character) ((c) & '\177') < (character) ' ');
  }
#endif
}

static
character controlchar (c)
  character c;
{
  if ((c) == '\177') {
	return '?';
  } else {
	return (c) + '@';
  }
}


/*======================================================================*\
|*			Output						*|
\*======================================================================*/

/*
 * Bad_write () is called when a write failed. Notify the user.
 */
void
bad_write (fd)
  int fd;
{
  if (fd == output_fd) {	/* Cannot write to terminal? */
	panicio ("Write error on terminal", serror ());
  }

  clear_buffer (); /* out_count = 0; */
  ring_bell ();
  error ("Write aborted (File incomplete): ", serror ());
}

/*
 * Flush the I/O buffer on filedescriptor fd.
flush () is (void) flush_buffer (output_fd) unless with curses
 */
int
flush_buffer (fd)
  int fd;
{
  if (out_count <= 0) {		/* There is nothing to flush */
	return FINE;
  }
#ifdef conio
  if (fd == output_fd) {
	cputs (screen);
  }
#else
#ifdef BorlandC
  if (fd == output_fd) {
	screen [out_count] = '\0';
/* don't ask me why that crazy compiler doesn't work with write () below */
	printf ("%s", screen);
  }
#endif
#endif
  else {
	char * writepoi = screen;
	int written = 0;
	int none_count = 0;
/*	int less_count = 0;	*/

	while (out_count > 0) {
		written = write (fd, writepoi, out_count);
		if (written == -1) {
			bad_write (fd);
			return ERRORS;
		}
/*
		if (written != out_count) {
			less_count ++;
		}
		if (less_count > 0) {
fprintf (stderr, "   written %d/%d     \n", written, out_count); sleep (2);
		}
*/
		if (written == 0) {
			none_count ++;
			if (none_count > 20) {
				bad_write (fd);
				return ERRORS;
			}
		}
		out_count -= written;
		writepoi += written;
	}
  }
  clear_buffer (); /* Empty buffer: out_count = 0; */
  return FINE;
}

/*
 * writeout does a buffered output to screen.
 */
int
writeout (c)
  char c;
{
  char clow;

  if (c == '\n') {
	if (writeout ('\015') == ERRORS) {
		return ERRORS;
	}
  }
  if (translate_output == True) {
	if ((c & '\200') != '\0') {
		altcset_on ();
		clow = (c & '\177');
		if ((int) clow < translen) {
			writeout (transout [(int) clow]);
		} else {
			writeout (clow);
		}
		altcset_off ();
		return FINE;
	}
  }
  screen [out_count ++] = c;
  if (out_count == screen_BUFL) {
	return flush_buffer (output_fd);
  }
#ifdef DEBUG
  flush ();
#endif
  return FINE;
}

/*
 * Putoutstring writes the given string out to the terminal.
 * (buffered via writechar via misused screen buffer!)
putstring (str) is putoutstring (str) unless with curses
 */
void
putoutstring (text)
  register char * text;
{
  while (* text != '\0') {
	(void) writeout (* text ++);
  }
}

/*
 * Print string on terminal, printing controls with ^.
 */
static int lpos = 0;
static int bottom_lines = 0;
static int print_utfcount = 0;
static unsigned long print_unichar;

static
void
print_char (c)
  register character c;
{
  char_on_status_line = True;

  if (lpos > XBREAK_bottom - 1) {
	putstring ("\r\n");
	lpos = 0;
	pagewrapped = True;
	bottom_lines ++;
  }

  if (iscontrol (c)) {
	put_char ('^');
	lpos ++;
	putcharacter (controlchar (c));
	lpos ++;
  } else if (utf8_text == True) {
	put_char (c);

	/* trace sequences of UTF-8 bytes through this function 
	   for the sole purpose of adjusting lpos for wide characters;
	   increase lpos at the end of each character accordingly */
	if (c < 0x80) {
		if (print_utfcount > 0) {
			/* illegal UTF-8 terminated */
			lpos ++;
			print_utfcount = 0;
		}
		lpos ++;
	} else if ((c & 0xC0) == 0x80) {
	    if (print_utfcount == 0) {
			/* illegal UTF-8 */
			lpos ++;
	    } else {
		print_unichar = (print_unichar << 6) | (c & 0x3F);
		print_utfcount --;

		if (print_utfcount == 0) {
			/* final UTF-8 byte */
			lpos += uniwidth (print_unichar);
		} else {
			/* character continues */
		}
	    }
	} else { /* first UTF-8 byte */
		if (print_utfcount > 0) {
			/* illegal UTF-8 terminated */
			lpos ++;
			print_utfcount = 0;
		}
		if ((c & 0xE0) == 0xC0) {
			print_utfcount = 2;
			print_unichar = c & 0x1F;
		} else if ((c & 0xF0) == 0xE0) {
			print_utfcount = 3;
			print_unichar = c & 0x0F;
		} else if ((c & 0xF8) == 0xF0) {
			print_utfcount = 4;
			print_unichar = c & 0x07;
		} else if ((c & 0xFC) == 0xF8) {
			print_utfcount = 5;
			print_unichar = c & 0x03;
		} else if ((c & 0xFE) == 0xFC) {
			print_utfcount = 6;
			print_unichar = c & 0x01;
		} else {	/* illegal UTF-8 code */
			print_utfcount = 1;
			lpos ++;
		}
		print_utfcount --;
	}
  } else {
	putcharacter (c);
	lpos ++;
  }

  char_on_status_line = False;
}

static
void
printlim_string (text, limit)
  register char * text;
  register int limit;
{
  int utfcount;
  unsigned long unichar;
  unsigned long combipref;
  int charwidth;

  lpos = 0;
  char_on_status_line = True;
  while (* text != '\0') {
	if (utf8_text == True) {
		utf8_info (text, & utfcount, & unichar);
		charwidth = uniwidth (unichar);

		combipref = '\0';
		if (iscombining (unichar)) {
			if (combining_mode == False 
				&& combining_screen == True)
			{	/* enforce separated display */
			/*	disp_separate_combining = True;	*/
			/*	combiningdisp_on ();		*/

#ifdef disp_combining_at_basechar_width
				if (basechar_width == 2)
#else	/* display on its own virtual width property */
				if (iswide (unichar))
#endif
				{
					combipref = 0x3000;
					charwidth = 2;
				} else {
					combipref = ' ';
					charwidth = 1;
				}
			} else if (combining_screen == False) {
			/*	disp_separate_combining = True;	*/
			/*	combiningdisp_on ();		*/
			}
		} else {
#ifdef disp_combining_at_basechar_width
			basechar_width = charwidth;
#endif
		}
	} else if (iscontrol (* text)) {
		charwidth = 2;
	} else if (Chinese == True && multichar (* text)) {
		charwidth = 2;
	} else {
		charwidth = 1;
	}

	if (lpos > limit - charwidth) {
		putmark (SHIFT_MARK, NIL_PTR);
		break;
	}

	if (iscontrol (* text)) {
		lpos += 2;
		put_char ('^');
		putcharacter (controlchar (* text ++));
	} else {
		lpos += charwidth;
		if (utf8_text == True) {
			if (combipref != '\0') {
				put_unichar (combipref);
			}

			putcharacter (* text ++);
			utfcount --;
			while (utfcount > 0 && (* text & 0xC0) == 0x80) {
				putcharacter (* text ++);
				utfcount --;
			}
		} else if (Chinese == True && multichar (* text)) {
			putcharacter (* text ++);
			if (* text != '\0') {
				putcharacter (* text ++);
			}
		} else {
			putcharacter (* text ++);
		}
	}
  }
  char_on_status_line = False;
}

static
void
print_string (text)
  register char * text;
{
  while (* text != '\0') {
	print_char (* text ++);
  }
}

static
void
put_blanks (endpos)
  int endpos;
{
  int startpos = 0;
  while (startpos ++ <= endpos) {
	putchar (' ');
  }
}

void
clear_wholeline ()
{
  if (can_clear_eol == True) {
	clear_eol ();
  } else {
	put_blanks (XMAX);
  }
}

void
clear_lastline ()
{
  if (can_clear_eol == True) {
	clear_eol ();
  } else {
	put_blanks (XMAX - 1);
  }
}


/*======================================================================*\
|*			Text display (buffer oriented output)		*|
\*======================================================================*/

static
void
disp_syntax (syntax_marker_was, syntax_marker)
  char syntax_marker_was;
  char syntax_marker;
{
  if (syntax_marker != syntax_marker_was) {
	if (syntax_marker & syntax_JSP) {
		dispHTML_jsp ();
	} else if (syntax_marker & syntax_comment) {
		dispHTML_comment ();
	} else if (syntax_marker & syntax_HTML) {
		dispHTML_code ();
	} else {
		dispHTML_off ();
	}
  }
}

#define default_script -999

static
void
disp_script (script_colour)
  int script_colour;
{
  if (script_colour == default_script) {
	disp_normal ();
  } else {
	disp_colour (script_colour);
  }
}

static
struct colour_mapping {
	char * scriptname;
	int colour;
} colour_table [] =
{
#include "colours.h"
};

/**
   Determine display colour for character script.
 */
static
int
map_script_colour (scriptname)
  char * scriptname;
{
  int min = 0;
  int max = sizeof (colour_table) / sizeof (struct colour_mapping) - 1;
  int mid;
  int cmp;

  /* binary search in table */
  while (max >= min) {
    mid = (min + max) / 2;
    cmp = strcmp (colour_table [mid].scriptname, scriptname);
    if (cmp < 0) {
      min = mid + 1;
    } else if (cmp > 0) {
      max = mid - 1;
    } else {
      return colour_table [mid].colour;
    }
  }

  return default_script;
}

/*
 * Put_line prints the given line on the standard output.
 * If offset is not zero, printing will start at that x-coordinate.
 * If the FLAG clear_line is True, then the screen line will be cleared
 * when the end of the line has been reached.
line_print (line) is put_line (line, 0, True, False)
 * put_line is directly called only by S () and delete_text ()
 */
void
put_line (line, offset, clear_line, positioning)
  LINE * line;		/* Line to print */
  int offset;		/* Offset to start if positioning == False */
			/* position if positioning == True */
  FLAG clear_line;	/* Clear to eoln if True */
  FLAG positioning;	/* positioning inside line if True (for prop. fonts) */
{
  char * textp = line->text;
  char * parap;
  int count = get_shift (line->shift_count) * - SHIFT_SIZE;
  int count_ini = count;
  int tab_count;			/* Used in tab expansion */
  int tab_mid;
  int offset_start;
  int offset_stop;
  character c;
  unsigned long unichar;
  int follow;
  int charwidth;
#ifdef disp_combining_at_basechar_width
  int basechar_width = 0;
#endif
  FLAG disp_separate_combining;
  char syntax_marker = line->prev->syntax_marker;
  char syntax_marker_was = syntax_none;
  char syntax_marker_new = syntax_marker;
  char syntax_byte_printed;
  int script_colour = default_script;
  int last_script_colour;

  if (positioning == True) {
	offset_start = 0;
	offset_stop = offset;
  } else {
	offset_start = offset;
	offset_stop = XBREAK;
  }

/* Skip all chars as indicated by the offset_start and the shift_count field */
  while (count < offset_start) {
	syntax_marker_new = syntax_state (syntax_marker_new, textp);
	if (* textp == '<' || * textp == '>') {
		syntax_marker_was = syntax_marker;
		syntax_marker = syntax_marker_new;
	}
	advance_char_col (& textp, & count);
  }

  if (count == 0 && count_ini < 0 && SHIFT_BEG != '\0') {
	putmark (SHIFT_BEG, NIL_PTR);
	if (! is_tab (* textp)) {
		syntax_marker_new = syntax_state (syntax_marker_new, textp);
		if (* textp == '<' || * textp == '>') {
			syntax_marker_was = syntax_marker;
			syntax_marker = syntax_marker_new;
		}
		advance_char_col (& textp, & count);
	} else {
		count ++;
	}
  }
/* Check for shift-out left across double-width character boundary */
  if (count_ini < 0 && count > offset_start) {
	putmark (TABchar, NIL_PTR);
  }

  if (utf8_text == True) {
	utf8_info (textp, & follow, & unichar);
	charwidth = uniwidth (unichar);
  } else if (Chinese == True && multichar (* textp)) {
	charwidth = 2;
  } else {
	charwidth = 1;
  }

  if (dim_HTML == True) {
	disp_syntax (0, syntax_marker);
  }

  while (* textp != '\n' && count < offset_stop + 1 - charwidth) {
	if (dim_HTML == True) {
		syntax_marker_new = syntax_state (syntax_marker_new, textp);
		if (* textp == '<' || * textp == '>') {
			syntax_marker_was = syntax_marker;
			syntax_marker = syntax_marker_new;
		}
		if (* textp == '<') {
			disp_syntax (syntax_marker_was, syntax_marker);
		}
	}

	if (is_tab (* textp)) {		/* Expand tabs to spaces */
		tab_count = tab (count);
		tab_mid = count + (tab_count - count) / 2 + 1;
		if (Chinese == True) {
			if ((count & 1) == 1) {
				count ++;
				__putchar (' ');
			}
			while (count < offset_stop && count < tab_count) {
				count ++;
				count ++;
				putChinmark ("");
			}
		} else if (TABchar0 != '\0') {
			if (TABchar0 != TABchar) {
				count ++;
				putmarkmode (TABchar0, UTF_TAB0);
			}
			while (count < offset_stop && count < tab_count - 1) {
				count ++;
				putmarkmode (TABchar, UTF_TAB);
			}
			if (count < offset_stop && count < tab_count) {
				count ++;
				if (TABchar2 != '\0') {
					putmarkmode (TABchar2, UTF_TAB2);
				} else {
					putmarkmode (TABchar, UTF_TAB);
				}
			}
		}
		else while (count < offset_stop && count < tab_count) {
			count ++;
			if (count == tab_mid && TABcharmid != '\0') {
				putmarkmode (TABcharmid, UTF_TABmid);
			} else {
				putmarkmode (TABchar, UTF_TAB);
			}
		}
		endmarkmode ();
		if (dim_HTML == True) {
			disp_syntax (0, syntax_marker);
		}

		textp ++;
	}
	else if (iscontrol (* textp)
		|| ((character) * textp == (character) 0xA0 && utf8_text == False)
		)
	{
		put_uniend ();
		if (use_block_graphics == True && enable_block_graphics == True
		    && (
			   * textp == 1
			|| (* textp >= 0x0B && * textp <= 0x0F)
			|| * textp == 0x12
			|| (* textp >= 0x15 && * textp <= 0x19)
		)) {
			altcset_on ();
			putcharacter (* textp + 0x5F);
			altcset_off ();
		} else if ((character) * textp == (character) 0xA0) {
			unimarkdisp_on ();
			putcharacter ('');
			unimarkdisp_off ();
		} else {
			ctrldisp_on ();
			putcharacter (controlchar (* textp));
			ctrldisp_off ();
		}
		if (dim_HTML == True) {
			disp_syntax (0, syntax_marker);
		}

		textp ++;
		count ++;
	}
	else {
		syntax_byte_printed = * textp;

		if (utf8_text == True) {
			follow --;
			c = * textp ++;

			last_script_colour = script_colour;
			script_colour = map_script_colour (script (unichar));
			if (script_colour != last_script_colour) {
				disp_script (script_colour);
			}

			disp_separate_combining = False;
			if (iscombining (unichar)) {
				if (combining_mode == False 
					&& combining_screen == True)
				{	/* enforce separated display */
					disp_separate_combining = True;
					combiningdisp_on ();

#ifdef disp_combining_at_basechar_width
					if (basechar_width == 2)
#else	/* display on its own virtual width property */
					if (iswide (unichar))
#endif
					{
						put_unichar (0x3000);
						count += 2;
					} else {
						put_char (' ');
						count += 1;
					}
				}
				else {
					if (combining_screen == False) {
						disp_separate_combining = True;
						combiningdisp_on ();
						count += charwidth;
					}
				}
			}
			else {
#ifdef disp_combining_at_basechar_width
				basechar_width = charwidth;
#endif
				count += charwidth;
			}

			putcharacter (c);
			while (follow > 0 && (* textp & 0xC0) == 0x80) {
				put_char (* textp ++);
				follow --;
			}
			if (disp_separate_combining == True) {
				combiningdisp_off ();
				if (dim_HTML == True) {
					disp_syntax (0, syntax_marker);
				}
			}

		} else if (Chinese == True) {
			c = * textp ++;
			if (multichar (c)) {
				__putchar (c);
				c = * textp;
				if (c != '\n') {
					textp ++;
				}
				count ++;
			}
			__putchar (c);
			count ++;
		} else {
			c = * textp ++;
			putcharacter (c);
			count ++;
		}

		if (dim_HTML == True && syntax_byte_printed == '>') {
			disp_syntax (syntax_marker_was, syntax_marker);
		}
	}

	if (utf8_text == True) {
		utf8_info (textp, & follow, & unichar);
		charwidth = uniwidth (unichar);
	} else if (Chinese == True && multichar (* textp)) {
		charwidth = 2;
	} else {
		charwidth = 1;
	}
  }
  put_uniend ();

  if (script_colour != default_script) {
	disp_script (default_script);
  }

  if (dim_HTML == True) {
	disp_syntax (syntax_marker, 0);
  }

  if (positioning == True) {
	/* self-made cursor for terminals (such as xterm)
	   which have display problems with proportional screen fonts
	   and their cursor */
	reverse_on ();
	if (* textp != '\n') {
		if (utf8_text == True) {
			put_utfchar (textp);
			put_uniend ();
		} else {
			putcharacter (* textp);
		}
	} else if (Chinese == True) {
		putChinmark ("");
	} else if (RET_MARK != '\0') {
		if (PARA_MARK != '\0' && paradisp == True) {
			parap = textp;
			parap --;
			if (* parap == ' ' || * parap == '\t') {
				putret (line -> return_type);
			} else {
				putmark (PARA_MARK, UTF_PARA);
			}
		} else {
			putret (line -> return_type);
		}
	} else {
		putchar (' ');
	}
	reverse_off ();
	set_cursor (0, YMAX);
  }
  else /* (positioning == False) */ {
	/* If line is longer than XBREAK chars, print the shift_mark */
	if (count <= XBREAK && * textp != '\n') {
		putmark (SHIFT_MARK, NIL_PTR);
		count ++;
		if (count == XBREAK) {
			putchar (' ');
			count ++;
		}
	}

	/* Mark end of line if so desired */
	if (* textp == '\n') {
	    if (Chinese == True) {
		putChinmark ("");
		/* to assure it fits scrollbar_width is set to 1 
		   in Chinese mode even if scrollbar is switched off */
	    } else if (RET_MARK != '\0') {
		if (PARA_MARK != '\0' && paradisp == True) {
			parap = textp;
			parap --;
			if (* parap == ' ') {
				putret (line -> return_type);
			} else {
				putmark (PARA_MARK, UTF_PARA);
			}
		} else {
			putret (line -> return_type);
		}

		count ++;
		if (RET_BLANK) {
			while (count < XBREAK) {
				putmark (RET_BLANK, UTF_RETblank);
				count ++;
			}
			if (RET_BLANK2 && count <= XBREAK) {
				putmark (RET_BLANK2, UTF_RETblank2);
				count ++;
			}
		}
	    }
	}

	/* Clear the rest of the line if clear_line is True */
	if (clear_line == True) {
		if (can_clear_eol == True) {
			if (count <= XBREAK) {
				clear_eol ();
			}
		}
		else {
			while (count ++ <= XBREAK) {
				/* clear up to XMAX */
				putchar (' ');
			}
		}
	}
  }
}

/*
 * set_cursor_xy sets the cursor by either directly calling set_cursor
 * or, in the case of proportional font support, reprinting the line
 * up to the x position
 */
void
set_cursor_xy ()
{
  if (proportional == True) {
	set_cursor (0, y);
	if (x != 0) {
		put_line (cur_line, x, False, True);
	}
	/* cur_line may still be undefined if x == 0 */
  } else {
	set_cursor (x, y);
  }
}

/*
 * Proceed returns the count'th line after `line'. When count is negative
 * it returns the count'th line before `line'. When the next (previous)
 * line is the tail (header) indicating EOF (tof) it stops.
 */
LINE *
proceed (line, count)
  register LINE * line;
  register int count;
{
  if (count < 0) {
	while (count ++ < 0 && line != header) {
		line = line->prev;
	}
  } else {
	while (count -- > 0 && line != tail) {
		line = line->next;
	}
  }
  return line;
}

/*
 * Reset assigns bot_line, top_line and cur_line according to `head_line'
 * which must be the first line of the screen, and a y-coordinate,
 * which will be the current y-coordinate (if it isn't larger than last_y)
 */
void
reset (head_line, screen_y)
  LINE * head_line;
  int screen_y;
{
  register LINE * line;

  top_line = line = head_line;

/* Search for bot_line (might be last line in file) */
  for (last_y = 0; last_y < total_lines - 1 && last_y < SCREENMAX
					&& line->next != tail; last_y ++)
	line = line->next;

  bot_line = line;
  y = (screen_y > last_y) ? last_y : screen_y;

/* Set cur_line according to the new y value */
  cur_line = proceed (top_line, y);
  calc_line_no ();
}

/*
 * Display line at screen line y_pos if it lies between y_min and y_max.
 * If it is no text line (end of file), clear screen line.
 */
static
void
display_line_at (y_pos, line, y_min, y_max, first)
  int y_pos;
  int y_min;
  int y_max;
  register LINE * line;
  FLAG first;
{
  line = proceed (line, y_pos - y_min);
  if (y_pos >= y_min && y_pos <= y_max) {
	set_cursor (0, y_pos);
	if (line == tail) {
		clear_wholeline ();
	} else {
		if (first == False) {
			if (display_delay >= 0) {
				flush ();
			}
#ifdef msdos
			if (display_delay > 0) {
				delay (display_delay);
			}
#else
#ifdef use_usleep
			if (display_delay > 0) {
				(void) usleep (1000 * display_delay);
			}
#endif
#endif
		}
		line_print (line);
	}
  }
}

/*
 * Display () shows count + 1 lines on the terminal starting at the given 
 * coordinates. At end of file, the rest of the screen is blanked.
 * When count is negative, a backwards print from `line' will be done.
 */
void
display (y_coord, line, count, new_pos)
  int y_coord;
  int new_pos;
  register LINE * line;
  register int count;
{
  int y_max = y_coord + count;
  int y_off;

/* Find new startline if count is negative */
  if (count < 0) {
	line = proceed (line, count);
	count = - count;
  }

  clean_menus ();

  display_line_at (new_pos, line, y_coord, y_max, True);
  y_off = 0;
  while (y_off < count) {
	y_off ++;
	display_line_at (new_pos - y_off, line, y_coord, y_max, False);
	display_line_at (new_pos + y_off, line, y_coord, y_max, False);
  }
#ifdef UNUSED
/* old code, building the display from top to bottom (how boring): */
/* with this code, XBREAK must be set to XMAX - 1 */
/* Print the lines */
  set_cursor (0, y_coord);
  while (line != tail && count -- >= 0) {
	line_print (line);
	line = line->next;
  }

/* Print the blank lines (if any) */
  if (loading == False) {
	while (count -- >= 0) {
		clear_eol ();
		putchar ('\n');
	}
  }
#endif
}


/*======================================================================*\
|*			Terminal dialog					*|
\*======================================================================*/

/*
 * promptyn reads in a 'y' or 'n' character.
 */
character
promptyn ()
{
  register character c;
  while ((c = readchar_nokeymap ()) != 'y' && c != 'n' && c != '\033' && quit == False) {
	ring_bell ();
	flush ();
  }
  if (c == '\033') {
	quit = True;
  }
  return c;
}

#ifdef UNUSED
/*
 * In case of a QUIT signal, swallow the dummy char generated by catchquit ()
 * called by re_search () and change ()
 */
void
swallow_dummy_quit_char ()
{
  (void) readchar (); /* Swallow away a quit character delivered by QUIT */
/* Not needed because this character is ignored by being the CANCEL command */
}
#endif


/*======================================================================*\
|*			Scroll bar display				*|
\*======================================================================*/

/* mined displays last_y + 1 lines in a display area of YMAX lines */
/* the file has (approx.?) total_lines lines */
/* the file position was in line # line_number */

static int prev_disp_start = 0;
static int prev_disp_end = 0;

FLAG
display_scrollbar (update)
  FLAG update;
{
/* last_y / YMAX */
  int disp_start = ((long) line_number - y) * last_y / total_lines;
  int disp_end = ((long) line_number - y + last_y + 1) * last_y / total_lines;
  int i;
  FLAG painted;

#define debug_scrollbar
#ifdef debug_scrollbar
extern FLAG update_scrollbar_lazy;
#endif

  if (disp_scrollbar == True) {
    if (update == False ||
	(update == True && 
	(disp_start != prev_disp_start || disp_end != prev_disp_end))
       )
    {
#ifdef debug_scrollbar
if (update_scrollbar_lazy == True)
   printf ("scrollbar (%d) %d - %d (prev. %d - %d)\n", update, disp_start, disp_end, prev_disp_start, prev_disp_end);
/*printf (" - line %d of %d; cursor %d, last line %d (%d)\n", 
	line_number, total_lines, y, last_y, SCREENMAX);
*/
#endif

/* last_y / SCREENMAX */
	for (i = 0; i <= last_y; i ++) {
		set_cursor (XMAX, i);
		if (i >= disp_start && i <= disp_end) {
			if (Chinese == False) {
				disp_scrollbar_foreground ();
				putchar (' ');
				disp_scrollbar_off ();
			} else {
				reverse_on ();
				putchar ('<');
				reverse_off ();
			}
		} else {
			disp_scrollbar_background ();
			putchar (' ');
			disp_scrollbar_off ();
		}
	}
	painted = True;
    } else {
#ifdef debug_scrollbar
if (update_scrollbar_lazy == True)
   printf ("scrollbar (%d) nix\n", update);
#endif
	painted = False;
    }
  }

  prev_disp_start = disp_start;
  prev_disp_end = disp_end;
  return painted;
}


/*======================================================================*\
|*			Status line dialog and input			*|
\*======================================================================*/

/*
 * Display a line telling how many chars and lines the file contains. Also tell
 * whether the file is readonly and/or modified.
fstatus (mess, bytes, chars) is
	file_status ((mess), (bytes), file_name, \
		     total_lines, True, writable, modified, viewonly, (chars))
 */
/* directly called only from WB: file_status 
   (msg_done, chars_saved, file_name, lines_saved, False, True, False, False); */
void
file_status (message, bytecount, file, lines, textstat, writefl, changed, viewing, charcount)
  char * message;
  register long bytecount;	/* Contains number of bytes in file */
  register long charcount;	/* Contains number of characters in file */
  char * file;
  int lines;
  FLAG textstat;
  FLAG writefl;
  FLAG changed;
  FLAG viewing;
{
  register LINE * line;
  register int line_num = 0;
  static char msg [maxLINE_LEN + 40];	/* Buffer to hold line */
#ifdef buffer_info
  char yank_msg [maxLINE_LEN];	/* Buffer for msg of yank_file */
#endif

  /* determine information */
  if (bytecount < 0) {	/* # chars not given; count them */
	bytecount = 0L;
	charcount = 0L;
	for (line = header->next; line != tail; line = line->next) {
		bytecount += length_of (line->text);
		charcount += char_count (line->text);
		if (line->return_type == lineend_NONE) {
			charcount --;
			bytecount --;
		} else if (line->return_type == lineend_CRLF) {
			bytecount ++;
		} else if (line->return_type == lineend_LS
			|| line->return_type == lineend_PS) {
			bytecount += 2;
		}
		line_num ++;
		if (line == cur_line) {
			line_number = line_num;
		}
	}
  } else {		/* check for current line # only */
	for (line = header->next; line != cur_line->next; line = line->next) {
		line_num ++;
		if (line == cur_line) {
			line_number = line_num;
		}
	}
  }

#ifdef buffer_info
  if (yank_status == VALID && textstat == True) {
	/* Append buffer info */
	/* build_string (yank_msg, " Buffer: %ld char%s.", chars_saved,
					(chars_saved == 1L) ? "" : "s");
	*/
	yank_msg [0] = '\0';
	/* Empty paste buffer is only an initial condition and thus this 
	   would be no significant information; since buffer contents 
	   may be appended, the exact count is not available anyway. */
  } else {
	yank_msg [0] = '\0';
  }
#endif

  if (textstat == True) {
	build_string (msg,
		(lines_per_page != 0)
		? "%s %s%s%s%s, line %d of %d (chars/bytes: %ld/%ld), page %d"
		: "%s %s%s%s%s, line %d of %d (chars/bytes: %ld/%ld)",
		message,
		(rpipe == True && * message != '[') ?
			"standard input" : file,
		(viewing == True) ? " (View only)" : "",
		(changed == True) ? " (modified)" : "",
		(writefl == False) ? " (Readonly)" : "",
		line_number,
		lines,
		charcount,
		bytecount,
		(lines_per_page != 0)
			? (line_number - 1) / lines_per_page + 1
			: 1
#ifdef buffer_info
		, yank_msg
#endif
		);
  } else {
	build_string (msg,
		"%s %s%s%s%s, lines: %d (chars/bytes: %ld/%ld)",
		message,
		(rpipe == True && * message != '[') ?
			"standard input" : file,
		(viewing == True) ? " (View only)" : "",
		(changed == True) ? " (modified)" : "",
		(writefl == False) ? " (Readonly)" : "",
		lines,
		charcount,
		bytecount
		);
  }

  status_msg (msg);	/* display the information */
}

/*
 * Just count to determine the current line number
 */
void
calc_line_no ()
{
  register LINE * line;
  register int line_num = 0;

	for (line = header->next; line != cur_line; line = line->next) {
		line_num ++;
	}
	line_number = line_num + 1;
}

/*
 * Input () reads a string from the terminal.
 * Return values:
 *	when QUIT character typed => ERRORS
 *	when empty input and clearfl == True: NO_INPUT
 *	else: FINE
 */
static
character *
input_byte (inbuf, ptr, c)
  character * inbuf;
  character * ptr;
  character c;
{
	if (c > '\0' && (ptr - inbuf) < maxLINE_LEN) {
		* ptr ++ = c;
		* ptr = '\0';
		print_char (c);
		putstring (" \b");
	} else {
		ring_bell ();
	}
	return ptr;
}

static
character *
input_unicode (inbuf, ptr, unichar)
  character * inbuf;
  character * ptr;
  unsigned long unichar;
{
	character c;

	if (unichar < 0x80) {
	    if ((ptr - inbuf) >= maxLINE_LEN) {
		ring_bell ();
	    } else {
		* ptr ++ = unichar;
		print_char (unichar);
		* ptr = '\0';
	    }
	} else if (unichar < 0x800) {
	    if ((ptr - inbuf) + 2 > maxLINE_LEN) {
		ring_bell ();
	    } else {
		c = 0xC0 | (unichar >> 6);
		* ptr ++ = c;
		print_char (c);
		c = 0x80 | (unichar & 0x3F);
		* ptr ++ = c;
		print_char (c);
		* ptr = '\0';
	    }
	} else {
	    if ((ptr - inbuf) + 3 > maxLINE_LEN) {
		ring_bell ();
	    } else {
		c = 0xE0 | (unichar >> 12);
		* ptr ++ = c;
		print_char (c);
		c = 0x80 | ((unichar >> 6) & 0x3F);
		* ptr ++ = c;
		print_char (c);
		c = 0x80 | (unichar & 0x3F);
		* ptr ++ = c;
		print_char (c);
		* ptr = '\0';
	    }
	}
	putstring (" \b");
	return ptr;
}

static
character *
input_character (inbuf, ptr, c)
  character * inbuf;
  character * ptr;
  unsigned long c;
{
	if (utf8_text == True) {
		return input_unicode (inbuf, ptr, c);
	} else if ((c & 0xFF) == c) {
		return input_byte (inbuf, ptr, c);
	} else {
		return input_byte (inbuf, ptr, '');
	}
}

static
void
redraw_prompt ()
{
	set_cursor (0, 0);
	while (bottom_lines > 0) {
		scroll_reverse ();
		bottom_lines --;
	}
	rd_bottom_line ();
	if (can_clear_eol == True) {
		clear_eol ();
	}
}

static
int
input (inbuf, clearfl, term_input)
  character * inbuf;
  FLAG clearfl;
  char * term_input;
{
  character * ptr;
  character * inputptr;
  register character c;
  char * term_chars;

  unsigned long unichar;
  char mnemonic [maxLINE_LEN];
  char * mnemo;
  character c1;
  int utfcount;
  int i;

  ptr = inbuf;
  * ptr = '\0';
  bottom_lines = 0;

  while (quit == False) {
     if (lpos >= XBREAK_bottom) {
	pagewrapped = True;
     }
     if (lpos < 0) {
	redraw_prompt ();
     }

     flush ();
     c = readchar ();
     for (term_chars = term_input; * term_chars != '\0'; term_chars ++) {
	if (c == * term_chars) {
		c = '\n';
	}
     }

     switch (c) {

	case '\b' :		/* Erase previous char */
	case '\177' /* DEL */ :
	    if (ptr > inbuf) {
		ptr --;
		diagdisp_off ();
		if (iscontrol (* ptr)) {
			putstring (" \b\b\b  \b\b");
			lpos = lpos - 2;
		} else if (Chinese == True && inChinese (inbuf, ptr)) {
			ptr --;
			putstring (" \b\b\b  \b\b");
			lpos = lpos - 2;
		} else if (utf8_text == True && (* ptr & 0xC0) == 0x80) {
			/* UTF-8 sequence byte found; look for start */
			ptr ++;
			precede_char (& ptr, inbuf);
			utf8_info (ptr, & utfcount, & unichar);
			while (combining_mode == True 
				&& ptr != inbuf && iscombining (unichar))
			{
				precede_char (& ptr, inbuf);
				utf8_info (ptr, & utfcount, & unichar);
			}
			if (! iscombining (unichar)) {
				if (iswide (unichar)) {
					putstring (" \b\b\b  \b\b");
					lpos = lpos - 2;
				} else {
					putstring (" \b\b \b");
					lpos = lpos - 1;
				}
			}
		} else {
			ptr ++;
			precede_char (& ptr, inbuf);
			putstring (" \b\b \b");
			lpos = lpos - 1;
		}
		diagdisp_on ();
		putstring (" \b");
		* ptr = '\0';
	    } else {
		ring_bell ();
	    }
	    break;

	case QUITCHAR :
	case '\033' :
		quit = True;
		break;

	case '\n' :		/* End of input */
	case '\015' :
		/* If inbuf is empty clear status_line */
		return (ptr == inbuf && clearfl == True) ?
			NO_INPUT : FINE;

	case (character) FUNcmd:
		if (keyproc == insert_diaeresis) {
			ptr = input_character (inbuf, ptr, diaeresis (readchar ()));
		} else if (keyproc == insert_acute) {
			ptr = input_character (inbuf, ptr, acute (readchar ()));
		} else if (keyproc == insert_grave) {
			ptr = input_character (inbuf, ptr, grave (readchar ()));
		} else if (keyproc == insert_circumflex) {
			ptr = input_character (inbuf, ptr, circumflex (readchar ()));
		} else if (keyproc == insert_tilde) {
			ptr = input_character (inbuf, ptr, tilde (readchar ()));
		} else if (keyproc == insert_angstrom) {
			ptr = input_character (inbuf, ptr, angstrom (readchar ()));
		} else if (allow_keymap && keyproc == toggleKEYMAP) {
			reverse_off ();
			toggleKEYMAP ();
			redraw_prompt ();
			break;
		} else if (allow_keymap && keyproc == setupKEYMAP) {
			reverse_off ();
			setupKEYMAP ();
			redraw_prompt ();
			break;
		} else {
			return ERRORS;
		}

	default :
		if (c == control_prefix /* ^V/^P */) {
			c = readchar ();
			if (c == ring /*|| c == ','*/) {
				ptr = input_character (inbuf, ptr, angstrom (readchar ()));
			} else switch (c) {
			/*case ':':*/
			case '"':	{
					 ptr = input_character (inbuf, ptr, diaeresis (readchar ()));
					 break;
					}
			case '\'':	{
					 ptr = input_character (inbuf, ptr, acute (readchar ()));
					 break;
					}
			case '`':	{
					 ptr = input_character (inbuf, ptr, grave (readchar ()));
					 break;
					}
			case '^':	{
					 ptr = input_character (inbuf, ptr, circumflex (readchar ()));
					 break;
					}
			case '~':	{
					 ptr = input_character (inbuf, ptr, tilde (readchar ()));
					 break;
					}
			default:
			    if (utf8_text == True) {
				if (c == '#') {
				  i = 0;
				  unichar = 0;
				  while (i < 4 && quit == False) {
					c = readchar ();
					if (c == QUITCHAR || c == '\033') {
						quit = True;
					}

					if (c >= '0' && c <= '9') {
						i ++;
						unichar *= 16;
						unichar += c - '0';
					} else if (c >= 'A' && c <= 'F') {
						i ++;
						unichar *= 16;
						unichar += c - 'A' + 10;
					} else if (c >= 'a' && c <= 'f') {
						i ++;
						unichar *= 16;
						unichar += c - 'a' + 10;
					} else {
						quit = True;
					}
				  }
				  if (quit == True) {
					ring_bell ();
					break;
				  }
				} else if (c == ' ') {
					/* read mnemonic ... */
					mnemo = mnemonic;
					while (c != '\n' && quit == False) {
						c = readchar ();
						if (c == QUITCHAR || c == '\033') {
							quit = True;
						}

						if (c == ' ') {
							c = '\n';
						}
						if (c != '\n') {
							* mnemo ++ = c;
						}
					}
					* mnemo = '\0';
					if (quit == True) {
						ring_bell ();
						break;
					}

					unichar = compose_mnemo (mnemonic);
				} else if ( ((c > ' ') && (c <= 'Z'))
				            || (c > '^') ) {
					unichar = compose (c, readchar ());
				} else if (c > '\0' 
				           && (ptr - inbuf) < maxLINE_LEN) {
					* ptr ++ = c;
					* ptr = '\0';
					print_char (c);
					putstring (" \b");
					break;
				} else {
					ring_bell ();
					break;
				}

				/* enter unichar as utf-8 ... */
				inputptr = ptr;
				if (unichar < 0x80) {
				    if ((ptr - inbuf) >= maxLINE_LEN) {
					ring_bell ();
				    } else {
					* ptr ++ = unichar;
					print_char (unichar);
					* ptr = '\0';
				    }
				} else if (unichar < 0x800) {
				    if ((ptr - inbuf) + 2 > maxLINE_LEN) {
					ring_bell ();
				    } else {
					c = 0xC0 | (unichar >> 6);
					* ptr ++ = c;
					print_char (c);
					c = 0x80 | (unichar & 0x3F);
					* ptr ++ = c;
					print_char (c);
					* ptr = '\0';
				    }
				} else {
				    if ((ptr - inbuf) + 3 > maxLINE_LEN) {
					ring_bell ();
				    } else {
					c = 0xE0 | (unichar >> 12);
					* ptr ++ = c;
					print_char (c);
					c = 0x80 | ((unichar >> 6) & 0x3F);
					* ptr ++ = c;
					print_char (c);
					c = 0x80 | (unichar & 0x3F);
					* ptr ++ = c;
					print_char (c);
					* ptr = '\0';
				    }
				}
				if (combining_mode == True) {
				    if (iscombining (unichar)) {
					/* refresh display from base char */
					do {
						precede_char (& inputptr, inbuf);
						utf8_info (inputptr, & utfcount, & unichar);
					} while (inputptr != inbuf
						 && iscombining (unichar));
					if (! iscombining (unichar)) {
						if (iswide (unichar)) {
							putstring ("\b\b");
						} else {
							putstring ("\b");
						}
					}
					putstring (inputptr);
				    }
				}
				putstring (" \b");
			    } else {
				if ( ((c > ' ') && (c <= 'Z'))
				            || (c > '^') ) {
					c = compose (c, readchar ());
				} else {
					if (c != '\177') c = c & '\237';
				}

				if (c > '\0' 
				    && (ptr - inbuf) < maxLINE_LEN) {
					* ptr ++ = c;
					* ptr = '\0';
					print_char (c);
					putstring (" \b");
				} else {
					ring_bell ();
				}
			    }
			}

		} else if (utf8_text == True && multichar (c)) {
		   if (utf8_input == True && last_char_latin1 == False) {
		/* UTF-8 input for UTF-8 text */
			if ((c & 0xE0) == 0xC0)
				utfcount = 2;
			else if ((c & 0xF0) == 0xE0)
				utfcount = 3;
			else if ((c & 0xF8) == 0xF0)
				utfcount = 4;
			else if ((c & 0xFC) == 0xF8)
				utfcount = 5;
			else if ((c & 0xFE) == 0xFC)
				utfcount = 6;
			else /* ignore illegal UTF-8 code */
				utfcount = 0;
			if (utfcount > 0
			    && (ptr - inbuf) + utfcount <= maxLINE_LEN)
			{
				inputptr = ptr;
				* ptr ++ = c;
				print_char (c);
				utfcount --;
				while (utfcount > 0) {
					c = readchar ();
					* ptr ++ = c;
					print_char (c);
					utfcount --;
				}
				* ptr = '\0';
				if (combining_mode == True) {
				    utf8_info (inputptr, & utfcount, & unichar);
				    if (iscombining (unichar)) {
					/* refresh display from base char */
					do {
						precede_char (& inputptr, inbuf);
						utf8_info (inputptr, & utfcount, & unichar);
					} while (inputptr != inbuf
						 && iscombining (unichar));
					if (! iscombining (unichar)) {
						if (iswide (unichar)) {
							putstring ("\b\b");
						} else {
							putstring ("\b");
						}
					}
					putstring (inputptr);
				    }
				}
				putstring (" \b");
			} else {
				ring_bell ();
			}
		   } else {
		/* 8-bit input for UTF-8 text */
		      if ((ptr - inbuf) + 1 < maxLINE_LEN) {
			c1 = 0xC0 | (c >> 6);
			* ptr ++ = c1;
			print_char (c1);
			c1 = 0x80 | (c & 0x3F);
			* ptr ++ = c1;
			print_char (c1);
			* ptr = '\0';
			putstring (" \b");
		      } else
			ring_bell ();
		   }

		} else if (utf8_input == True && multichar (c) 
				&& last_char_latin1 == False) {
		/* UTF-8 input for 8-bit text */
			if ((c & 0xE0) == 0xC0) {
				utfcount = 2;
				unichar = c & 0x1F;
			} else if ((c & 0xF0) == 0xE0) {
				utfcount = 3;
				unichar = c & 0x0F;
			} else if ((c & 0xF8) == 0xF0) {
				utfcount = 4;
				unichar = c & 0x07;
			} else if ((c & 0xFC) == 0xF8) {
				utfcount = 5;
				unichar = c & 0x03;
			} else if ((c & 0xFE) == 0xFC) {
				utfcount = 6;
				unichar = c & 0x01;
			} else /* ignore illegal UTF-8 code */
				utfcount = 0;
			if (utfcount > 0
			    && (ptr - inbuf) < maxLINE_LEN) {
				utfcount --;
				while (utfcount > 0) {
					c = readchar ();
					unichar = (unichar << 6) | (c & 0x3F);
					utfcount --;
				}
				c = unichar & 0xFF;
				if (c == unichar) {
					* ptr ++ = c;
					print_char (c);
					* ptr = '\0';
					putstring (" \b");
				} else
					ring_bell ();
			} else
				ring_bell ();

		} else if (Chinese == True && multichar (c)) {
		/* 8/16-bit (Chinese) input for Chinese text */
		   if ((ptr - inbuf) + 1 < maxLINE_LEN) {
			* ptr ++ = c;
			print_char (c);
			c = readchar ();
			* ptr ++ = c;
			print_char (c);
			* ptr = '\0';
			putstring (" \b");
		   }
		   else
			ring_bell ();

		} else if ((ptr - inbuf) < maxLINE_LEN) {
		/* 8-bit input for 8-bit text */
			if ((c > '\0')) {
				* ptr ++ = c;
				* ptr = '\0';
				print_char (c);
				putstring (" \b");
			}
			else
				ring_bell ();

		} else
			ring_bell ();
     }
  }
  quit = False;
  return ERRORS;
}

/*
 * Show concatenation of s1 and s2 on the status line (bottom of screen)
 * If revfl is True, turn on reverse video on both strings. Set stat_visible
 * only if bottom_line is visible.
 * The return value is FINE except for get_string, where it is taken
 * from the call to input ().
status_line (str1, str2) is (void) bottom_line (ON, (str1), (str2), NIL_PTR, False, "")
status_msg (str)	 is status_line (str, NIL_PTR)
status_beg (str)	 is (void) bottom_line (ON, (str), NIL_PTR, NIL_PTR, True, "")
error (str1, str2)	 is (void) bottom_line (ON, (str1), (str2), NIL_PTR, False, "")
clear_status ()		 is (void) bottom_line (OFF, NIL_PTR, NIL_PTR, NIL_PTR, False, "")
get_string (str1, str2, fl, term_chars)
			 is bottom_line (ON, (str1), NIL_PTR, (str2), fl, term_chars)
 */
FLAG lastrevfl;
char * lastinbuf;
FLAG input_active = False;
char status_buf [maxLINE_LEN];

void
rd_bottom_line ()
{
  set_cursor (0, YMAX);
  diagdisp_on ();
  if (lastinbuf == NIL_PTR) {
	printlim_string (status_buf, XBREAK_bottom);
  } else {
	lpos = 0;
	print_string (status_buf);
	print_string (lastinbuf);
  }
  if (input_active == False) {
	diagdisp_off ();
	set_cursor_xy ();	/* Set cursor back to old position */
  }
  flush ();	/* Perform the actual screen output */
}

int
bottom_line (revfl, s1, s2, inbuf, statfl, term_input)
  FLAG revfl;
  char * s1;
  char * s2;
  char * inbuf;
  FLAG statfl;
  char * term_input;
{
  int ret = FINE;

  if (inbuf != NIL_PTR) {
	* inbuf = '\0';
  }
  lastrevfl = revfl;
  lastinbuf = inbuf;

  if (pagewrapped == True) {
	status_buf [0] = '\0';
	RD ();
	pagewrapped = False;
  }

  build_string (status_buf, " %s%s ", unnull (s1), unnull (s2));
		/* (s1 == NIL_PTR) ? "" : s1, (s2 == NIL_PTR) ? "" : s2); */

  if (revfl == ON && stat_visible == True) {
	set_cursor (0, YMAX);
	clear_lastline ();
  }
  set_cursor (0, YMAX);
  if (revfl == ON) {		/* Print rev. start sequence */
	diagdisp_on ();
	stat_visible = True;
  } else {			/* Used as clear_status () */
	diagdisp_off ();
	stat_visible = False;
  }

  if (inbuf == NIL_PTR) {
	printlim_string (status_buf, XBREAK_bottom);
  } else {
	lpos = 0;
	print_string (status_buf);
	input_active = True;
	ret = input (inbuf, statfl, term_input);
	input_active = False;
  }

  /* Print normal video */
  diagdisp_off ();
  set_cursor (lpos, YMAX);
  if (can_clear_eol == True) {
	clear_eol ();
  } else {
	put_blanks (XMAX - 1 - lpos);
	set_cursor (lpos, YMAX);
  }

  if (inbuf != NIL_PTR) {
	set_cursor (0, YMAX);
  } else if (statfl == True) {
	diagdisp_on ();
  } else {
	set_cursor_xy ();	/* Set cursor back to old position */
  }

  flush ();	/* Perform the actual screen output */
  if (ret != FINE) {
	clear_status ();
  }
  return ret;
}

/*
 * get_number () reads a number from the terminal.
 * get_hex_number () reads a hexadecimal number from the terminal.
 * If the parameter firstdigit is a valid digit, empty input (return) 
 * will yield a valid number; if it is '\0', it won't.
 * The last character typed in is returned.
 * ERRORS is returned on a bad number or on interrupted input.
 * The resulting number is put into the integer the argument points to.
 */
int
get_number (message, firstdigit, result)
  char * message;
  char firstdigit;
  int * result;
{
  register int index;
  register int count;

  status_beg (message);

  if (firstdigit > '\0') {
	index = firstdigit;
  } else {
	index = readchar ();
  }

  if (index == QUITCHAR || index == '\033') {
	quit = True;
  }
  if (quit == False && (index < '0' || index > '9')) {
	error ("Bad number", NIL_PTR);
	return ERRORS;
  }

/* Convert input to a decimal number */
  count = 0;
  while (index >= '0' && index <= '9' && quit == False) {
	print_char (index); flush ();
	if (lpos >= XBREAK_bottom) {
		pagewrapped = True;
	}
	count *= 10;
	count += index - '0';
	index = readchar ();
	if (index == QUITCHAR || index == '\033') {
		quit = True;
	}
  }

  clear_status ();
  if (quit == True) {
	clear_status ();
	return ERRORS;
  }
  * result = count;
  return index;
}

int
get_hex_number (message, result)
  char * message;
  unsigned long * result;
{
  register int index;
  register int count = 0;
  int i;
  FLAG ok = False;
  int base = 16;

  status_beg (message);

  i = 0;
  while (i < 9 && ok == False) {
	index = readchar ();
	if (i == 0 && index == '#') {
		base = 8;
		status_beg ("Enter octal value ...");
		index = readchar ();
	}

	if (index >= '0' && index <= '9' && (index <= '7' || base == 16)) {
		i ++;
		count *= base;
		count += index - '0';
		print_char (index); flush ();
	} else if (base == 16 && index >= 'A' && index <= 'F') {
		i ++;
		count *= base;
		count += index - 'A' + 10;
		print_char (index); flush ();
	} else if (base == 16 && index >= 'a' && index <= 'f') {
		i ++;
		count *= base;
		count += index - 'a' + 10;
		index = index - 'a' + 'A';
		print_char (index); flush ();
	} else if (index == '\b' || index == '\177') {
		if (i > 0) {
			i --;
			count /= base;
			putstring ("\b \b"); flush ();
			if (lpos >= XBREAK_bottom) {
				pagewrapped = True;
			}
			lpos --;
		}
	} else if (index == ' ' || index == '\n' || index == '\r') {
		ok = True;
	} else {
		quit = True;
		ok = True;
	}

	if (lpos >= XBREAK_bottom) {
		pagewrapped = True;
	}
  }

  clear_status ();
  if (quit == True) {
	if (base == 16) {
		error ("Bad hex value", NIL_PTR);
	} else {
		error ("Bad octal value", NIL_PTR);
	}
	quit = False;
	return ERRORS;
  }
  * result = count;
  return index;
}

/*
 * get_digits () reads in a number. In contrast to get_number, it does no
 * echoing, no messaging, and it does not require any digits at all.
 * The last character typed in is returned.
 * The resulting number is put into the integer the arguments points to.
 * It is for use within I/O routines (e.g. reading mouse escape sequences 
 * or cursor position report) and thus calls _readchar ().
 */
int
get_digits (result)
  int * result;
{
  register int index;
  register int count;

  index = _readchar ();
  if (index == QUITCHAR) {
	quit = True;
  }
  * result = -1;
/* Convert input to a decimal number */
  count = 0;
  while (index >= '0' && index <= '9' && quit == False) {
	count *= 10;
	count += index - '0';
	* result = count;
	index = _readchar ();
	if (index == QUITCHAR) {
		quit = True;
	}
  }

  if (quit == True) {
	return QUITCHAR;
  }
  return index;
}

/*
 * Get_file () reads a filename from the terminal.
 */
int
get_file (message, file)
  char * message;
  char * file;
{
  int ret = get_string (message, file, True, "");
#ifndef msdos
  char * filei;
  char file1 [maxLINE_LEN];

  if (file [0] == '~' && file [1] == '/') {
	filei = file; filei ++;
	build_string (file1, "%s%s", unnull (getenv ("HOME")), filei);
	build_string (file, file1);
  }
#endif
  return ret;
}


/*======================================================================*\
|*				End					*|
\*======================================================================*/
